
#include "stdafx.h"

#ifndef _WIN32
#include <stdarg.h>
#endif

#include "grt_manager.h"
#include "grt_shell.h"
#include "myx_shared_util_functions.h"


using namespace grt;
using namespace bec;


#define SNIPPETS_FILENAME "shell_snippets.txt"
#define HISTORY_FILENAME "shell_history.txt"



static void _free(gpointer data, gpointer udata)
{
  g_free(data);
}


ShellBE::ShellBE(GRTManager *grtm, GRTDispatcher *dispatcher)
  : _grtm(grtm), _dispatcher(dispatcher)
{
  _grt= grtm->get_grt();

  _shell= 0;

  _save_history= false;
  _completer= NULL;
  _completer_items= NULL;

  g_static_mutex_init(&_text_queue_mutex);

  _history_ptr= _history.begin();
}


ShellBE::~ShellBE()
{
  store_history();

  if (_completer_items)
  {
    g_list_foreach(_completer_items, _free, NULL);
    g_list_free(_completer_items);
  }
  if (_completer)
    g_completion_free(_completer);
}


void ShellBE::set_save_directory(const std::string &path)
{
  _savedata_dir= path;
}


void ShellBE::enable_completion()
{
  if (!_completer)
    _completer= g_completion_new(NULL);
  
  update_completer();
}


grt::ValueRef ShellBE::get_shell_variable(const std::string &varname)
{
  return _shell->get_global_var(varname);
}


void ShellBE::set_saves_history(bool flag)
{
  _save_history= flag;
  if (!flag)
  {
    _history.clear();
    _history_ptr= _history.end();
  }
}


void ShellBE::run_script_grt(grt::GRT *grt, const std::string &path)
{
  int rc= _shell->run_file(path, false);

  if (rc != 0)
    throw grt::grt_runtime_error("Error running script "+path,
                                 "There was an error during execution of GRT script "+path);
}




void ShellBE::process_line_async(const std::string &line)
{
  GRTShellTask *task= new GRTShellTask("User shell command", _dispatcher, line);

  task->signal_message().connect(sigc::mem_fun(this, &ShellBE::handle_msg));

  task->signal_finished().connect(sigc::bind<std::string>(sigc::mem_fun(this, &ShellBE::shell_finished_cb), line));

  _dispatcher->add_task(task);
}


void ShellBE::shell_finished_cb(ShellCommand result, const std::string &prompt, const std::string &line)
{
  if (result == ShellCommandExit)
  {
    _grtm->terminate();
    _current_statement.clear();
  }
  else if (result == ShellCommandUnknown)
  { 
    if (_current_statement.empty())
      _current_statement= line;
    else
      _current_statement+= "\n"+line;
  }
  else if (result == ShellCommandStatement)
  {
    if (_current_statement.empty())
      _current_statement= line;
    else
      _current_statement+= "\n"+line;

    if (_save_history && _current_statement != "\n" && _current_statement != "")
      save_history_line(_current_statement);

    _current_statement.clear();
  }
  else
  {
    if (_current_statement.empty())
      _current_statement= line;
    else
      _current_statement+= "\n"+line;

    if (_save_history && _current_statement != "\n" && _current_statement != "")
      save_history_line(_current_statement);

    _current_statement.clear();
  }

  if (_ready_slot)
    _ready_slot(prompt);
}



bool ShellBE::setup()
{
  if (!_grt->init_shell("lua"))
    return false;

  _shell= _grt->get_shell();
  _grt->get_shell()->set_disable_quit(false);

  _shell->print_welcome();

  start();

  return true;
}


void ShellBE::start()
{
  restore_history();

  // trigger a command that will in turn make the GUI receive a ready
  // callback with the prompt that should be shown
  process_line_async("print(\"Ready.\\n\")");
}


void ShellBE::save_history_line(const std::string &line)
{
  //remove empty cmd if it is on top of the stack
  if (!_history.empty() && (*_history.begin()).empty())
    _history.pop_front();

  _history.push_front(line);
  _history_ptr= _history.begin();
}


void ShellBE::clear_history()
{
  _history.clear();
  _history_ptr= _history.end();
}



bool ShellBE::previous_history_line(const std::string &current_line,
                                    std::string &line)
{
  if (_history_ptr == _history.end())
    return false;

  // save the current line to the history
  if (_history_ptr == _history.begin())
    save_history_line(current_line);

  std::list<std::string>::const_iterator tmp= _history_ptr;
  ++tmp;
  if (tmp == _history.end())
    return false;

  ++_history_ptr;
  line= *_history_ptr;

  return true;
}


bool ShellBE::next_history_line(std::string &line)
{
  if (_history_ptr != _history.begin())
  {
    --_history_ptr;
    line= *_history_ptr;
    if (_history_ptr == _history.begin())
    {
      _history.pop_front();
      _history_ptr= _history.begin();
    }
    return true;
  }
  return false;
}

void ShellBE::reset_history_position()
{
  _history_ptr= _history.begin();
}


void ShellBE::set_ready_handler(const sigc::slot<void,const std::string&> &slot)
{
  _ready_slot= slot;
}


void ShellBE::set_output_handler(const sigc::slot<void, const std::string&> &slot)
{
  _output_slot= slot;
}



void ShellBE::handle_msg(const Message &msg)
{
  // the default msg handler outputs everything to shell
  switch (msg.type)
  {
  case grt::ErrorMsg:
    write_line("ERROR: "+msg.text+"\n");
    break;
  case grt::WarningMsg:
    write_line("WARNING: "+msg.text+"\n");
    break;
  case grt::InfoMsg:
    write_line("INFO: "+msg.text+"\n");
    break;
  case grt::ProgressMsg:
    write_line("Progress: "+msg.text+"\n");
    break;
  case grt::OutputMsg:
    write(msg.text);
    break;
  default:
    write_line("Message: "+msg.text+"\n");
    break;
  }
}


void ShellBE::writef(const char *fmt, ...)
{
  va_list ap;
  char *tmp;
  std::string line;

  va_start(ap, fmt);
  tmp= g_strdup_vprintf(fmt, ap);
  line= tmp;
  g_free(tmp);
  va_end(ap);

  if (_grtm->is_threaded())
  {
    g_static_mutex_lock(&_text_queue_mutex);
    _text_queue.push_back(line);
    g_static_mutex_unlock(&_text_queue_mutex);

    // if we're in the main thread, flush the message queue
    if (_grtm->in_main_thread())
      flush_shell_output();
  }
  else // if not threaded, just print directly
    _output_slot(line);
}


void ShellBE::write_line(const std::string &line)
{
  writef("%s\n", line.c_str());
}


void ShellBE::write(const std::string &text)
{
  writef("%s", text.c_str());
}


void ShellBE::flush_shell_output()
{
  std::string line;

  g_static_mutex_lock(&_text_queue_mutex);

  while (!_text_queue.empty())
  {
   line= _text_queue.front();
    _text_queue.pop_front();
    g_static_mutex_unlock(&_text_queue_mutex);

    _output_slot(line);

    g_static_mutex_lock(&_text_queue_mutex);
  }

  g_static_mutex_unlock(&_text_queue_mutex);
}



void ShellBE::update_completer()
{
  if (_completer)
  {
    g_completion_clear_items(_completer);

    g_list_foreach(_completer_items, _free, NULL);
    g_list_free(_completer_items);

    _completer_items= _shell->get_tokens();

    g_completion_add_items(_completer, _completer_items);
  }
}


GList *ShellBE::complete_word(const std::string &prefix, std::string &nprefix)
{
  if (_completer)
  {
    char *tmp= NULL;
    GList *list;

    list= g_completion_complete(_completer, prefix.c_str(), &tmp);
    if (tmp)
      nprefix= tmp;
    g_free(tmp);

    return list;
  }
  return NULL;
}



std::string ShellBE::get_snippet_data()
{
  std::string path= make_path(_savedata_dir, SNIPPETS_FILENAME);
  gchar *contents;
  gsize length;
  
  if (g_file_get_contents(path.c_str(), &contents, &length, NULL))
  {
    std::string text= std::string(contents, contents+length);
    g_free(contents);
    return text;
  }
  return "";
}


void ShellBE::set_snippet_data(const std::string &data)
{
  std::string path= make_path(_savedata_dir, SNIPPETS_FILENAME);

  // Make sure path exists, if not create it with privileges 755
  g_mkdir_with_parents(_savedata_dir.c_str(), 0755);

  if (!g_file_set_contents(path.c_str(), data.c_str(), data.size(), NULL))
  {
    throw std::runtime_error("Could not save file "+path);
  }
}


void ShellBE::store_history()
{
  std::string path= make_path(_savedata_dir, HISTORY_FILENAME);

  // Make sure path exists, if not create it with privileges 744
  g_mkdir_with_parents(_savedata_dir.c_str(), 0755);

  FILE *f= myx_fopen(path.c_str(), "w+");
  if (!f)
    throw std::runtime_error("Could not save file "+path);

  for (std::list<std::string>::const_iterator i= _history.begin();
       i != _history.end(); ++i)
  {
    char **lines= g_strsplit(i->c_str(), "\n", 0);
    for (int j= 0; lines[j]; j++)
      fprintf(f, " %s\n", lines[j]);
    g_strfreev(lines);
    fprintf(f, "\n");
  }
  fclose(f);
}


void ShellBE::restore_history()
{
  char line[1024];
  std::string path= make_path(_savedata_dir, HISTORY_FILENAME);
  std::string command;
  FILE *f= myx_fopen(path.c_str(), "r");
  if (!f)
    return;

  _history.clear();

  while (!feof(f) && fgets(line, sizeof(line), f))
  {
    if (*line==' ')
      command+= line+1;
    else
    {
      _history.push_back(command.substr(0, command.size()-1));
      command= "";
    }
  }

  fclose(f);

  _history_ptr= _history.begin();
}

