#ifndef _TREE_MODEL_H_
#define _TREE_MODEL_H_

/*
 * TreeModel is the base class for all list or tree based
 * backend classes. 
 */
#include <algorithm>
//#include <list>
//#include <vector>

#include <grtpp.h>
#include "grt/icon_manager.h"
#include "grt/common.h"
#include "grt/ui_form.h" // for menu stuff

#include "grt_backend_public_interface.h"
#include <ctype.h>

#include <set>
#include <algorithm>

namespace bec
{
  /** A tree node index.
   * Used to index nodes in a tree or list backend that inherits from ListModel or TreeModel.
   * A nodeId is like an index, in the simplest case of a flat list, it will contain
   * a single integer index. For trees, it will contain one index for each parent node
   * until the leaf node it refers to.
   * 
   * Since nodeIds are just indices, a nodeId may not point to the same item after
   * the list/tree contents change.
   * 
   * @ingroup begrt
   */ 

  /**
  */  
  // TODO: Add threshold so we do not accumulate megabytes of allocated index vectors
  template <typename T>
  class MYSQLGRTBACKEND_PUBLIC_FUNC Pool
  {
    private:
      Pool(const Pool&)
      {
        throw std::runtime_error("Copy of Pool forbidden");
      }
      Pool& operator=(const Pool&)
      {
        throw std::runtime_error("Assignment of Pool forbidden");
      }
    public:
      static void delete_object(T* obj)
      {
        delete obj;
      }
  
      Pool()
          : _pool(4)
          , _sync(g_mutex_new())
      {}
    
      ~Pool()
      {
        {
          GMutexLock lock(_sync);
          std::for_each(_pool.begin(), _pool.end(), Pool::delete_object);
        }
        
        g_mutex_free(_sync);
      }
    
      T* get()
      {
        T* item = 0;
        try
        {
          GMutexLock lock(_sync);
          if ( _pool.size() > 0 )
          { 
            item = _pool.back();
            _pool.pop_back();
          }
        }
        catch (...)
        {
          //TODO: check when pop may throw
        }
      
        if ( !item )
        {
          item = new T;
        }
      
        return item;
      }
    
      void put(T* item)
      {
         GMutexLock lock(_sync);
        _pool.push_back(item);
      }
  
    private:
      std::vector<T*>  _pool;
      GMutex          *_sync; //! Protect against concurrent access within threads
  };

  /**
    \class NodeId
    \brief descibes path to a node starting from root for a Tree or it will contain an index (single entry) for List

    Imagine we have a tree with a two root nodes and two children of each root. So to address first child of the
    first root node we need to have a path like that: "0.0". To address the second child of the first root node:
    "0.1"
  */
  struct MYSQLGRTBACKEND_PUBLIC_FUNC NodeId 
  {
    typedef std::string*                uid;   //!< To map short-living NodeId path to a persistent value
                                               //!< This is needed for Gtk::TreeModel iterators
    typedef std::vector<short>          Index; 
    static Pool<Index>                 *_pool; //!< Pool of allocated std::vectors (Index)
    Index                              *index; //!< Path itself

    static Pool<Index>* pool()
    {
      return _pool ? _pool : (_pool = new Pool<Index>);
    }

    // NodeId will get index from the pool
    NodeId()
        : index(0)
    {
      index = pool()->get();
    }

    NodeId(const NodeId &copy)
        : index(0)
    {
      index = pool()->get();
      if ( copy.index )
        *index = *copy.index;
    }

    NodeId(const short i)
        : index(0)
    {
      index = pool()->get();
      index->push_back(i);
    }

    NodeId(const std::string &str)
        : index(0)
    {
      index = pool()->get();
      try
      {
        const char* chr = str.c_str();
        const int size = (int)str.length();
        std::string num;
        num.reserve(size);
        
        for ( int i = 0; i < size; i++ )
        {
          if ( isdigit(chr[i]) )
            num.push_back(chr[i]);
          else if ( '.' == chr[i] || ':' == chr[i] )
          {
            if ( !num.empty() )
            {
              index->push_back(atoi(num.c_str()));
              num.clear();
            }
          }
          else
            throw std::runtime_error("Wrong format of NodeId");
        }
        
        if ( !num.empty() )
          index->push_back(atoi(num.c_str()));
      }
      catch (...)
      {
        index->clear();
        pool()->put(index);
        index = 0;
        throw;
      }
    }

    ~NodeId() 
    {
      index->clear();
      pool()->put(index);
    }

    inline NodeId &operator = (const NodeId &node)
    {
      *index = *node.index;
      
      return *this;
    }

    inline bool operator == (const NodeId &node) const
    {
      return equals(node);
    }

    bool equals(const NodeId &node) const
    {
      // TODO: Check if we need to compare content of the index and node.index vectors
      return index && node.index && *node.index == *index;
    }

    int depth() const
    {
      return (int)index->size();
    }

    inline short& operator[] (unsigned short i) const
    {
      if ( i < index->size() )
        return const_cast<short&>((*index)[i]);
      throw std::range_error("invalid index");
    }

    inline short back() const
    {
      if ( index->size() > 0 )
        return (*index)[index->size() - 1];
      throw std::logic_error("invalid node id");
    }
    
    //! Advance leaf to the next node. E.g. for node with path "1.3.2" it will become "1.3.3"
    inline bool next() const
    {
      bool ret = false;
      if ( index->size() > 0 )
      {
        ++((*index)[index->size() - 1]);
        ret = true;
      }
      return ret;
    }
    
    inline bool is_valid() const
    {
      return index->size() != 0;
    }
    
    inline NodeId parent() const
    {
      if (depth() < 2)
        return NodeId();
      NodeId copy(*this);
      copy.index->pop_back();
      return copy;
    }

    inline std::string repr(const char separator = '.') const
    {
      std::string r;
      const int depth = (int)index->size();
      for (int i= 0; i < depth; i++)
      {
        char buf[30];
        g_snprintf(buf, sizeof(buf), "%i", (*index)[i]);
        if (i > 0)
          r= r + separator + buf;
        else
          r= buf;
      }
      return r;
    }

    inline NodeId &append(const short i)
    {
      if (i < 0)
        throw std::invalid_argument("negative node index is invalid");
      
      index->push_back(i);
      
      return *this;
    }
  };

  //----------------------------------------------------------------------------
  //----------------------------------------------------------------------------
  /**
    \class NodeIds
    \brief Mapper of short-living NodeId to a persistent item
    
    This class was added cause GtkTreeIter which is used in TreeModel for Gtk::TreeView
    has only three int-size fields describing iterator state. And it is not possible to fit entire NodeId
    there, moreover GtkTreeIter along with Gtk::TreeIter have no indication when iterator is 
    destroyed. So the was a need to have something small which can be fit into those fields.
    Lifetime of the items named 'uid' is defined by the initial mapping when the uid is created
    and the issue of NodeIds::flush call. The mapped entity 'uid' is a pointer to a std::string stored
    in the std::set. For example if we need to have an item which lifetime is longer that NodeId("1.2.3")
    we should obtain mapped uid of the NodeId("1.2.3"). That mapped item - uid will be a pointer
    to a string in the std::set and the string itself will store "1.2.3". Having that uid allows us to get back
    path which can be stored between NodeId creations. That is used for example when we obtain Gtk::TreeIter
    and need to move to the next node in the Tree thus advance iterator.
    It is not advisable to dereference 'uid' to obtain std::string from std::string*, as future implementation
    may change that.
  */
  class NodeIds
  {
    public:
      NodeIds() {}
      
      //! Resets map of NodeId paths to uid
      void flush();
        
      //! Maps path with type of std::string from NodeId. This function is used for
      //! convience. See map_node_id(const NodeId&)
      NodeId::uid map_node_id(const std::string& path_from_nodeid);
      NodeId::uid map_node_id(const NodeId& nid)
      {
        return map_node_id(nid.repr());
      }
      
      //! Reverse mapping from 'uid' to a path
      const std::string& map_node_id(const NodeId::uid nodeid);
  
    private:
      typedef std::set<std::string> Map;
      
      Map          _map;
  };
  
  //------------------------------------------------------------------------------
  inline void NodeIds::flush()
  {
    _map.clear();
  }
  
  //------------------------------------------------------------------------------
  inline NodeId::uid NodeIds::map_node_id(const std::string& path_from_nodeid)
  {
    const Map::const_iterator it = _map.find(path_from_nodeid);
    if ( _map.end() != it )
      return (const NodeId::uid)&(*it);
    else
    {
      //TODO: make a faster way. Probably we can use item from insert
      _map.insert(path_from_nodeid);
      return map_node_id(path_from_nodeid);
    }
  }
  
  //------------------------------------------------------------------------------
  inline const std::string& NodeIds::map_node_id(const NodeId::uid nodeid)
  {
    // Note that dereference of nodeid is dangerous after flush was called
    // That should be protected by stamp approach in TreeModel wrapper.
    static std::string empty;
    return nodeid ? *nodeid : empty;
  }

  /** Base list model class.
   * 
   * @ingroup begrt
   */
  class MYSQLGRTBACKEND_PUBLIC_FUNC ListModel
  {
   private:
    NodeIds               _nodeid_map;
    sigc::signal<void>    _tree_changed_signal;
    
   public:
    virtual ~ListModel() {};

    virtual int count()= 0;
    virtual NodeId get_node(int index);
    virtual bool has_next(const NodeId &node);
    virtual NodeId get_next(const NodeId &node);

    sigc::signal<void>& tree_changed_signal()
    {
      return _tree_changed_signal;
    }

    void tree_changed()
    {
      _tree_changed_signal.emit();
      _nodeid_map.flush();
    }

    NodeId::uid nodeid_to_uid(const NodeId& nodeid)
    {
      return _nodeid_map.map_node_id(nodeid);
    }

    NodeId::uid nodeid_path_to_uid(const std::string& path)
    {
      return _nodeid_map.map_node_id(path);
    }

    const std::string& nodeuid_to_path(const NodeId::uid nodeuid)
    {
      return _nodeid_map.map_node_id(nodeuid);
    }    
//    virtual void set_data(NodeId node, void *data)= 0;
//    virtual void *get_data(NodeId node)= 0;

    virtual bool get_field(const NodeId &node, int column, std::string &value);
    virtual bool get_field(const NodeId &node, int column, int &value);
    virtual bool get_field(const NodeId &node, int column, bool &value);
    virtual bool get_field(const NodeId &node, int column, double &value);
    
    // representation of the field as a GRT value
    virtual grt::ValueRef get_grt_value(const NodeId &node, int column);

    virtual std::string get_field_description(const NodeId &node, int column);
    virtual IconId get_field_icon(const NodeId &node, int column, IconSize size);
    
    virtual void refresh()= 0;

    virtual void reorder(const NodeId &node, int index) { throw std::logic_error("not implemented"); }
    void reorder_up(const NodeId &node);
    void reorder_down(const NodeId &node);

    virtual bool activate_node(const NodeId &node) { throw std::logic_error("not implemented"); return false; }
    virtual int get_node_popup_items(const NodeId &node, MenuItemList &items) { throw std::logic_error("not implemented"); }
    virtual void activate_node_popup_item(const NodeId &node, const std::string &name) { throw std::logic_error("not implemented"); }

    virtual bool delete_node(const NodeId &node) { throw std::logic_error("not implemented"); }

    // for editable lists only
    virtual grt::Type get_field_type(const NodeId &node, int column);
 

    virtual bool set_field(const NodeId &node, int column, const std::string &value);
    virtual bool set_field(const NodeId &node, int column, int value);
    virtual bool set_field(const NodeId &node, int column, double value);

    virtual bool set_convert_field(const NodeId &node, int column, const std::string &value);

  protected:
    // for internal use only
    virtual bool get_field_grt(const NodeId &node, int column, grt::ValueRef &value);

    grt::ValueRef parse_value(grt::Type type, const std::string &value);
  };



  /** Base tree model class.
   * 
   * @ingroup begrt
   */
  class MYSQLGRTBACKEND_PUBLIC_FUNC TreeModel : public ListModel
  {
  public:
    virtual int count();
    virtual NodeId get_node(int index);

    virtual NodeId get_root() const;
    virtual int get_node_depth(const NodeId &node);
    inline NodeId get_parent(const NodeId &node) const { return node.parent(); }

    virtual int count_children(const NodeId &parent)= 0;
    virtual NodeId get_child(const NodeId &parent, int index)= 0;
    virtual bool has_next(const NodeId &node);
    virtual NodeId get_next(const NodeId &node);

    virtual bool is_expandable(const NodeId &node_id);
    virtual bool expand_node(const NodeId &node);
    virtual void collapse_node(const NodeId &node);
    virtual bool is_expanded(const NodeId &node);

    void save_expand_info(const std::string &path);


    virtual void dump(int show_field);
  protected:
    
  };



  class MYSQLGRTBACKEND_PUBLIC_FUNC GridModel : public ListModel
  {
  public:
    enum ColumnType
    {
      UnknownType,
      StringType,
      NumericType,
      DatetimeType,
      BlobType
    };

    virtual int get_column_count()= 0;
    virtual std::string get_column_caption(int column)= 0;
    virtual ColumnType get_column_type(int column)= 0;
  };

};

#ifdef _MSC_VER
#pragma make_public(::bec::TreeModel)
#pragma make_public(::bec::NodeId)
#pragma make_public(::bec::GridModel)
#endif


#endif /* _TREE_MODEL_H_ */
