#ifndef _EDITOR_TABLE_H_
#define _EDITOR_TABLE_H_

#include "grtdb/editor_dbobject.h"
#include "grtdb/editor_table_inserts.h"
#include "grt/tree_model.h"

#include "grtdb/charset_list.h"

#include "grts/structs.db.mgmt.h"

#include "grtdb_backend_public_interface.h"

#define TableEditorBE_VERSION 2


namespace bec {

  class TableEditorBE;
  class IndexListBE;
  class FKConstraintListBE;
  
  class MYSQLGRTDBBACKEND_PUBLIC_FUNC TableColumnsListBE : public ListModel
  {
  public:
    enum ColumnListColumns {
      Name,
        Type,
        IsPK,
        IsNotNull,
        Flags,
        Default,
        CharsetCollation,
        Charset,
        Collation,
        Comment,
        HasCharset,
        LastColumn
    };
    
    TableColumnsListBE(TableEditorBE *owner);
    
    bool get_row(const NodeId &node,
                 std::string &name,
                 std::string &type,
                 bool &ispk,
                 bool &notnull,
                 std::string &flags,
                 std::string &defvalue,
                 std::string &charset,
                 std::string &collation,
                 std::string &comment
                 );

    virtual IconId get_field_icon(const NodeId &node, int column, IconSize size);

    virtual void refresh();
    virtual int count();
    int real_count();

    bool set_column_type(const NodeId &node, const GrtObjectRef &type);

    virtual bool set_field(const NodeId &node, int column, const std::string &value);
    virtual bool set_field(const NodeId &node, int column, int value);

    virtual void reorder(const NodeId &node, int nindex);

    std::vector<std::string> get_datatype_flags(const ::bec::NodeId &node);
    void set_column_flag(const ::bec::NodeId &node, const std::string& flag_name, int is_set);
    int get_column_flag(const ::bec::NodeId &node, const std::string& flag_name);

    virtual std::string quote_value_if_needed(const db_ColumnRef &column, const std::string &value);
    virtual int get_node_popup_items(const NodeId &node, MenuItemList &items);

  protected:
    TableEditorBE *_owner;
    
    
    
    // for internal use only
    virtual bool get_field_grt(const NodeId &node, int column, grt::ValueRef &value);

    virtual void reorder_column(int column, int new_index);
  };
  
  
  
  class MYSQLGRTDBBACKEND_PUBLIC_FUNC IndexColumnsListBE : public ListModel
  {
  public:
    enum IndexColumnsListColumns {
        Name,
        Descending,
        Length,
        OrderIndex
    };
    
    IndexColumnsListBE(IndexListBE *owner);
    
    virtual void refresh();
    virtual int count();

    void set_column_enabled(const NodeId &node, bool flag);
    bool get_column_enabled(const NodeId &node);

    virtual bool set_field(const NodeId &node, int column, int value);
    virtual bool set_field(const NodeId &node, int column, const std::string &value);

    int get_max_order_index();

  protected:
    IndexListBE *_owner;
    
    // for internal use only
    virtual bool get_field_grt(const NodeId &node, int column, grt::ValueRef &value);
    
    db_IndexColumnRef get_index_column(const db_ColumnRef &column);
    int get_index_column_index(const db_ColumnRef &column);
    void set_index_column_order(const db_IndexColumnRef &column, int order);
  };
  
  
  
  class MYSQLGRTDBBACKEND_PUBLIC_FUNC IndexListBE : public ListModel
  {
    friend class IndexColumnsListBE;
    friend class TableEditorBE;
    
  public:
    enum IndexListColumns {
      Name,
        Type,
        Comment,
        LastColumn
    };
    
    IndexListBE(TableEditorBE *owner);

    virtual void refresh();
    virtual int count();
    int real_count();
    
    // for editable lists only
    virtual bool set_field(const NodeId &node, int column, const std::string &value);
    
    IndexColumnsListBE *get_columns() { return &_column_list; }
    
    
    db_IndexRef get_selected_index();
    void select_index(const NodeId &node);

    bool index_editable(const db_IndexRef &index);

    TableEditorBE *get_owner() { return _owner; }
    
  protected:
    // for internal use only
    virtual bool get_field_grt(const NodeId &node, int column, grt::ValueRef &value);

    NodeId add_column(const db_ColumnRef &column, const db_IndexRef &index= db_IndexRef());
    void remove_column(const NodeId &node);

  protected:
    IndexColumnsListBE _column_list;
    TableEditorBE *_owner;
    NodeId _selected;
  };
  
  
  
  class MYSQLGRTDBBACKEND_PUBLIC_FUNC FKConstraintColumnsListBE : public ListModel
  {
  public:
    enum FKConstraintColumnsListColumns {
      Enabled,
      Column,
      RefColumn
    };
    
    FKConstraintColumnsListBE(FKConstraintListBE *owner);
    
    virtual void refresh();
    virtual int count();

    std::vector<std::string> get_ref_columns_list(const NodeId &node);

    // for editable lists only
    virtual bool set_field(const NodeId &node, int column, const std::string &value);
    virtual bool set_field(const NodeId &node, int column, int value);

    void set_column_is_fk(const NodeId &node, bool flag);
    int get_fk_column_index(const NodeId &node);
    bool get_column_is_fk(const NodeId &node);
    
    FKConstraintListBE *get_owner() { return _owner; }
    
  protected:
    // for internal use only
    virtual bool get_field_grt(const NodeId &node, int column, grt::ValueRef &value);
    
    FKConstraintListBE *_owner;
  };
  
  
  
  class MYSQLGRTDBBACKEND_PUBLIC_FUNC FKConstraintListBE : public ListModel
  {
    friend class FKConstraintColumnsListBE;
    
  public:
    enum FKConstraintListColumns {
      Name,
        OnDelete,
        OnUpdate,
        RefTable,
        Comment
    };
    FKConstraintListBE(TableEditorBE *owner);
    
    NodeId add_column(const db_ColumnRef &column, const db_ColumnRef &refcolumn, 
      const db_ForeignKeyRef &fk= db_ForeignKeyRef());

    virtual void remove_column(const NodeId &node);

    virtual void refresh();
    virtual int count();
    int real_count();

    // for editable lists only
    virtual bool set_field(const NodeId &node, int column, const std::string &value);
    
    void select_fk(const NodeId &node);
    db_ForeignKeyRef get_selected_fk();
    
    TableEditorBE *get_owner() { return _owner; }
    
    FKConstraintColumnsListBE *get_columns() { return &_column_list; }

  protected:
    // for internal use only
    virtual bool get_field_grt(const NodeId &node, int column, grt::ValueRef &value);
    
  protected:
    FKConstraintColumnsListBE _column_list;
    TableEditorBE *_owner;
    NodeId _selected_fk;
  };
  
  
  class MYSQLGRTDBBACKEND_PUBLIC_FUNC TableEditorBE : public DBObjectEditorBE
  {
  public:
    virtual db_DatabaseObjectRef get_dbobject() { return get_table(); }
    virtual db_TableRef get_table()= 0;

    virtual TableColumnsListBE *get_columns()= 0;
    virtual IndexListBE *get_indexes()= 0;
    FKConstraintListBE *get_fks() { return &_fk_list; }
    TableInsertsGridBE *get_inserts_grid() { return &_inserts_grid; }

    virtual void set_name(const std::string &name);

    // table options
    virtual void set_table_option_by_name(const std::string& name, const std::string& value) = 0;
    virtual std::string get_table_option_by_name(const std::string& name) = 0;

    // column editing
    virtual NodeId add_column(const std::string &name);
    virtual void remove_column(const NodeId &column);
    void rename_column(const db_ColumnRef &column, const std::string &name);

    db_ColumnRef get_column_with_name(const std::string &name);

    // fk editing
    virtual NodeId add_fk(const std::string &name);
    virtual bool remove_fk(const NodeId &fk);
    virtual NodeId add_fk_with_columns(const std::vector<NodeId> &columns);  

    // index editing
    virtual NodeId add_index(const std::string &name);
    virtual bool remove_index(const NodeId &index);

    virtual NodeId add_index_with_columns(const std::vector<NodeId> &columns);

    // helper utils for columns
    virtual bool parse_column_type(const std::string &str, db_ColumnRef &column);
    virtual std::string format_column_type(db_ColumnRef &column);

    virtual std::vector<std::string> get_index_types()= 0;

    virtual void set_triggers_sql(const std::string &sql);
    grt::ValueRef parse_triggers_sql(grt::GRT*, grt::StringRef sql);

    bool update_inserts_grid();
    void update_from_inserts_grid();
    bool set_inserts(const std::string &text);
    std::string get_inserts();

  protected:
    db_TableRef _table;
    FKConstraintListBE _fk_list;
    TableInsertsGridBE _inserts_grid;

    TableEditorBE(GRTManager *grtm,
                  const db_TableRef &table,
                  const db_mgmt_RdbmsRef &rdbms);
    
    void undo_called(grt::UndoAction *action, grt::UndoAction *expected);
  };

};

#endif /* _EDITOR_TABLE_H_ */
