
#ifndef _WIZARDFORM_H_
#define _WIZARDFORM_H_

#include <mforms/mforms.h>
#include <vector>

#include <grtpp.h>


#ifdef _WIN32
# ifdef DECL_GRTUI_EXPORT
#  define GRTUI_EXPORT __declspec(dllexport)
# else
#  define GRTUI_EXPORT __declspec(dllimport)
# endif
#else
# define GRTUI_EXPORT
#endif



namespace bec {
  class GRTManager;
};

namespace grtui {

  class WizardPage;

  class GRTUI_EXPORT WizardForm : public ::mforms::Form
  {
  public:
    WizardForm(bec::GRTManager *mgr);
    virtual ~WizardForm();

    virtual grt::DictRef run_modal(grt::DictRef &values);

    void set_top_icon(const std::string &path);

    void add_page(WizardPage *page);

    int default_width() const;

    void update_buttons();

    void set_problem(const std::string &text);
    void clear_problem();

    virtual void reset();
    
    void switch_to_page(WizardPage *page, bool advancing);
    WizardPage *get_active_page() { return _active_page; }
    int get_active_page_number();
    
    WizardPage *get_page_with_id(const std::string &id);

    bec::GRTManager *grtm() { return _grtm; }
    
    grt::DictRef values() { return _values; }
    
  protected:    
    bec::GRTManager *_grtm;
    
    ::mforms::Table _toplevel;

    ::mforms::Panel _header_back;
    ::mforms::Box _hheader;
    ::mforms::ImageBox _sakila;
    ::mforms::Box _header;
    ::mforms::Label _top_title;
    ::mforms::Label _top_subtitle;
    ::mforms::Panel _contents;
    ::mforms::Panel _side;
    ::mforms::Table _side_table;
    
    ::mforms::Box _button_box;
    ::mforms::Button _next;
    ::mforms::Button _back;
    ::mforms::Button _cancel;
    ::mforms::Button _advanced;

    grt::DictRef _values;
     
    std::string _problem;

    WizardPage *_active_page;
    std::vector<WizardPage*> _pages;
    std::list<WizardPage*> _turned_pages;

    struct StepItem;
    std::vector<StepItem*> _step_items;

    virtual WizardPage *get_next_page(WizardPage *current);
    void rebuild_step_list();
    void refresh_step_list();

  public:
    void go_to_next();
    void go_to_back();
    void finish();
    void cancel();

    void toggle_advanced();
  };


  /** A page of a wizard.
   * 
   * The sequence of calls when navigating through the wizard is:
   * - p1.load
   * - p1.enter
   * - <next>
   *   - p1.execute
   *   - p1.leave
   *   - p2.load
   *   - p2.enter
   * - <back>
   *   - p2.leave
   *   - p1.enter
   */
  class GRTUI_EXPORT WizardPage : public ::mforms::Box
  {
  public:
    WizardPage(WizardForm *form, const std::string &pageid);

    std::string get_id() { return _id; }
    
    std::string get_title() { return _title; }
    std::string get_subtitle() { return _subtitle; }
    
    void set_title(const std::string &title);
    void set_subtitle(const std::string &title);

    virtual mforms::View *get_advanced_panel() { return 0; }

    void show_advanced(bool flag);
    bool advanced_shown() const;

    virtual void load() {}
    virtual void execute() {}

    virtual bool allow_next() { return true; }
    virtual bool allow_back() { return true; }
    virtual bool allow_cancel() { return true; }

    //! return true if this is the last page and pressing next should close wizard
    virtual bool next_closes_wizard() { return false; }

    //! return true if execute_overriden_next_action() should be called when next is pressed
    virtual bool override_next_action();
    //! return true if next was overriden
    virtual bool execute_overriden_next_action();
    //! overrider may return "" for default caption
    virtual std::string next_button_caption() { return "_Next >"; }

    void validate();

    sigc::signal<void, grt::DictRef> signal_enter() { return _signal_enter; }
    sigc::signal<void, grt::DictRef> signal_leave() { return _signal_leave; }
    
  public:


  protected:
    friend class WizardForm;

    void enter(grt::DictRef values);
    void leave(grt::DictRef values);

    void enable_file_browsing(mforms::TextEntry *entry, mforms::Button *button, mforms::FileChooserType type,
                              const std::string &extensions= "");

    grt::DictRef values() { return _form->values(); }

    //! Subclasses must override this to implement validation.
    //! If there is a validation error, it must call _form->set_problem()
    virtual void do_validate() { }

  protected:
    WizardForm *_form;
    std::string _id;
    sigc::signal<void, grt::DictRef> _signal_enter;
    sigc::signal<void, grt::DictRef> _signal_leave;
    std::string _title;
    std::string _subtitle;

    bool _advanced_shown;

  private:
    void browse_file_callback(mforms::TextEntry *entry, mforms::FileChooserType type,
                              const std::string &extensions);
  };

};


#endif
