
#include "wizard_progress_page.h"
#include "grt/common.h"
#include "grt/grt_manager.h"

using namespace grtui;


void WizardProgressPage::TaskRow::set_state(WizardProgressPage::TaskState state)
{
  std::string file;
  switch (state)
  {
  case WizardProgressPage::StateNormal:
    file= "task_unchecked.png";
    break;
  case WizardProgressPage::StateBusy:
    file= "task_unchecked.png";
    break;
  case WizardProgressPage::StateDone:
    file= "task_checked.png";
    break;
  case WizardProgressPage::StateError:
    file= "task_error.png";
    break;
  case WizardProgressPage::StateDisabled:
    file= "task_disabled.png";
    break;
  }

  std::string path= bec::IconManager::get_instance()->get_icon_path(file);

  icon.set_image(path);
}


void WizardProgressPage::TaskRow::set_enabled(bool flag)
{
  set_state(flag ? StateNormal : StateDisabled);
  label.set_enabled(flag);

  enabled= flag;
}



WizardProgressPage::WizardProgressPage(WizardForm *form, const std::string &id)
  : WizardPage(form, id), _progress(true), _content(false),
_log_panel(true)
{
  _preview_before_execute= false;
  
  _progress_box= 0;
  _progress_bar= 0;
  _progress_label= 0;
  
  _busy= false;
  _current_task= 0;

  _progress.set_title(_("Progress"));
  _progress.add(&_content);

  _label1.set_text(_("Tasks to Execute"));
  _label1.set_style(mforms::BoldStyle);
  _label2.set_text(_("The following tasks will now be executed. Please monitor the execution.\n"
                     "Press [Advanced] to see the logs."));

  _content.set_padding(20);
  _content.set_spacing(8);

  _content.add(&_label1, false, false);
  _content.add(&_label2, false, false);

  _log_text.set_padding(12);
  _log_text.set_read_only(true);

  _log_panel.set_title(_("Message Log"));
  _log_panel.add(&_log_text);

  add(&_progress, false, true);
  add(&_log_panel, true, true);
  _log_panel.show(false);

  set_spacing(12);
}


WizardProgressPage::~WizardProgressPage()
{
  for (std::vector<TaskRow*>::iterator iter= _tasks.begin();
       iter != _tasks.end(); ++iter)
    delete *iter;
  
  delete _progress_box;
  delete _progress_bar;
  delete _progress_label;
}


WizardProgressPage::TaskRow *WizardProgressPage::add_async_task(const std::string &caption,
                                                                const sigc::slot<bool> &execute,
                                                                const std::string &status_text)
{
  return add_task(true, caption, execute, status_text);
}



WizardProgressPage::TaskRow *WizardProgressPage::add_task(const std::string &caption,
                                                          const sigc::slot<bool> &execute,
                                                          const std::string &status_text)
{
  return add_task(false, caption, execute, status_text);
}


WizardProgressPage::TaskRow *WizardProgressPage::add_task(bool async,
                                                          const std::string &caption, 
                                                          const sigc::slot<bool> &execute,
                                                          const std::string &status_text)
{
  TaskRow *row= new TaskRow;

  row->icon.set_image(bec::IconManager::get_instance()->get_icon_path("task_unchecked.png"));
  row->label.set_text(caption);

  row->box.set_spacing(4);
  row->box.add(&row->icon, false, false);
  row->box.add(&row->label, true, true);

  row->execute= execute;
  row->status_text= status_text;

  row->async= async;
  row->async_running= false;
  row->async_failed= false;

  _tasks.push_back(row);
  
  _content.add(&row->box, false, false);
  
  return row;
}


void WizardProgressPage::end_adding_tasks(bool add_progressbar,
                                          const std::string &finish_message,
                                          bool preview_tasks_before_execute)
{
  _status_text.set_style(mforms::BoldStyle);
  _status_text.set_text("Preparing");
  _content.add(&_status_text, false, false);

  if (add_progressbar)
  {
    _progress_box= new mforms::Box(true);
    _progress_bar= new mforms::ProgressBar();
    _progress_label= new mforms::Label();

    _progress_label->set_text("");
  
    _progress_box->set_padding(12);
    _progress_box->set_spacing(12);
    _progress_box->add(_progress_bar, false, false);
    _progress_bar->set_size(400, -1);
    _progress_box->add(_progress_label, true, true);

    _content.add(_progress_box, false, false);
  }
  _finish_message= finish_message;
  
  _preview_before_execute= preview_tasks_before_execute;
  
  _status_text.set_text("");
}


void WizardProgressPage::reset_tasks()
{
  for (std::vector<TaskRow*>::iterator task= _tasks.begin();
       task != _tasks.end(); ++task)
  {
    (*task)->async_running= false;
    (*task)->async_failed= false;
    (*task)->set_state(WizardProgressPage::StateNormal);
  }
}


void WizardProgressPage::start_tasks()
{
  _got_warning_messages= false;
  _got_error_messages= false;
  _current_task= 0;
  _busy= true;

  _form->update_buttons();
  
  perform_tasks();
}


void WizardProgressPage::perform_tasks()
{
  bool failed= false;

  if (!_form->grtm()->in_main_thread())
    throw std::logic_error("BAD THREAD");
  
  while (_current_task < (int)_tasks.size())
  {
    TaskRow *task= _tasks[_current_task];

    _form->flush_events();
    _form->grtm()->perform_idle_tasks();

    // check if we're being called because an async task finished
    if (task->async_running)
    {
      task->async_running= false;
      if (task->async_failed)
      {
        failed= true;

        break;
      }
      else
      {
        task->set_state(StateDone);
        //task.ProcessTaskMsg((int)Msg_type.MT_info, ""); // delimit task log messages with empty line
        _current_task++;
        continue;
      }
    }

    set_status_text(task->status_text);

    if (task->enabled)
    {
      try
      {
        task->set_state(StateBusy);

        _form->flush_events();

        bool flag= task->execute();

        if (task->async && flag)
        {
          task->async_running= true;
          // if the task is asynchronous, return here to allow other stuff to be
          // executed. When the task is finished, process_task_finish() (or _task_fail())
          // will get called which will in turn, call this method again
          return;
        }

        task->set_state(StateDone);

        _current_task++;
      }
      catch (std::exception &exc)
      {
        failed= true;

        set_status_text(std::string("Error: ").append(exc.what()), true);
        break;
      }
    }
    else
      _current_task++;
  }

  if (!failed)
  {
    if (_got_error_messages)
      set_status_text(_("Operation has completed with errors. Please see logs for details."));
    else if (_got_warning_messages)
      set_status_text(_("Operation has completed with warnings. Please see logs for details."));
    else
      set_status_text(_finish_message);
  }
  else
  {
    while (_current_task < (int)_tasks.size())
    {
      TaskRow *task= _tasks[_current_task++];
      task->set_state(StateError);
    }
  }

  _done= true;
  _busy= false;
  _signal_finished(!failed);

  validate();
}


void WizardProgressPage::set_status_text(const std::string &text, bool is_error)
{
  if (!_form->grtm()->in_main_thread())
    throw std::logic_error("BAD THREAD");

  _status_text.set_text(text);
}


void WizardProgressPage::update_progress(float pct, const std::string &caption)
{
  if (!_form->grtm()->in_main_thread())
    throw std::logic_error("BAD THREAD");

  if (_progress_bar)
    _progress_bar->set_value(pct);
  if (_progress_label)
    _progress_label->set_text(caption);
}


bool WizardProgressPage::allow_cancel()
{
  return !_busy;
}


bool WizardProgressPage::allow_next()
{
  return !_busy;
}


bool WizardProgressPage::override_next_action()
{
  return _preview_before_execute;
}


bool WizardProgressPage::execute_overriden_next_action()
{
  if (!_done)
  {
    start_tasks();
    return true;
  }
  return false;
}


std::string WizardProgressPage::next_button_caption()
{
  if (_done || !_preview_before_execute)
    return "";
  return _("_Execute >");
}


void WizardProgressPage::execute()
{
  _done= false;

  if (_preview_before_execute)
    set_status_text(_("Press [Execute >] to perform described actions."));
  else
    start_tasks();
}


void WizardProgressPage::add_log_text(const std::string &text)
{
  _log_text.append_text(text+"\n");
}





//--------------------------------------------------------------------------------


void WizardProgressPage::execute_grt_task(const sigc::slot<grt::ValueRef,grt::GRT*> &slot, bool sync)
{
  bec::GRTTask *task= new bec::GRTTask("wizard task", _form->grtm()->get_dispatcher(), slot);

  task->signal_message().connect(sigc::mem_fun(this, &WizardProgressPage::process_grt_task_message));
  task->signal_failed().connect(sigc::mem_fun(this, &WizardProgressPage::process_grt_task_fail));
  task->signal_finished().connect(sigc::mem_fun(this, &WizardProgressPage::process_grt_task_finish));

  if (sync)
    _form->grtm()->get_dispatcher()->add_task_and_wait(task);
  else
    _form->grtm()->get_dispatcher()->add_task(task);
}


void WizardProgressPage::process_grt_task_message(const grt::Message &msg)
{
  std::string msgTypeStr;

  switch (msg.type)
  {
  case grt::ErrorMsg:
    {
      _got_error_messages= true;
      /*
      //if (!continueOnErrMsg)
      */
      _tasks[_current_task]->async_errors++;

      msgTypeStr = "ERROR: ";
    }
    break;
  case grt::WarningMsg:
    {
      _got_warning_messages= true;

      msgTypeStr = "WARNING: ";
    }
    break;
  case grt::OutputMsg:
  case grt::InfoMsg:
    msgTypeStr = "";
    break;

  case grt::ProgressMsg:
    update_progress(msg.progress, msg.text);
    return;
    
  case grt::ControlMsg:
    return;
  }

  add_log_text(msgTypeStr + msg.text);
}


void WizardProgressPage::process_grt_task_fail(const std::exception &error)
{  
  _tasks[_current_task]->async_failed= true;

  if (_tasks[_current_task]->process_fail)
  {
    // if process_fail returns true, the error was recovered
    if (_tasks[_current_task]->process_fail())
      _tasks[_current_task]->async_failed= false;
    else
      set_status_text(std::string("Error: ").append(error.what()), true);
  }
  else
  {
    add_log_text(std::string("Operation failed: ").append(error.what()));
    set_status_text(std::string("Error: ").append(error.what()), true);
  }
  // continue with task execution
  perform_tasks();
}


void WizardProgressPage::process_grt_task_finish(const grt::ValueRef &result)
{
  g_message("FINISHED");

  _form->grtm()->perform_idle_tasks();

  if (_tasks[_current_task]->process_finish)
    _tasks[_current_task]->process_finish(result);

  // continue with task execution
  perform_tasks();
}
