#include "stdafx.h"

#include "grt/common.h"
#include "figure_common.h"
/**
 * @file  figure_common.cpp
 * @brief 
 */



mdc::FontSpec parse_font_spec(const std::string &font)
{
  std::vector<std::string> parts= bec::split_string(font, " ");

  mdc::FontSpec fs;
  int size= 12;

  if (sscanf(parts.back().c_str(), "%i", &size) == 1)
  {
    fs.size= (float) size;
    parts.pop_back();
  }

  for (int i= 0; i < 2 && !parts.empty(); i++)
  {
    if (g_strcasecmp(parts.back().c_str(), "bold")==0)
    {
      fs.toggle_bold(true);
      parts.pop_back();
    }

    if (g_strcasecmp(parts.back().c_str(), "italic")==0)
    {
      fs.toggle_italic(true);
      parts.pop_back();
    }
  }

  if (!parts.empty())
  {
    fs.family= parts[0];
    for (unsigned int i= 1; i < parts.size(); i++)
      fs.family+= " " + parts[i];
  }
  return fs;
}




using namespace wbfig;


Titlebar::Titlebar(mdc::Layer *layer, bool expander)
  : mdc::Box(layer, mdc::Box::Horizontal), _icon_text(layer)
{
  _last_click= 0;

  set_padding(6, 4);
  set_spacing(6);

  _corners= mdc::CNone;
  _back_color= mdc::Color(1, 1, 1);
  _border_color= mdc::Color(0.7, 0.7, 0.7);

  _icon_text.set_allow_shrinking(true);
  //_icon_text.set_auto_sizing(false);

  add(&_icon_text, true, true);
  
  if (expander)
  {
    _expander= new mdc::Button(layer, mdc::ExpanderButton);
    _expander->signal_activate().connect(sigc::mem_fun(this, &Titlebar::expand_toggled));
    _expander->set_pen_color(mdc::Color(0.4, 0.4, 0.4));
    
    add(mdc::autodelete(_expander), false, false);
  }
  else
    _expander= 0;
}



Titlebar::~Titlebar()
{
}


void Titlebar::set_icon(cairo_surface_t *icon)
{
  _icon_text.set_icon(icon);
}


void Titlebar::set_title(const std::string &text)
{
  _icon_text.set_text(text);
}


void Titlebar::set_expanded(bool flag)
{
  if (_expander && _expander->get_active() != flag)
    _expander->set_active(flag);
}


bool Titlebar::get_expanded()
{
  return _expander->get_active();
}


void Titlebar::expand_toggled()
{
  _signal_expand_toggle.emit(_expander->get_active());
}


bool Titlebar::on_click(mdc::CanvasItem *target, const mdc::Point &point, mdc::MouseButton button, mdc::EventState state)
{
  if (button == mdc::ButtonLeft && (target == this || target->get_parent() == this))
  {
    mdc::Timestamp now= mdc::get_time();
    
    if (now - _last_click < DOUBLE_CLICK_DELAY)
    {
      _signal_activate.emit((state & mdc::SControlMask)!=0);
      _last_click= 0;
      return true;
    }
    else
      _last_click= now;
  }
  return mdc::Box::on_click(target, point, button, state);
}


void Titlebar::set_color(const mdc::Color &color)
{
  _back_color= color;
  set_needs_render();
}


void Titlebar::set_text_color(const mdc::Color &color)
{
  _icon_text.set_pen_color(color);
}


void Titlebar::set_border_color(const mdc::Color &color)
{
  _border_color= color;
}


void Titlebar::set_font(const mdc::FontSpec &font)
{
  _icon_text.set_font(font);
}


void Titlebar::set_rounded(mdc::CornerMask corners)
{
  _corners= corners;
}


void Titlebar::render(mdc::CairoCtx *cr)
{
  cr->save();

  mdc::stroke_rounded_rectangle(cr, get_bounds(), _corners, 8);

  cr->set_color(_back_color);
  cr->set_line_width(1);

  cr->fill_preserve();
  cr->set_color(_border_color);
  cr->stroke();
  cr->restore();

  mdc::Box::render(cr);
}


void Titlebar::set_auto_sizing(bool flag)
{
  mdc::Box::set_auto_sizing(flag);
  _icon_text.set_auto_sizing(flag);
}


//-------------------------------------------------------------------------------------------------------------


FigureItem::FigureItem(mdc::Layer *layer)
: mdc::IconTextFigure(layer), _dirty(true) 
{
  set_font(mdc::FontSpec("Helvetica", mdc::SNormal, mdc::WNormal, 11));
}


mdc::Point FigureItem::get_intersection_with_line_to(const mdc::Point &p)
{
  mdc::Point point(mdc::CanvasItem::get_intersection_with_line_to(p));
  mdc::Rect bounds(get_root_bounds());

  if (point.x-bounds.xmin() < bounds.width()/2)
    point.x= bounds.xmin();
  else
    point.x= bounds.xmax();

  point.y= (bounds.ymax()+bounds.ymin())/2;
  
  return point;
}


mdc::Rect FigureItem::get_root_bounds() const
{
  if (!get_visible() || !get_parents_visible())
    return get_toplevel()->get_root_bounds();
  return super::get_root_bounds();
}


void FigureItem::draw_state(mdc::CairoCtx *cr)
{
  switch (get_state())
  {
  case Hovering:
    cr->save();
    cr->set_color(get_view()->get_hover_color(), 0.5);
    stroke_outline(cr, 0);
    cr->fill();
    cr->restore();
    break;
  case Highlighted:
    cr->save();
    cr->set_color(get_view()->get_hover_color(), 0.3);
    stroke_outline(cr, 0);
    cr->fill();
    cr->restore();
    break;
  default:
    mdc::IconTextFigure::draw_state(cr);
  }
}


//-------------------------------------------------------------------------------------------------------------

BaseFigure::BaseFigure(mdc::Layer *layer)
: mdc::Box(layer, mdc::Box::Vertical)
{
  set_cache_toplevel_contents(true);
  set_draggable(true);
  _last_click= 0;
  _manual_resizing= false;
  _resizing= false;
}


void BaseFigure::unset_color()
{
  set_draw_background(false);
}


void BaseFigure::set_color(const mdc::Color &color)
{
  set_background_color(color);
  set_draw_background(true);
}


void BaseFigure::set_allow_manual_resizing(bool flag)
{
  _manual_resizing= flag;
  invalidate_min_sizes();
  if (!flag)
    set_fixed_size(mdc::Size(-1,-1));
}


void BaseFigure::invalidate_min_sizes(mdc::CanvasItem *item)
{
  item->set_needs_relayout();
  if (dynamic_cast<mdc::Layouter*>(item))
    dynamic_cast<mdc::Layouter*>(item)->foreach(sigc::ptr_fun(&BaseFigure::invalidate_min_sizes));
}


void BaseFigure::invalidate_min_sizes()
{
  invalidate_min_sizes(this);
}


bool BaseFigure::on_click(mdc::CanvasItem *target, const mdc::Point &point, mdc::MouseButton button, mdc::EventState state)
{
  if (button == mdc::ButtonLeft && (target == this || target->get_parent() == this))
  {
    mdc::Timestamp now= mdc::get_time();
    
    if (now - _last_click < DOUBLE_CLICK_DELAY)
    {
      _signal_activate((state & mdc::SControlMask) != 0);
      _last_click= 0;
      return true;
    }
    else
      _last_click= now;
  }
  return super::on_click(target, point, button, state);
}


bool BaseFigure::on_drag_handle(mdc::ItemHandle *handle, const mdc::Point &pos, bool dragging)
{
  if (!_manual_resizing)
    set_allow_manual_resizing(true);

  if (dragging)
  {
    if (!_resizing)
      _initial_bounds= get_bounds();
    _resizing= true;
  }

  bool flag= super::on_drag_handle(handle, pos, dragging);

  if (!dragging)
  {
    _resizing= false;
    _signal_resize.emit(_initial_bounds);
  }
  if (flag)
    set_fixed_size(get_size());

  return flag;
}


void BaseFigure::set_content_font(const mdc::FontSpec &font)
{
  _content_font= font;
}


void BaseFigure::set_state_drawing(bool flag)
{
  _disable_state_drawing= !flag;
}


static BaseFigure::ItemList::iterator find_item(BaseFigure::ItemList &list,
                                           const std::string &id)
{
  for (BaseFigure::ItemList::iterator i= list.begin(); i != list.end(); ++i)
  {
    if ((*i)->get_id() == id)
      return i;
  }
  return list.end();
}



BaseFigure::ItemList::iterator BaseFigure::begin_sync(mdc::Box &box, ItemList &list)
{
  get_layer()->get_view()->lock();
  get_layer()->get_view()->lock_redraw();
  return list.begin();
}


BaseFigure::ItemList::iterator BaseFigure::sync_next(mdc::Box &box, ItemList &list,
                                                     ItemList::iterator iter,
                                                     const std::string &id,
                                                     cairo_surface_t *icon,
                                                     const std::string &text,
                                                     const CreateItemSlot &create_item,
                                                     const UpdateItemSlot &update_item)

{
  ItemList::iterator i= find_item(list, id);
  
  if (i == list.end()) // item is new, insert it
  {
    FigureItem *item;

    if (create_item)
      item= create_item(get_layer());
    else
      item= new FigureItem(get_layer());

    if (update_item)
      update_item(item);

    if (_manual_resizing)
      item->set_auto_sizing(false);
    item->set_allow_shrinking(true);
    item->set_spacing(2);
    item->set_padding(4, 4);
    
    item->set_font(_content_font);

    item->set_icon(icon);
    item->set_text(text);
    item->set_id(id);

    list.insert(iter, item); // insert before

    _signal_item_added.emit(item);

    return iter; // return same
  }
  else if (i == iter) // item is in the same position, just update
  {
    if ((*i)->get_icon() != icon || (*i)->get_text() != text)
    {
      (*i)->set_icon(icon);
      (*i)->set_text(text);
      (*i)->set_dirty();
    }

    if (update_item)
      update_item(*i);

    ++iter; // go to next
    return iter;
  }
  else // item has moved, update and move up
  {
    FigureItem *item= *i;

    (*i)->set_icon(icon);
    (*i)->set_text(text);
    (*i)->set_dirty();
    if (update_item)
      update_item(*i);

    list.erase(i);
    list.insert(iter, item);

    return iter;
  }
}


void BaseFigure::end_sync(mdc::Box &box, ItemList &list, ItemList::iterator iter)
{
  // everything after iter is outdated, so just delete everything
  while (iter != list.end())
  {
    ItemList::iterator next= iter;

    ++next;
    
    delete *iter;
    list.erase(iter);

    iter= next;
  }

  bool dirty= false;

  // first check if anything changed
  for (ItemList::const_iterator i= list.begin(); i!=list.end(); ++i)
  {
    if ((*i)->get_dirty())
      dirty= true;
    (*i)->set_dirty(false);
  }

  // resync box
  if (dirty)
  {
    box.remove_all();
    for (ItemList::const_iterator i= list.begin(); i!=list.end(); ++i)
    {
      box.add(*i, false, true, true);
    }
    box.set_needs_relayout();

    if (_manual_resizing)
    {
      mdc::Size size= get_min_size();
      mdc::Size fsize= get_fixed_size();

      if (size.height > fsize.height)
        set_fixed_size(mdc::Size(fsize.width, size.height));
    }
  }
  get_layer()->get_view()->unlock_redraw();
  get_layer()->get_view()->unlock();
}


//---------------------------------------------------------------------------------------------------

TextFigure::TextFigure(mdc::Layer *layer)
: mdc::TextFigure(layer)
{
}


bool TextFigure::on_click(mdc::CanvasItem *target, const mdc::Point &point, mdc::MouseButton button, mdc::EventState state)
{
  return mdc::TextFigure::on_click(target, point, button, state);
}


bool TextFigure::on_enter(mdc::CanvasItem *target, const mdc::Point &point)
{
  _signal_crossing.emit(true);
  return true;
}


bool TextFigure::on_leave(mdc::CanvasItem *target, const mdc::Point &point)
{
  _signal_crossing.emit(false);
  return true;
}
