#include "stdafx.h"

#include "wb_config.h"

#include "table_figure.h"
#include "connection_figure.h"
/**
 * @file  table_figure.cpp
 * @brief 
 */

using namespace wbfig;



//--------------------------------------------------------------------------------

TableColumnItem::TableColumnItem(mdc::Layer *layer)
: FigureItem(layer)
{
  _flags= (ColumnFlags)0;
}

                                                             
void TableColumnItem::set_column_flags(ColumnFlags flags)
{
  if (_flags != flags)
  {
    _flags= flags;
    set_needs_relayout();
  }
}


mdc::Size TableColumnItem::calc_min_size()
{
  mdc::Size size= FigureItem::calc_min_size();
  cairo_text_extents_t extents;
  mdc::CairoCtx *cr= get_layer()->get_view()->cairoctx();

  std::vector<std::string> flags;
  if (_flags & ColumnUnsigned)
    flags.push_back("UN");
  if (_flags & ColumnNotNull)
    flags.push_back("NN");
  if (_flags & ColumnAutoIncrement)
    flags.push_back("AI");

  mdc::FontSpec font(get_font());
  font.size= 8.0;
  for (std::vector<std::string>::const_iterator iter= flags.begin(); iter != flags.end(); ++iter)
  {
    cr->get_text_extents(font, iter->c_str(), extents);
    size.width+= extents.x_advance + 4;
  }

  size.width= ceil(size.width);

  return size;
}

void TableColumnItem::draw_contents(mdc::CairoCtx *cr)
{
  FigureItem::draw_contents(cr);

  cairo_text_extents_t extents;
  mdc::Size text_size(get_text_size());
  mdc::Point pos(get_position());

  if (text_size.width >= get_size().width - 2*_xpadding)
    return;
  
  pos.x+= text_size.width + _xpadding;

  mdc::FontSpec font(get_font());
  font.size= 8.0;

  std::vector<std::string> flags;
  if (_flags & ColumnUnsigned)
    flags.push_back("UN");
  if (_flags & ColumnNotNull)
    flags.push_back("NN");
  if (_flags & ColumnAutoIncrement)
    flags.push_back("AI");

  cr->set_font(font);
  for (std::vector<std::string>::const_iterator iter= flags.begin(); iter != flags.end(); ++iter)
  {
    cr->get_text_extents(font, *iter, extents);

    pos.x+= 4;
    cr->move_to(pos.x, pos.y + (get_size().height + text_size.height)/2);
    pos.x+= extents.x_advance;

    cr->show_text(*iter);
  }
  cr->stroke();
}


//--------------------------------------------------------------------------------


static int find_index(BaseFigure::ItemList::iterator begin,
                      BaseFigure::ItemList::iterator end,
                      FigureItem *item)
{
  int i= 0;
  for (BaseFigure::ItemList::iterator iter= begin; iter != end; ++iter, ++i)
    if (*iter == item)
      return i;
  return -1;
}


bool TableIndexItem::on_enter(mdc::CanvasItem *target, const mdc::Point &point)
{
  int i= find_index(_owner->get_indexes()->begin(),
                    _owner->get_indexes()->end(),
                    this);

  if (i >= 0)
    _owner->signal_index_crossed().emit(i, true);

  return FigureItem::on_enter(target, point);
}


bool TableIndexItem::on_leave(mdc::CanvasItem *target, const mdc::Point &point)
{
  int i= find_index(_owner->get_indexes()->begin(),
                    _owner->get_indexes()->end(),
                    this);

  if (i >= 0)
    _owner->signal_index_crossed().emit(i, false);

  return FigureItem::on_leave(target, point);
}


TableIndexItem::TableIndexItem(mdc::Layer *layer, Table *owner)
  : FigureItem(layer), _owner(owner)
{
}

//-------------------------------------------------------------------------------------------

Table::Table(mdc::Layer *layer, bool collapsible)
: BaseFigure(layer), _background(layer), _title(layer, collapsible)
{
  _original_column_box_height= 0.0;

  _hide_columns= false;
  _hide_indexes= false;
  _hide_triggers= false;

  _show_flags= false;

  add_magnet(_sides_magnet= new mdc::BoxSideMagnet(this));
  _sides_magnet->set_compare_slot(sigc::mem_fun(this, &Table::compare_connection_position));
}


bool Table::on_enter(mdc::CanvasItem *target, const mdc::Point &point)
{
  _signal_enter.emit();
  return super::on_enter(target, point);
}


bool Table::on_leave(mdc::CanvasItem *target, const mdc::Point &point)
{
  _signal_leave.emit();
  return super::on_enter(target, point);
}


void Table::set_title_font(const mdc::FontSpec &font)
{
  _title.set_font(font);
}


void Table::set_section_font(const mdc::FontSpec &font)
{
  if (get_index_title())
    get_index_title()->set_font(font);
  if (get_trigger_title())
    get_trigger_title()->set_font(font);
}


void Table::set_content_font(const mdc::FontSpec &font)
{
  super::set_content_font(font);
}


void Table::set_show_flags(bool flag)
{
  _show_flags= flag;
}


wbfig::FigureItem *Table::create_column_item(mdc::Layer *layer)
{
  return new TableColumnItem(layer);
}


void Table::update_column_item(wbfig::FigureItem *item, ColumnFlags flags)
{
  if (_show_flags)
    dynamic_cast<TableColumnItem*>(item)->set_column_flags(flags);
}


wbfig::FigureItem *Table::create_index_item(mdc::Layer *layer)
{
  return new TableIndexItem(layer, this);
}


bool Table::compare_connection_position(mdc::Connector *a, mdc::Connector *b, 
                                        mdc::BoxSideMagnet::Side side)
{
  wbfig::ConnectionLineLayouter *layouter;
  mdc::Point a_pos, b_pos;
  mdc::CanvasItem *item;
  mdc::Connector *a_conn, *b_conn;

  layouter= dynamic_cast<wbfig::ConnectionLineLayouter*>(
    dynamic_cast<mdc::Line*>(a->get_owner())->get_layouter());
  if (!layouter) return false;
  if (layouter->get_start_connector() == a)
  {
    a_conn= layouter->get_end_connector();
    item= a_conn->get_connected_item();
  }
  else
  {
    a_conn= layouter->get_start_connector();
    item= a_conn->get_connected_item();
  }
  if (item)
    a_pos= item->get_root_bounds().center();
  else
    a_pos= a_conn->get_position();

  layouter= dynamic_cast<wbfig::ConnectionLineLayouter*>(
    dynamic_cast<mdc::Line*>(b->get_owner())->get_layouter());
  if (!layouter) return false;
  if (layouter->get_start_connector() == b)
  {
    b_conn= layouter->get_end_connector();
    item= b_conn->get_connected_item();
  }
  else
  {
    b_conn= layouter->get_start_connector();
    item= b_conn->get_connected_item();
  }
  if (item)
    b_pos= item->get_root_bounds().center();
  else
    b_pos= b_conn->get_position();

  // the strict ordering between 2 connectors is important 
  // ie, compare(a,b) and compare(b,a) must never be the same
  // otherwise we get an infinite loop when a_pos == b_pos
  if (side == mdc::BoxSideMagnet::Top || side == mdc::BoxSideMagnet::Bottom)
    return a_pos.x < b_pos.x || a_pos.x == b_pos.x && a_conn < b_conn;
  else
    return a_pos.y < b_pos.y || a_pos.y == b_pos.y && a_conn < b_conn;
}
