#ifndef _LF_TREEVIEW_H_
#define _LF_TREEVIEW_H_

#include <mforms/mforms.h>

#include "lf_view.h"

namespace mforms {
namespace gtk {


class TreeViewImpl : public ViewImpl
{
private:
  class ColumnRecord : public Gtk::TreeModelColumnRecord
  {
  public:
    std::vector<Gtk::TreeModelColumnBase*> columns;
    
    template<class C>
      const Gtk::TreeModelColumn<C> &get(int column)
      {
        Gtk::TreeModelColumnBase* c= columns[column];
        
        return *static_cast<Gtk::TreeModelColumn<C>*>(c);
      }


    virtual ~ColumnRecord()
    {
      for (std::vector<Gtk::TreeModelColumnBase*>::iterator iter= columns.begin();
           iter != columns.end(); ++iter)
        delete *iter;
    }

    int add_string(Gtk::TreeView *tree, const std::string &title, bool editable)
    {
      Gtk::TreeModelColumn<Glib::ustring> *column= new Gtk::TreeModelColumn<Glib::ustring>();
      
      columns.push_back(column);
      add(*column);
            
      if (editable)
        tree->append_column_editable(title, *column);
      else
        tree->append_column(title, *column);
      
      return columns.size()-1;
    }


    int add_integer(Gtk::TreeView *tree, const std::string &title, bool editable)
    {
      Gtk::TreeModelColumn<int> *column= new Gtk::TreeModelColumn<int>();
      
      columns.push_back(column);
      add(*column);

      if (editable)
        tree->append_column_editable(title, *column);
      else
        tree->append_column(title, *column);

      return columns.size()-1;
    }


    int add_check(Gtk::TreeView *tree, const std::string &title, bool editable)
    {
      Gtk::TreeModelColumn<bool> *column= new Gtk::TreeModelColumn<bool>();
      
      columns.push_back(column);
      add(*column);

      if (editable)
        tree->append_column_editable(title, *column);
      else
        tree->append_column(title, *column);

      return columns.size()-1;
    }
  };

  ColumnRecord _columns;
  

  Gtk::ScrolledWindow *_swin;
  Gtk::TreeView *_tree;

  Glib::RefPtr<Gtk::ListStore> _list_store;

  virtual Gtk::Widget *get_outer() const { return _swin; }
  virtual Gtk::Widget *get_inner() const { return _tree; }


  TreeViewImpl(TreeView *self, bool show_header)
    : ViewImpl(self)
  {
    _swin= Gtk::manage(new Gtk::ScrolledWindow());
    _swin->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
    _swin->set_shadow_type(Gtk::SHADOW_IN);
    _tree= Gtk::manage(new Gtk::TreeView());
    _tree->show();

    _swin->add(*_tree);

    _tree->set_headers_visible(show_header);
  }
  
  
  void string_edited(const Glib::ustring &path, const Glib::ustring &new_text, int column)
  {
    if (_list_store)
    {
      Gtk::TreeRow row= *_list_store->get_iter(Gtk::TreePath(path));

      row[_columns.get<Glib::ustring>(column)]= new_text;
    }
  }


  int add_column(TreeColumnType type, const std::string &name, bool editable)
  {
    int column;
    switch (type)
    {
    case StringColumnType:
      column= _columns.add_string(_tree, name, editable);
      if (editable)
      {
        ((Gtk::CellRendererText*)_tree->get_column(column)->get_first_cell_renderer())->signal_edited().
          connect(sigc::bind(sigc::mem_fun(this, &TreeViewImpl::string_edited), column));
      }
      break;
    case IntegerColumnType: column= _columns.add_integer(_tree, name, editable); break;
    case CheckColumnType: column= _columns.add_check(_tree, name, editable); break;
    default: return -1;
    }
    
    _tree->get_column(column)->set_resizable(true);
    
    return column;
  }


  void end_columns()
  {
    _list_store= Gtk::ListStore::create(_columns);
    _tree->set_model(_list_store);
  }



  static bool create(TreeView *self, bool show_header)
  {
    return new TreeViewImpl(self, show_header) != 0;
  }

  static int add_column(TreeView *self, TreeColumnType type, const std::string &name, bool editable)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();

    return tree->add_column(type, name, editable);
  }
  
  static void end_columns(TreeView *self)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();

    tree->end_columns();
  }

  static void clear_rows(TreeView *self)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();

    if (tree->_list_store)
      tree->_list_store->clear();
  }
  
  static void delete_row(TreeView *self, int row)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreePath path;
    
    path.append_index(row);

    if (tree->_list_store)
      tree->_list_store->erase(tree->_list_store->get_iter(path));
  }
  
  static int add_row(TreeView *self)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();

    if (tree->_list_store)
      return tree->_list_store->get_path(tree->_list_store->append()).front();
    
    return -1;
  }
  
  static int count(TreeView *self)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();

    if (tree->_list_store)
      return tree->_list_store->children().size();
    
    return 0;
  }

  static void set_string(TreeView *self, int row, int column, const std::string &value)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreeRow tree_row;
    Gtk::TreePath path;

    path.append_index(row);

    if (tree->_list_store)
      tree_row= *tree->_list_store->get_iter(path);
    else
      return;

    tree_row[tree->_columns.get<Glib::ustring>(column)]= value;
  }
  
  static void set_integer(TreeView *self, int row, int column, int value)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreeRow tree_row;
    Gtk::TreePath path;

    path.append_index(row);

    if (tree->_list_store)
      tree_row= *tree->_list_store->get_iter(path);
    else
      return;

    tree_row[tree->_columns.get<int>(column)]= value;
  }
  
  static void set_check(TreeView *self, int row, int column, bool check)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreeRow tree_row;
    Gtk::TreePath path;

    path.append_index(row);

    if (tree->_list_store)
      tree_row= *tree->_list_store->get_iter(path);
    else
      return;

    tree_row[tree->_columns.get<bool>(column)]= check;
  }
  
  
  static std::string get_string(TreeView *self, int row, int column)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreeRow tree_row;
    Gtk::TreePath path;

    path.append_index(row);

    if (tree->_list_store)
      tree_row= *tree->_list_store->get_iter(path);
    else
      return "";

    return ((Glib::ustring)tree_row[tree->_columns.get<Glib::ustring>(column)]).c_str();
  }
  
  static int get_int(TreeView *self, int row, int column)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreeRow tree_row;
    Gtk::TreePath path;

    path.append_index(row);

    if (tree->_list_store)
      tree_row= *tree->_list_store->get_iter(path);
    else
      return 0;

    return tree_row[tree->_columns.get<int>(column)];
  }
    
  static bool get_check(TreeView *self, int row, int column)
  {
    TreeViewImpl* tree = self->get_data<TreeViewImpl>();
    Gtk::TreeRow tree_row;
    Gtk::TreePath path;

    path.append_index(row);

    if (tree->_list_store)
      tree_row= *tree->_list_store->get_iter(path);
    else
      return false;

    return tree_row[tree->_columns.get<bool>(column)];
  }

public:
  static void init()
  {
    ::mforms::ControlFactory *f = ::mforms::ControlFactory::get_instance();

    f->_treeview_impl.create= &TreeViewImpl::create;
    f->_treeview_impl.add_column= &TreeViewImpl::add_column;
    f->_treeview_impl.end_columns= &TreeViewImpl::end_columns;
    f->_treeview_impl.clear_rows= &TreeViewImpl::clear_rows;
    f->_treeview_impl.delete_row= &TreeViewImpl::delete_row;
    f->_treeview_impl.add_row= &TreeViewImpl::add_row;
    f->_treeview_impl.count= &TreeViewImpl::count;
    f->_treeview_impl.set_string= &TreeViewImpl::set_string;
    f->_treeview_impl.set_int= &TreeViewImpl::set_integer;
    f->_treeview_impl.set_check= &TreeViewImpl::set_check;
    f->_treeview_impl.get_string= &TreeViewImpl::get_string;
    f->_treeview_impl.get_int= &TreeViewImpl::get_int; 
    f->_treeview_impl.get_check= &TreeViewImpl::get_check;
  }
};

  
}
}

#endif
