//#include "stdafx.h"

#include "grtdiff.h"
#include <assert.h>
#include "diffchange.h"
#include "changefactory.h"
#include "grtlistdiff.h"

#include "grtpp_util.h"
#include "grts/structs.h"

namespace grt
{

DiffChange *diff_make(const ValueRef &source, const ValueRef &target, const Omf *omf)
{
  return GrtDiff().diff(source, target, omf);
}

void diff_apply(ValueRef &source, DiffChange &change)
{
  change.apply(source);
}

void diff_dump(const DiffChange &change)
{
  change.dump_log(0);
}


void diff_delete(DiffChange *change)
{
  delete change;
}


bool is_any(const ValueRef &v)
{
  return !v.valueptr() || v.type() == AnyType;
}

inline bool XOR(bool a, bool b)
{
  return a ^ b;
}

bool are_compatible(const ValueRef &source, const ValueRef &target, Type *cmptype)
{
  Type st= source.type();
  Type tt= target.type();

  if (cmptype) 
    *cmptype= (st == tt || tt == AnyType)? st: tt;

  return ((st == tt) && !is_any(source)) || XOR(is_any(source) , is_any(target));
}


bool are_compatible_lists(const BaseListRef &source, const BaseListRef &target, Type *cmptype)
{
  Type stl= is_any(source)? AnyType: source.content_type();
  Type ttl= is_any(target)? AnyType: target.content_type();

  Type type= (stl == ttl || ttl == AnyType)? stl: ttl;
  if (cmptype)
    *cmptype= type;

  return ((stl == ttl) && !is_any(source)) || XOR(is_any(source) , is_any(target)) 
      && (is_simple_type(type) || type == ObjectType);
}

DiffChange *GrtDiff::diff(const ValueRef &source, const ValueRef &target, const Omf* omf)
{
  this->omf= omf? omf: NULL;
  return on_value(NULL, source, target);
}

DiffChange *GrtDiff::on_value(DiffChange *parent, const ValueRef &source, const ValueRef &target, bool ci /* case insensitive */)
{
  Type type;
  if (!are_compatible(source, target, &type))
    return on_uncompatible(parent, source, target);

  if (is_any(source))
    return ChangeFactory::create_value_added_change(parent, source, target);
  
  if (is_any(target))
    return ChangeFactory::create_value_removed_change(parent, source, target);

  // TODO use omf here to determine if we should replace value

  switch(type)
  {
  case IntegerType:
  case DoubleType:
  case StringType:
    return on_simple_value(parent, source, target, ci);
    break;
  case ListType:
    return on_list(parent, BaseListRef::cast_from(source), BaseListRef::cast_from(target));
    break;
  case DictType:
    return on_dict(parent, DictRef::cast_from(source), DictRef::cast_from(target));
    break;
  case ObjectType:
    return on_object(parent, ObjectRef::cast_from(source), ObjectRef::cast_from(target));
    break;
  default:
    break;
  }
  assert(0);
  return NULL;
}

typedef unsigned int uint;

DiffChange *GrtDiff::on_simple_value(DiffChange *parent, const ValueRef &source, const ValueRef &target, bool ci /* case insensitive */)
{
  return ChangeFactory::create_simple_value_change(parent, source, target, ci);
}

DiffChange *GrtDiff::on_object(DiffChange *parent, const ObjectRef &source, const ObjectRef &target)
{
  ChangeSet changes;
  MetaClass *meta= source.get_metaclass();
  
  do
  {
    for (MetaClass::MemberList::const_iterator iter= meta->get_members_partial().begin(); iter != meta->get_members_partial().end(); ++iter)
    {
      if (iter->second.overrides) continue;
      
      std::string name1= iter->second.name;

      std::string attr;

      int dontfollow= !iter->second.owned_object;

      attr= meta->get_member_attribute(name1, "dontdiff");
      int dontdiff= attr.size() && attr[0] == '1';
      attr= meta->get_member_attribute(name1, "caseinsensitive");
      int ci= attr.size() && attr[0] == '1';
      
      if (dontdiff)
        continue;
      
      ValueRef v1= source.get_member(name1);
      ValueRef v2= target.get_member(name1);

      if (dontfollow && (name1 != "owner"))
      {
        grt::Type t= UnknownType;

        if (v1.is_valid())
          t= v1.type();

        if (v2.is_valid())
        {
          if(t == UnknownType)
            t= v2.type();
          else
            assert(t == v2.type());
        }
       
        if((t == IntegerType) || (t == DoubleType) || 
          (t == StringType) || (t == ObjectType))
        {
          changes.append(
            ChangeFactory::create_object_attr_modified_change(
              parent, 
              source, 
              target, 
              name1, 
              ChangeFactory::create_simple_value_change(parent, v1, v2, false)
            ));
        }
        continue;
      }

      // this block is temporary here: for db.Index::indexType
      // 1. should be no difference between "INDEX" and "FOREIGN"
      // 2. for IndexColumn names should be ignored
      {
        // 1.
        if (name1 == "indexType" && source.is_instance("db.Index") && target.is_instance("db.Index") &&
            StringRef::can_wrap(v1) && StringRef::can_wrap(v2))
        {
          StringRef s1(StringRef::cast_from(v1)), s2(StringRef::cast_from(v2));
          if (s1 == "INDEX" && s2 == "FOREIGN")
            continue;
          if (s2 == "INDEX" && s1 == "FOREIGN")
            continue;
        }
        // 2.
        else if(name1 == "name" && source.is_instance("db.IndexColumn") && target.is_instance("db.IndexColumn") &&
                StringRef::can_wrap(v1) && StringRef::can_wrap(v2))
        {
          continue;
        }
      } 
      
      if ((v1.valueptr() || v2.valueptr()) && name1 != "owner")
      {
        if((name1 == "name") || (name1 == "oldName"))
        {
          const char *n1= StringRef::cast_from(v1).c_str();
          char *n1up= g_utf8_strup(n1, g_utf8_strlen(n1, -1));
          const char *n2= StringRef::cast_from(v2).c_str();
          char *n2up= g_utf8_strup(n2, g_utf8_strlen(n2, -1));
          bool eq= (strcmp(n1up, n2up) == 0);
          g_free(n1up);
          g_free(n2up);
          if(eq)
            continue;
        }

        changes.append(ChangeFactory::create_object_attr_modified_change(
          parent, source, target, name1, on_value(NULL, v1, v2)));
      }


    }
    
    meta= meta->parent();
  }
  while (meta != 0);
  
  return ChangeFactory::create_object_modified_change(parent, source, target, changes);
}
  
  
DiffChange *GrtDiff::on_list(DiffChange *parent, const BaseListRef &source, const BaseListRef &target)
{
  Type type;

  if (!are_compatible_lists(source, target, &type))
    return on_uncompatible(parent, source, target);

  return GrtListDiff().diff(source, target, omf);
}


DiffChange *GrtDiff::on_dict(DiffChange *parent, const DictRef &source, const DictRef &target)
{
  ChangeSet changes;

  for (internal::Dict::const_iterator iter= source.begin(); iter != source.end(); ++iter)
  {
    std::string key= iter->first;
    ValueRef source_item(iter->second);

    if (!target.has_key(key))
      changes.append(ChangeFactory::create_dict_item_removed_change(parent, source, target, key));
    else
      changes.append(ChangeFactory::create_dict_item_modified_change(parent, source, target, key, on_value(NULL, source_item, target.get(key))));
  }

  for (internal::Dict::const_iterator iter= target.begin(); iter != target.end(); ++iter)
  {
    std::string key= iter->first;
    ValueRef target_item(iter->second);

    if (!source.has_key(key))
      changes.append(ChangeFactory::create_dict_item_added_change(parent, source, target, key, target_item));
  }

  return ChangeFactory::create_dict_change(parent, source, target, changes);
}

DiffChange *GrtDiff::on_uncompatible(DiffChange *parent, const ValueRef &source, const ValueRef &target)
{
  return ChangeFactory::create_value_added_change(parent, source, target);
}

}
