

#include "grtpp.h"
#include "grtpp_util.h"
#include "grtpp_shell.h"
#include "grtpp_module_cpp.h"
#include "grtpp_undo_manager.h"

#include <algorithm>

#include "grtpp_shell_lua.h"

#include <glib.h>

#include "serializer.h"
#include "unserializer.h"

using namespace grt;




std::string grt::type_to_str(Type type)
{
  switch (type)
  {
    case UnknownType:
      return "";
    case IntegerType:
      return "int";
    case DoubleType:
      return "real";
    case StringType:
      return "string";
    case ListType:
      return "list";
    case DictType:
      return "dict";
    case ObjectType:
      return "object";
  }
  return "";
}


Type grt::str_to_type(const std::string &type)
{
  char ini= type[0];
  if (ini == 'i' && type == "int")
    return IntegerType;
  else if ((ini == 'd' && type == "double") || (ini == 'r' && type == "real"))
    return DoubleType;
  else if (ini == 's' && type == "string")
    return StringType;
  else if (ini == 'l' && type == "list")
    return ListType;
  else if (ini == 'd' && type == "dict")
    return DictType;
  else if (ini == 'o' && type == "object")
    return ObjectType;
  return UnknownType;
}



std::string Message::format(bool withtype) const
{
  std::string text;

  if (withtype)
  {
    switch (type)
    {
    case InfoMsg:  text= "Info: "; break;
    case WarningMsg: text= "Warning: "; break;
    case ErrorMsg: text= "Error: "; break;
    }
  }

  text+= this->text;

  if (!detail.empty())
    text+= " ("+detail+")";

  return text;
}

//--------------------------------------------------------------------------------------------------

type_error::type_error(Type expected, Type actual)
: std::logic_error(std::string("Type mismatch: expected type ").append(type_to_str(expected)).append(", but got ").append(type_to_str(actual)))
{
}

type_error::type_error(Type expected, Type actual, Type container)
: std::logic_error(std::string("Type mismatch: expected content-type ").append(type_to_str(expected)).append(", but got ").append(type_to_str(actual)))
{
}

type_error::type_error(const std::string &expected, const std::string &actual)
: std::logic_error(std::string("Type mismatch: expected object of type ").append(expected).append(", but got ").append(actual))
{
}

type_error::type_error(const std::string &expected, const std::string &actual, Type container)
: std::logic_error(std::string("Type mismatch: expected content object of type ").append(expected).append(", but got ").append(actual))
{
}  

//--------------------------------------------------------------------------------------------------

StringRef StringRef::format(const char *format, ...)
{
  va_list args;
  char *tmp;
  StringRef ret;

  va_start(args, format);
  tmp= g_strdup_vprintf(format, args);
  va_end(args);

  ret= StringRef(tmp);
  g_free(tmp);

  return ret;
}

//--------------------------------------------------------------------------------------------------

GRT::GRT()
: _check_serialized_crc(false), _verbose(false)
{
  _scanning_modules= false;

  _tracking_changes= 0;
  _shell= 0;
  
  if (getenv("GRT_VERBOSE"))
    _verbose= true;

  _undo_manager= new UndoManager(this);

  add_module_loader(new CPPModuleLoader(this));

  g_static_rec_mutex_init(&_message_mutex);
  
  // register metaclass for base class
  add_metaclass(MetaClass::create_base_class(this));
  
  _root= grt::DictRef(this);
}


GRT::~GRT()
{
  delete _shell;
  delete _undo_manager;
  

  /* this is crashing, must be tracked
  for (std::vector<Module*>::iterator iter= _modules.begin();
       iter != _modules.end(); ++iter)
  {
    delete *iter;
  }
  _modules.clear();
  */

  for (std::map<std::string,Interface*>::iterator iter= _interfaces.begin();
       iter != _interfaces.end(); ++iter)
    delete iter->second;
  _interfaces.clear();
  
  for (std::list<ModuleLoader*>::iterator iter= _loaders.begin();
       iter != _loaders.end(); ++iter)
    delete *iter;
  _loaders.clear();

  for (std::map<std::string,MetaClass*>::iterator iter= _metaclasses.begin();
       iter != _metaclasses.end(); ++iter)
    delete iter->second;
  _metaclasses.clear();

  g_static_rec_mutex_free(&_message_mutex);
}


void GRT::start_tracking_changes()
{
  _tracking_changes++;
}

void GRT::stop_tracking_changes()
{
  _tracking_changes--;
}

void GRT::set_verbose(bool flag)
{
  _verbose= flag;
}


UndoGroup *GRT::begin_undoable_action(UndoGroup *group)
{
  start_tracking_changes();
  return _undo_manager->begin_undo_group(group);
}


void GRT::end_undoable_action(const std::string &group_description)
{
  if (!_undo_manager->end_undo_group(group_description, true))
  {
    if (getenv("DEBUG_UNDO"))
      g_warning("'%s' was empty", group_description.c_str());
  }
  stop_tracking_changes();
}


void GRT::cancel_undoable_action()
{
  _undo_manager->cancel_undo_group();
  stop_tracking_changes();
}


void GRT::lock() const
{
}


void GRT::unlock() const
{
}

//--------------------------------------------------------------------------------------------------


void GRT::load_metaclasses(const std::string &file, std::list<std::string> *requires)
{
  xmlNodePtr root;
  xmlDocPtr doc;
  
  doc= internal::Unserializer::load_xmldoc(file);
  
  root= xmlDocGetRootElement(doc);
  
  if (root && xmlStrcmp(root->name, (xmlChar*)"gstructs")==0)
  {
    root= root->children;
    while (root)
    {
      if (xmlStrcmp(root->name, (xmlChar*)"gstruct")==0)
      {
        MetaClass *gstruct= MetaClass::from_xml(this, file, root);
        if (gstruct)
        {
          MetaClass *tmp;
          if ((tmp= get_metaclass(gstruct->name())) == 0)
          {
            add_metaclass(gstruct);
          }
          else if (tmp != gstruct)
          {
            delete gstruct;
            throw std::runtime_error("Duplicate struct "+gstruct->name());
          }

          // add it to the ordered list of metaclasses, we dont want placeholders here
          _metaclasses_list.push_back(gstruct);
        }
      }
      else if (xmlStrcmp(root->name, (xmlChar*)"requires")==0)
      {
        xmlChar *path= xmlGetProp(root, (xmlChar*)"file");
        if (path)
        {
          if (requires)
            requires->push_back((char*)path);          
          xmlFree(path);
        }
      }
      root= root->next;
    }
  }
  xmlFreeDoc(doc);
}


int GRT::scan_metaclasses_in(const std::string &directory, std::multimap<std::string,std::string> *requires)
{
  GDir *dir;
  const char *entry;
  size_t old_count= _metaclasses.size();
  
  dir= g_dir_open(directory.c_str(), 0, NULL);
  if (!dir)
    throw grt::os_error("Invalid path "+directory);

  while ((entry= g_dir_read_name(dir)) != NULL)
  {
    if ( (g_str_has_prefix(entry, "structs.")) && (g_str_has_suffix(entry, ".xml")) )
    {
      char *path= g_build_filename(directory.c_str(), entry, NULL);
      std::list<std::string> reqs;

      reqs.clear();
      load_metaclasses(path, &reqs);
      
      if (requires)
      {
        for (std::list<std::string>::const_iterator i= reqs.begin(); i != reqs.end(); ++i)
          requires->insert(std::pair<std::string,std::string>(path, *i));
      }
      
      g_free(path);
    }
  }
  
  g_dir_close(dir);
  
  return _metaclasses.size() - old_count;
}


void GRT::end_loading_metaclasses(bool check_class_binding)
{
  bool undefined= false;
  bool validate_error= false;

  for (std::map<std::string,MetaClass*>::iterator iter= _metaclasses.begin();
       iter != _metaclasses.end(); ++iter)
  {
    if (iter->second->placeholder())
    {
      undefined= true;
      g_warning("MetaClass '%s' is undefined but was referred in '%s'",
                iter->second->name().c_str(), iter->second->source().c_str());
    }
    if (!iter->second->validate())
      validate_error= true;
  }
  if (undefined)
    throw std::runtime_error("One or more undefined metaclass were referred by other structs");
  if (validate_error)
    throw std::runtime_error("Validation error in loaded metaclasses");

  // register GRT object classes
  internal::ClassRegistry::get_instance()->register_all(this);
  
  if (check_class_binding)
  {
    // check if there are any metaclasses with unbound members
    for (std::map<std::string,MetaClass*>::iterator iter= _metaclasses.begin();
         iter != _metaclasses.end(); ++iter)
    {
      if (!iter->second->is_bound())
        g_warning("Allocation function of '%s' is unbound, which probably means the implementing C++ class was not registered",
                  iter->second->name().c_str());
    }
  }
}


MetaClass *GRT::get_metaclass(const std::string &name) const
{
  std::map<std::string,MetaClass*>::const_iterator iter;
  
  if ((iter= _metaclasses.find(name)) == _metaclasses.end())
    return 0;
  return iter->second;
}


void GRT::add_metaclass(MetaClass *stru)
{
  _metaclasses[stru->name()]= stru;
}

//--------------------------------------------------------------------------------------------------


void GRT::set_root(const ValueRef &root)
{
  AutoLock(this);
  _root= root;
  if (_root.is_valid())
    _root.valueptr()->mark_global();
}



ValueRef GRT::get(const std::string &path) const
{
  AutoLock(this);
 
  return get_value_by_path(_root, path);
}


void GRT::set(const std::string &path, const ValueRef &value)
{
  AutoLock(this);
  
  if (!set_value_by_path(_root, path, value))
    throw grt::bad_item("Invalid path "+path);
}


ObjectRef GRT::find_object_by_id(const std::string &id, const std::string &subpath)
{
  /*
  std::map<std::string,internal::Object*>::const_iterator iter= _objects.find(id);
  if (iter == _objects.end())
    return ObjectRef(iter->second);
  return ObjectRef();
   */
  
  ValueRef start= get(subpath);

  if (start.is_valid())
  {
    switch (start.type())
    {
    case ListType:
      return find_child_object(BaseListRef::cast_from(start), id);
    case DictType:
      return find_child_object(DictRef::cast_from(start), id);
    case ObjectType:
      return find_child_object(ObjectRef::cast_from(start), id);
    default:
      throw std::invalid_argument("Value at "+subpath+" is not a container");
    }
  }
  return ObjectRef();
}


//--------------------------------------------------------------------------------------------------



void GRT::serialize(const ValueRef &value, const std::string &path,
                    const std::string &doctype, const std::string &version)
{
  internal::Serializer ser(this);
  
  ser.save_to_xml(value, path, doctype, version);
}


ValueRef GRT::unserialize(const std::string &path)
{
  internal::Unserializer unser(this, _check_serialized_crc);

  if (!g_file_test(path.c_str(), G_FILE_TEST_EXISTS))
    throw os_error(path);

  try
  {
    return unser.load_from_xml(path);
  }
  catch (std::exception &exc)
  {
    throw std::runtime_error(std::string("Error unserializing GRT data from ").append(path).append(": ").append(exc.what()));
  }
  return ValueRef();
}


ValueRef GRT::unserialize(const std::string &path, std::string &doctype_ret, std::string &version_ret)
{
  internal::Unserializer unser(this, _check_serialized_crc);
  
  if (!g_file_test(path.c_str(), G_FILE_TEST_EXISTS))
    throw os_error(path);
  try
  {
    return unser.load_from_xml(path, &doctype_ret, &version_ret);
  }
  catch (std::exception &exc)
  {
    throw grt_runtime_error("Error unserializing GRT data from "+path, 
                            exc.what());
  }
  return ValueRef();
}


xmlDocPtr GRT::load_xml(const std::string &path)
{
  return internal::Unserializer::load_xmldoc(path);
}


void GRT::get_xml_metainfo(xmlDocPtr doc, std::string &doctype_ret, std::string &version_ret)
{  
  internal::Unserializer::get_xmldoc_metainfo(doc, doctype_ret, version_ret);
}


ValueRef GRT::unserialize_xml(xmlDocPtr doc, const std::string &source_path)
{
  internal::Unserializer unser(this, _check_serialized_crc);

  try
  {
    return unser.unserialize_xmldoc(doc, source_path);
  }
  catch (std::exception &exc)
  {
    throw grt_runtime_error("Error unserializing GRT data" , exc.what());
  }
  return ValueRef();
}


std::string GRT::serialize_xml_data(const ValueRef &value, const std::string &doctype, const std::string &version)
{
  return internal::Serializer(this).serialize_to_xmldata(value, doctype, version);
}


ValueRef GRT::unserialize_xml_data(const std::string &data)
{
  return internal::Unserializer(this, _check_serialized_crc).unserialize_xmldata(data.data(), data.size());
}

//--------------------------------------------------------------------------------

void GRT::add_module_loader(ModuleLoader *loader)
{
  _loaders.push_back(loader);
}


bool GRT::load_module(const std::string &path, bool refresh)
{
  for (std::list<ModuleLoader*>::iterator loader= _loaders.begin();
       loader != _loaders.end(); ++loader)
  {
    if ((*loader)->check_file_extension(path))
    {
      if (_verbose)
        g_message("Trying to load module '%s' (%s)", path.c_str(), (*loader)->get_loader_name().c_str());
      
      Module *module= (*loader)->init_module(path);
      if (!module)
      {
//        if (_verbose)
          g_message("Failed loading module '%s' (%s)", path.c_str(), (*loader)->get_loader_name().c_str());
      }
      else
      {
        if (refresh)
          refresh_module(module);
        else
          register_new_module(module);
        return true;
      }
    }
  }
  return false;
}


ModuleLoader *GRT::get_module_loader(const std::string &name)
{
  for (std::list<ModuleLoader*>::const_iterator iter= _loaders.begin();
       iter != _loaders.end(); ++iter)
  {
    if ((*iter)->get_loader_name() == name)
      return *iter;
  }
  return 0;
}


int GRT::scan_modules_in(const std::string &path, const std::list<std::string> &exts,
                         bool reload)
{
  GDir *dir;
  const char *entry;
  int count= 0;
  GError *error= NULL;
  
  dir= g_dir_open(path.c_str(), 0, &error);
  if (!dir)
  {
    send_warning(strfmt("Cannot open module directory %s: %s", path.c_str(), error->message));
    g_error_free(error);
    return -1;
  }

  _scanning_modules= true;

  if (verbose())
    send_info(strfmt("Scanning module directory %s.", path.c_str()));

  while ((entry= g_dir_read_name(dir)) != NULL)
  {
    char *p= g_build_filename(path.c_str(), entry, NULL);
    std::string module_path;
    
    // check if it's a bundle directory
    module_path= module_path_in_bundle(p);
    if (module_path.empty())
      module_path= p;
    g_free(p);

    // if we were given a list of extensions, check if it matches
    if (!exts.empty())
    {
      std::string::size_type end= module_path.rfind('.');
      if (end == std::string::npos)
        continue;
      std::string name= module_path.substr(0, end);
      bool ok= false;

      for (std::list<std::string>::const_iterator ext= exts.begin();
        ext != exts.end(); ++ext)
      {
        if (g_str_has_suffix(name.c_str(), ext->c_str()))
        {
          ok= true;
          break;
        }
      }

      if (!ok)
        continue;
    }

    try
    {
      if (load_module(module_path, reload))
      {
        count++;
      }
    }
    catch (std::exception &exc)
    {
      send_warning(strfmt("Could not load %s: %s", entry, exc.what()));
    }
  }

  g_dir_close(dir);

  _scanning_modules= false;

  refresh_loaders();

  return count;
}


Module *GRT::get_module(const std::string &name)
{
  for (std::vector<Module*>::iterator iter= _modules.begin(); iter != _modules.end(); ++iter)
  {
    if ((*iter)->name() == name)
      return *iter;
  }
  return 0;
}


std::vector<Module*> GRT::find_modules_matching(const std::string &interface_name, const std::string &name_pattern)
{
  std::vector<Module*> result;

  for (std::vector<Module*>::const_iterator module= _modules.begin(); 
       module != _modules.end(); ++module)
       
  {
    bool ok= true;
    if (!interface_name.empty())
    {
      ok= false;
      
      if (std::find((*module)->get_interfaces().begin(), (*module)->get_interfaces().end(), 
                    interface_name) != (*module)->get_interfaces().end())
        ok= true;
    }

    if (ok && (name_pattern.empty() || g_pattern_match_simple(name_pattern.c_str(),
                                                              (*module)->name().c_str())))
    {
      result.push_back(*module);
    }
  }

  return result;
}


void GRT::refresh_loaders()
{
  for (std::list<ModuleLoader*>::iterator iter= _loaders.begin();
       iter != _loaders.end(); ++iter)
  {
    (*iter)->refresh();
  }
}


void GRT::register_new_module(Module *module)
{
  module->validate();

  if (get_module(module->name()))
    throw std::runtime_error("Duplicate module "+module->name());

  _modules.push_back(module);
  
  if (!_scanning_modules)
    refresh_loaders();
}


void GRT::refresh_module(Module *module)
{
  bool found= false;

  module->validate();

  for (std::vector<Module*>::iterator iter= _modules.begin();
    iter != _modules.end(); ++iter)
  {
    if ((*iter)->name() == module->name())
    {
      delete *iter;
      
      *iter= module;
      found= true;
      break;
    }
  }
  if (!found)
    register_new_module(module);
}


void GRT::register_new_interface(Interface *iface)
{
  _interfaces[iface->name()]= iface;
} 


const Interface *GRT::get_interface(const std::string &name)
{
  std::map<std::string,Interface*>::const_iterator iter;
  if ((iter= _interfaces.find(name)) == _interfaces.end())
    return 0;
  return iter->second;
}

//--------------------------------------------------------------------------------

void GRT::set_context_data(const std::string &key, void *value, void(*free_value)(void*))
{
  unset_context_data(key);

  _context_data[key].first= value; 
  _context_data[key].second= free_value;
}


void GRT::unset_context_data(const std::string &key)
{
  if (_context_data.find(key)!=_context_data.end())
  {
    if ( _context_data[key].second)
      (*_context_data[key].second)(_context_data[key].first);
    _context_data.erase(key);
  }
}


void *GRT::get_context_data(const std::string &key)
{
  return _context_data[key].first; 
}


//--------------------------------------------------------------------------------

bool GRT::init_shell(const std::string &shell_type)
{
  if (shell_type == "lua")
    _shell= new LuaShell(this);
  else
    throw std::runtime_error("Invalid shell type "+shell_type);
  
  _shell->init();

  return true;
}


Shell *GRT::get_shell()
{
  return _shell;
}

//--------------------------------------------------------------------------------

MessageSlot GRT::set_message_handler(const MessageSlot &slot)
{
  g_static_rec_mutex_lock(&_message_mutex);
  MessageSlot old(_message_slot);
  
  _message_slot= slot;
  g_static_rec_mutex_unlock(&_message_mutex);
  
  return old;
}


StatusQuerySlot GRT::set_status_query_handler(const StatusQuerySlot &slot)
{
  StatusQuerySlot old(_status_query_slot);
  
  _status_query_slot= slot;
  
  return old;
}


void GRT::make_output_visible()
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= ControlMsg;
  msg.text= "show";
  msg.detail= "";
  msg.timestamp= time(NULL);
  msg.progress= 0.0;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
}


void GRT::send_error(const std::string &message, const std::string &details)
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= ErrorMsg;
  msg.text= message;
  msg.detail= details;
  msg.timestamp= time(NULL);
  msg.progress= 0.0;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
  
  if (verbose())
    g_message("ERROR: %s    %s", message.c_str(), details.c_str());
}


void GRT::send_warning(const std::string &message, const std::string &details)
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= WarningMsg;
  msg.text= message;
  msg.detail= details;
  msg.timestamp= time(NULL);
  msg.progress= 0.0;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
  
  if (verbose())
    g_message("WARNING: %s    %s", message.c_str(), details.c_str());
}


void GRT::send_info(const std::string &message, const std::string &details)
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= InfoMsg;
  msg.text= message;
  msg.detail= details;
  msg.timestamp= time(NULL);
  msg.progress= 0.0;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
}


void GRT::send_progress(float percentage, const std::string &message, const std::string &details)
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= ProgressMsg;
  msg.text= message;
  msg.detail= details;
  msg.timestamp= time(NULL);
  msg.progress= percentage;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
}


void GRT::send_verbose(const std::string &message)
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= VerboseMsg;
  msg.text= message;
  msg.detail= "";
  msg.timestamp= time(NULL);
  msg.progress= 0.0;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
}


void GRT::send_output(const std::string &text)
{
  g_static_rec_mutex_lock(&_message_mutex);
  Message msg;
  msg.type= OutputMsg;
  msg.text= text;
  msg.detail= "";
  msg.timestamp= time(NULL);
  msg.progress= 0.0;
  _message_slot(msg);
  g_static_rec_mutex_unlock(&_message_mutex);
}
