//#include "stdafx.h"

#include "grtpp_undo_manager.h"
#include "grtpp_util.h"

#ifdef _WIN32
#undef max
#endif

/**
 * @file  grtpp_undo_manager.cpp
 * @brief 
 */


using namespace grt;


void UndoAction::set_description(const std::string &description)
{
  _description= description;
}


//---------------------------------------------------------------------------------------------------

void SimpleUndoAction::dump(int indent) const
{
  g_message("%*s custom action (%s)", indent, "", _description.c_str());
}


//---------------------------------------------------------------------------------------------------

UndoObjectChangeAction::UndoObjectChangeAction(const ObjectRef &object, const std::string &member)
: _object(object), _member(member)
{
  _value= _object.get_member(_member);
}


UndoObjectChangeAction::UndoObjectChangeAction(const ObjectRef &object, const std::string &member, const ValueRef &value)
: _object(object), _member(member), _value(value)
{
}


void UndoObjectChangeAction::undo(UndoManager *owner)
{
  //owner->add_undo(new UndoObjectChangeAction(_object, _member));
  //owner->set_action_description(description());

  owner->get_grt()->start_tracking_changes();
  _object.set_member(_member, _value);
  owner->set_action_description(description());
  owner->get_grt()->stop_tracking_changes();
}


void UndoObjectChangeAction::dump(int indent) const
{
  g_message("%*s change '%s' in '%s' (%s)", indent, "", _member.c_str(), _object.id().c_str(), description().c_str());
}

//---------------------------------------------------------------------------------------------------


UndoListInsertAction::UndoListInsertAction(const BaseListRef &list, size_t index)
: _list(list), _index(index)
{
}


void UndoListInsertAction::undo(UndoManager *owner)
{
  /*
  if (_index == BaseListRef::npos)
  {
    owner->add_undo(new UndoListRemoveAction(_list, _list.get(_list.count()-1)));
    owner->set_action_description(description());

    _list.remove(_list.count()-1);
  }
  else
  {
    owner->add_undo(new UndoListRemoveAction(_list, _list.get(_index)));
    owner->set_action_description(description());

    _list.remove(_index);
  }*/
  if (_index == BaseListRef::npos)
  {
    owner->get_grt()->start_tracking_changes();
    _list.remove(_list.count()-1);
    owner->set_action_description(description());
    owner->get_grt()->stop_tracking_changes();
  }
  else
  {
    owner->get_grt()->start_tracking_changes();
    _list.remove(_index);
    owner->set_action_description(description());
    owner->get_grt()->stop_tracking_changes();
  }
}


void UndoListInsertAction::dump(int indent) const
{
  g_message("%*s insert %p to list %p[%li] of %s (%s)", indent, "",
            _index == BaseListRef::npos ? _list.get(_list.count()-1).valueptr() : _list.get(_index).valueptr(),
            _list.valueptr(),
            _index,
            fmt_simple_type_spec(_list->content_type_spec()).c_str(),
            description().c_str());
}

//---------------------------------------------------------------------------------------------------

UndoListReorderAction::UndoListReorderAction(const BaseListRef &list, size_t oindex, size_t nindex)
: _list(list), _oindex(oindex), _nindex(nindex)
{
}


void UndoListReorderAction::undo(UndoManager *owner)
{
  /*
  owner->add_undo(new UndoListReorderAction(_list, _nindex, _oindex));
  owner->set_action_description(description());
  _list.reorder(_nindex, _oindex);
   */
  owner->get_grt()->start_tracking_changes();
  _list.reorder(_nindex, _oindex);
  owner->set_action_description(description());
  owner->get_grt()->stop_tracking_changes();
}


void UndoListReorderAction::dump(int indent) const
{
  g_message("%*s reorder list of %s %li->%li (%s)", indent, "",
            fmt_simple_type_spec(_list->content_type_spec()).c_str(), _oindex, _nindex, 
            description().c_str());
}

//---------------------------------------------------------------------------------------------------

UndoListSetAction::UndoListSetAction(const BaseListRef &list, size_t index)
: _list(list), _index(index)
{
  _value= list.get(index);
}


void UndoListSetAction::undo(UndoManager *owner)
{
  /*
  owner->add_undo(new UndoListSetAction(_list, _index));
  owner->set_action_description(description());
  _list.gset(_index, _value);
   */
  owner->get_grt()->start_tracking_changes();
  _list.gset(_index, _value);
  owner->set_action_description(description());
  owner->get_grt()->stop_tracking_changes();
}


void UndoListSetAction::dump(int indent) const
{
  g_message("%*s set %li at list of %s (%s)", indent, "",
            _index, fmt_simple_type_spec(_list->content_type_spec()).c_str(),
            description().c_str());
}

//---------------------------------------------------------------------------------------------------

UndoListRemoveAction::UndoListRemoveAction(const BaseListRef &list, const ValueRef &value)
: _list(list), _value(value)
{
  bool found= false;
  for (size_t c= list.count(), i= 0; i < c; i++)
  {
    if (list[i].valueptr() == value.valueptr())
    {
      _index= i;
      found= true;
      break;
    }
  }
  // if the undo action is added after the value is deleted, it wont work
  if (!found)
    throw std::logic_error("attempt to add invalid undo operation");
}


UndoListRemoveAction::UndoListRemoveAction(const BaseListRef &list, size_t index)
: _list(list), _value(list.get(index)), _index(index)
{
}


void UndoListRemoveAction::undo(UndoManager *owner)
{
  /*
  owner->add_undo(new UndoListInsertAction(_list));
  owner->set_action_description(description());

  _list.ginsert(_value, _index);
   */
  owner->get_grt()->start_tracking_changes();
  _list.ginsert(_value, _index);
  owner->set_action_description(description());
  owner->get_grt()->stop_tracking_changes();
}


void UndoListRemoveAction::dump(int indent) const
{
  g_message("%*s remove from list %p[%li] of %s (%s)", indent, "",
            _list.valueptr(),
            _index,
            fmt_simple_type_spec(_list->content_type_spec()).c_str(),
            description().c_str());
}

//---------------------------------------------------------------------------------------------------


UndoDictSetAction::UndoDictSetAction(const DictRef &dict, const std::string &key)
: _dict(dict), _key(key)
{
  if (_dict.has_key(key))
  {
    _value= _dict.get(_key);
    _had_value= true;
  }
  else
    _had_value= false;
}


void UndoDictSetAction::undo(UndoManager *owner)
{
  if (_had_value)
  {
    /*
    owner->add_undo(new UndoDictSetAction(_dict, _key));
    owner->set_action_description(description());

    _dict.set(_key, _value);
     */
    owner->get_grt()->start_tracking_changes();
    _dict.set(_key, _value);
    owner->set_action_description(description());
    owner->get_grt()->stop_tracking_changes();
  }
  else
  {/*
    owner->add_undo(new UndoDictRemoveAction(_dict, _key));
    owner->set_action_description(description());

    _dict.remove(_key);
    */
    owner->get_grt()->start_tracking_changes();
    _dict.remove(_key);
    owner->set_action_description(description());
    owner->get_grt()->stop_tracking_changes();
  }
}


void UndoDictSetAction::dump(int indent) const
{
  g_message("%*s set %s in dict (%s)", indent, "", _key.c_str(), description().c_str());
}


//---------------------------------------------------------------------------------------------------

UndoDictRemoveAction::UndoDictRemoveAction(const DictRef &dict, const std::string &key)
: _dict(dict), _key(key)
{
  if (_dict.has_key(key))
  {
    _value= _dict.get(_key);
    _had_value= true;
  }
  else
    _had_value= false;
}


void UndoDictRemoveAction::undo(UndoManager *owner)
{
  if (_had_value)
  {
    /*
    owner->add_undo(new UndoDictSetAction(_dict, _key));
    owner->set_action_description(description());

    _dict.set(_key, _value);
     */
    owner->get_grt()->start_tracking_changes();
    _dict.set(_key, _value);
    owner->set_action_description(description());
    owner->get_grt()->stop_tracking_changes();
  }
  else
  {
    // nop
    owner->add_undo(new UndoDictRemoveAction(_dict, _key));
    owner->set_action_description(description());
  }
}


void UndoDictRemoveAction::dump(int indent) const
{
  g_message("%*s remove %s from dict (%s)", indent, "", _key.c_str(), description().c_str());
}

//---------------------------------------------------------------------------------------------------


UndoGroup::UndoGroup()
{
  _is_open= true;
}

UndoGroup::~UndoGroup()
{
  for (std::list<UndoAction*>::reverse_iterator iter= _actions.rbegin(); iter != _actions.rend(); ++iter)
    delete *iter;
}


void UndoGroup::undo(UndoManager *owner)
{
  owner->begin_undo_group();

  for (std::list<UndoAction*>::reverse_iterator iter= _actions.rbegin(); iter != _actions.rend(); ++iter)
  {
    (*iter)->undo(owner);
  }

  owner->end_undo_group();
  owner->set_action_description(UndoAction::description());
}


void UndoGroup::trim()
{ 
  std::list<UndoAction*>::iterator next, iter;
  next= _actions.begin();
  // delete closed groups that are empty or have a single action
  while (next != _actions.end())
  {
    UndoGroup *subgroup= dynamic_cast<UndoGroup*>(*next);
    iter= next++;

    if (subgroup && !subgroup->is_open())
    {
      subgroup->trim();

      if (subgroup->_actions.size() == 1)
      {
        UndoAction *content= subgroup->_actions.front();
        subgroup->_actions.clear();
        delete subgroup;

        // replace the group with the content
        *iter= content;
      }
      else if (subgroup->empty())
      {
        // remove this subgroup
        _actions.erase(iter);
        delete subgroup;
      }
    }
  }
}



UndoGroup *UndoGroup::get_deepest_open_subgroup()
{
  if (!_actions.empty())
  {
    UndoGroup *group= dynamic_cast<UndoGroup*>(_actions.back());
    if (group && group->is_open())
    {
      return group->get_deepest_open_subgroup();
    }
  }
  return _is_open ? this : 0;
}


void UndoGroup::close()
{
  // close the topmost open undo group
  UndoGroup *group= get_deepest_open_subgroup();
  if (group)
    group->_is_open= false;
  else
    g_warning("trying to close already closed undo group");
}


void UndoGroup::add(UndoAction *op)
{
  // add the action to the topmost open undo group
  UndoGroup *subgroup= get_deepest_open_subgroup();

  if (subgroup)
    subgroup->_actions.push_back(op);
  else
    throw std::logic_error("trying to add an action to a closed undo group");
}


bool UndoGroup::empty() const
{
  return _actions.empty();
}


void UndoGroup::set_description(const std::string &description)
{
  if (!_actions.empty() && _is_open)
  {
    UndoGroup *subgroup= dynamic_cast<UndoGroup*>(_actions.back());
    if (subgroup)
    {
      if (subgroup->_is_open)
        _actions.back()->set_description(description);
      else
        subgroup->set_description(description);
      return;
    }
    else
      _actions.back()->set_description(description);
  }

  if (!_is_open)
    UndoAction::set_description(description);
}


std::string UndoGroup::description() const
{
  if (!_actions.empty() && _is_open)
  {
    UndoGroup *subgroup= dynamic_cast<UndoGroup*>(_actions.back());
    if (subgroup && subgroup->_is_open)
    {
      return _actions.back()->description();
    }
  }
  return UndoAction::description();
}


void UndoGroup::dump(int indent) const
{
  g_message("%*s %sgroup (%s) {", indent, "", _is_open ? "open " : "", description().c_str());
  for (std::list<UndoAction*>::const_iterator iter= _actions.begin(); iter != _actions.end(); ++iter)
  {
    (*iter)->dump(indent+2);
  }
  g_message("%*s }", indent, "");
}

//---------------------------------------------------------------------------------------------------


UndoManager::UndoManager(GRT *grt)
{
  _owner= grt;
  _is_undoing= false;
  _is_redoing= false;
  _undo_limit= 0;
  _blocks= 0;
  g_static_rec_mutex_init(&_mutex);
}


UndoManager::~UndoManager()
{
  g_static_rec_mutex_free(&_mutex);
  reset();
}



void UndoManager::lock() const
{
  g_static_rec_mutex_lock(&_mutex);
}


void UndoManager::unlock() const
{
  g_static_rec_mutex_unlock(&_mutex);
}



void UndoManager::disable()
{
  _blocks++;
}


void UndoManager::enable()
{
  if (_blocks == 0)
    return;
  _blocks--;
}


void UndoManager::set_undo_limit(size_t limit)
{
  _undo_limit= limit;

  trim_undo_stack();
}


void UndoManager::trim_undo_stack()
{
  lock();
  if (_undo_limit > 0)
    _undo_stack.erase(_undo_stack.begin(), _undo_stack.begin() + std::max(0, (int)(_undo_stack.size() - _undo_limit)));
  unlock();
}


bool UndoManager::can_undo() const
{
  lock();
  bool empty= _undo_stack.empty();
  unlock();

  return !empty;
}


bool UndoManager::can_redo() const
{
  lock();
  bool empty= _redo_stack.empty();
  unlock();

  return !empty;
}


std::string UndoManager::undo_description() const
{
  std::string d;
  lock();
  if (can_undo())
    d= _undo_stack.back()->description();
  unlock();
  return d;
}


std::string UndoManager::redo_description() const
{
  std::string d;
  lock();
  if (can_redo())
    d= _redo_stack.back()->description();
  unlock();
  return d;
}


UndoAction *UndoManager::get_latest_undo_action() const
{
  lock();
  if (_undo_stack.empty())
  {
    unlock();
    return 0;
  }
  UndoAction *action= _undo_stack.back();

  UndoGroup *group= dynamic_cast<UndoGroup*>(action);
  while (group && group->is_open() && !group->empty())
  {
    action= group->get_actions().back();
    group= dynamic_cast<UndoGroup*>(action);
  }
  unlock();
  return action;
}


void UndoManager::reset()
{
  lock();
  for (std::deque<UndoAction*>::iterator iter= _undo_stack.begin(); iter != _undo_stack.end(); ++iter)
    delete *iter;
  _undo_stack.clear();

  for (std::deque<UndoAction*>::iterator iter= _redo_stack.begin(); iter != _redo_stack.end(); ++iter)
    delete *iter;
  _redo_stack.clear();

  unlock();
  _changed_signal.emit();
}


bool UndoManager::empty() const
{
  return _undo_stack.empty() && _redo_stack.empty();
}


UndoGroup *UndoManager::begin_undo_group(UndoGroup *group)
{
  if (_blocks > 0) 
  { // if blocked, delete the group given and return 0
    delete group;
    return 0;
  }

  // if no group given, just create the default type
  if (!group)
    group= new UndoGroup();

  add_undo(group);
  
  return group;
}


/** Closes an open undo group in the stack and sets its description 
 * 
 * @param description for the undo group
 * @param trim whether undo groups with a single action should be removed and replaced
 * with the containing action
 * 
 * @return true if the undo group was closed and added to the undo stack
 */ 
bool UndoManager::end_undo_group(const std::string &description, bool trim)
{
  if (_blocks > 0) 
    return false;
  
  UndoGroup *group= 0;
  std::deque<UndoAction*> *stack;

  if (_is_undoing)
    stack= &_redo_stack;
  else
    stack= &_undo_stack;

  if (stack->empty())
    throw std::logic_error("unmatched undo group (undo stack is empty)");

  group= dynamic_cast<UndoGroup*>(stack->back());
  if (!group)
    throw std::logic_error("unmatched undo group");

  if (group->empty())
  {
    stack->pop_back();
    delete group;
    if (getenv("DEBUG_UNDO"))
      g_message("undo group '%s' was empty, so it was deleted", description.c_str());

    return false;
  }
  else
  {
    group->close();
    if (!description.empty())
      group->set_description(description);
        
    /* have to 1st merge or check for signal_apply from the deleted groups
    if (!getenv("DEBUG_UNDO"))
    {
      if (trim)
        group->trim();

      // if the group has a single action, we can remove the group and leave the action
      if (trim && !group->is_open() && group->get_actions().size() == 1)
      {
        UndoAction *action= group->get_actions().front();
        if (action->description().empty() || action->description() == group->description())
        {
          action->set_description(group->description());
          stack->pop_back();
          stack->push_back(action);
          group->get_actions().clear();
          delete group;
        }
      }
    }
     */
    
    return true;
  }
}


void UndoManager::cancel_undo_group()
{
  // close (deepest open) group and undo it to revert half-changes
  if (end_undo_group())
  {
    // disable undo registration to avoid getting a redo action after undo
    disable();

    std::deque<UndoAction*> *stack;
    // undo the deepest open undo group
    if (_is_undoing)
      stack= &_redo_stack;
    else
      stack= &_undo_stack;

    UndoGroup *group;
    if (!stack->empty() && ((group= dynamic_cast<UndoGroup*>(stack->back()))))
      group->undo(this);

    enable();
  }
}


void UndoManager::set_action_description(const std::string &descr)
{
  if (_blocks > 0) return;  // added by tax (instructed by alfredo)
  
  lock();
  if (_is_undoing)
  {
    if (!_redo_stack.empty()) 
      _redo_stack.back()->set_description(descr);
  }
  else
  {
    if (!_undo_stack.empty())
      _undo_stack.back()->set_description(descr);
  }
  unlock();
  _changed_signal.emit();
}


std::string UndoManager::get_action_description() const
{
  if (_is_undoing)
    return _redo_stack.back()->description();
  else
    return _undo_stack.back()->description();
}


std::string UndoManager::get_running_action_description() const
{
  if (_is_redoing)
    return _redo_stack.back()->description();
  else if (_is_undoing)
    return _undo_stack.back()->description();
  return "";
}


void UndoManager::undo()
{
  if (_is_undoing)
    throw std::logic_error("unexpected nested undo"); // is this allowed??

  lock();
  if (can_undo())
  {
    UndoAction *cmd= _undo_stack.back();
    _is_undoing= true;
    unlock();

    cmd->undo(this);

    lock();
    _is_undoing= false;

    _undo_stack.pop_back();
    unlock();

    _undo_signal.emit(cmd);

    delete cmd;
  }
  else
    unlock();
}


void UndoManager::redo()
{
  if (_is_redoing)
    throw std::logic_error("unexpected nested redo"); // is this allowed??

  lock();
  if (can_redo())
  {
    UndoAction *cmd= _redo_stack.back();
    _is_redoing= true;
    unlock();

    cmd->undo(this);

    lock();
    _is_redoing= false;

    _redo_stack.pop_back();
    unlock();

    _redo_signal.emit(cmd);

    delete cmd;
  }
  else
    unlock();
}


void UndoManager::add_undo(UndoAction *cmd)
{
  if (_blocks > 0)
  {
    delete cmd;
    return;
  }

  lock();
  if (_is_undoing)
  {
    bool flag= false;
    if (!_redo_stack.empty())
    {
      UndoGroup *group= dynamic_cast<UndoGroup*>(_redo_stack.back());
      if (group && group->is_open())
      {
        group->add(cmd);
        flag= true;
      }
    }
    if (!flag)
      _redo_stack.push_back(cmd);
  }
  else
  {
    bool flag= false;
    if (!_undo_stack.empty())
    {
      UndoGroup *group= dynamic_cast<UndoGroup*>(_undo_stack.back());
      if (group && group->is_open())
      {
        group->add(cmd);
        flag= true;
      }
    }
    if (!flag)
    {
      if (getenv("DEBUG_UNDO") && !dynamic_cast<UndoGroup*>(cmd))
        g_message("added undo action that's not a group to top");
      _undo_stack.push_back(cmd);
      trim_undo_stack();
    }

    // if we're not undoing neither redoing, then reset the redo stack
    if (!_is_redoing)
    {
      for (std::deque<UndoAction*>::iterator iter= _redo_stack.begin(); iter != _redo_stack.end(); ++iter)
        delete *iter;
      _redo_stack.clear();
    }
  }
  unlock();

  _changed_signal.emit();
}


void UndoManager::add_simple_undo(const sigc::slot<void> &slot)
{
  add_undo(new SimpleUndoAction(slot));
}
