#include "stdafx.h"

#include <glib/gunicode.h>
#include "mysql_sql_parser_fe.h"
#include "myx_statement_parser.h"

#define MYSQL_LEX 1
#define MYSQL_SERVER

extern "C"
{
int pthread_dummy(int) { return 0; }
}

typedef void * YYSTYPE;
#define YYSTYPE_IS_DECLARED

// to stop complaint from compiler about a certain ATTRIBUTE_FORMAT in my_dbug.h
#define DBUG_OFF

#include "myx_lex_helpers.h"
#include "sql_lex.h"
#include "my_sys.h"
//!#include <sstream>
#include "myx_sql_parser.tab.hh"


int MYSQLlex(void **arg, void *yylex);
//void lex_init(void);
//void lex_start(LEX *lex, const uchar *buf, uint length);

const char *MYSQL_DEFAULT_CHARSET= "utf8_bin";
int yylex(void **yylval);


struct Context
{
  Mysql_sql_parser_fe::fe_process_sql_statement_callback cb;
  void* data;
  int err_count;
  bool ignore_dml;
  bool processing_create_statements;
  bool processing_alter_statements;
  bool processing_drop_statements;
};


class Lex_helper
{
public:
  Lex_helper(const char *statement)
  {
    lex_start(&_lex, reinterpret_cast<const unsigned char *>(statement), (unsigned int)strlen(statement));
    _lex.charset= get_charset_by_name(MYSQL_DEFAULT_CHARSET, MYF(0));
    lex_args.arg1= &_yystype;
    lex_args.arg2= &_lex;
    myx_set_parser_source(statement);
  }
  ~Lex_helper()
  {
    myx_free_parser_source();
  }
  LEX * lex() { return &_lex; }
private:
  YYSTYPE _yystype;
  LEX _lex;
};


#define LEX_HELPER(statement) Lex_helper _lex_helper(statement);


bool is_statement_relevant(const char *statement, const Context *context)
{
  LEX_HELPER(statement)

  void *token= NULL;
  yylex(&token);
  if (token)
  {
    MyxSQLTreeItem *item= static_cast<MyxSQLTreeItem *>(token);
    const char *value= item->value();
    if (are_strings_eq_ci("USE", value)
        || (context->processing_create_statements && are_strings_eq_ci("CREATE", value))
        || (context->processing_alter_statements && are_strings_eq_ci("ALTER", value))
        || (context->processing_drop_statements && are_strings_eq_ci("DROP", value))
        || are_strings_eq_ci("BEGIN", value)
        || are_strings_eq_ci("SELECT", value)
        )
      return true;
  }

  return false;
}


// this function removes comment braces of form /*!NUMBER */
// making their contents a part of the query itself

void remove_versioning_comments(std::string& sql, CHARSET_INFO *cs, bool *ignore_statement)
{
  const char *begin= sql.c_str();
  const char *ptr= begin;
  const char *endptr= ptr + sql.length();

  while(1)
  {
    for(; (ptr < endptr) && (*ptr != '/'); ptr += max(my_mbcharlen(cs, *ptr),1))
      ;

    if((ptr + 3) >= endptr)
      break;

    if((ptr[1] != '*') || (ptr[2] != '!'))
    {
      ptr += my_mbcharlen(cs, *ptr);
      continue;
    }
    
    const char *start_start= ptr;

    int digit_count= 0;
    for(ptr= ptr + 3; (ptr < endptr) && (my_isdigit(cs, *ptr)); ptr += max(my_mbcharlen(cs, *ptr),1), digit_count++)
      ;

    if(digit_count == 0)
      continue;

    const char *start_end= ptr;

    // in case of parsing mysqldump files, there are 'create table' statements embraced with comments.
    // they are to be ignored, because they relate to views, not tables.
    if (ignore_statement)
      *ignore_statement= (0 == strncmp(ptr, " CREATE TABLE", 13));

    {
      register bool quoted= false;
      register bool escaped= false;
      register bool commented= false;
      int nested_comments_count= 1;
      char quot_sym= 0;

      for(; ptr < endptr-1; ++ptr)
      {
        escaped= (!commented && quoted && ('\\' == ptr[0])) ? !escaped : false;

        switch (ptr[0])
        {
        case '#':
          if ((1 == nested_comments_count) && !quoted)
            commented= true;
          break;

        case '\n':
          commented= false;
          break;

        case '/':
          if (!commented && !quoted && ('*' == ptr[1]))
            ++nested_comments_count;
          break;

        case '*':
          if (!commented && !quoted && ('/' == ptr[1]))
          {
            if (0 == --nested_comments_count)
              goto end;
          }
          break;

        case '"':
        case '\'':
          if (!escaped && !commented)
          {
            if (quoted)
            {
              if (ptr[0] == quot_sym)
              {
                quoted= false;
                quot_sym= 0;
              }
            }
            else
            {
              quoted= true;
              quot_sym= ptr[0];
            }
          }
          break;
        }
      }
end: ;
    }

    if(ptr >= endptr)
      break;

    // replace comments in-place with spaces
    sql.replace(start_start - begin, start_end - start_start, start_end - start_start, ' ');
    sql.replace(ptr - begin, 2, 2, ' ');

    ptr += 2;
  }
}


Mysql_sql_parser_fe::Mysql_sql_parser_fe()
:
ignore_dml(true),
processing_create_statements(true),
processing_alter_statements(true),
processing_drop_statements(true)
{
  MyxSQLTreeItem::tree(NULL);
  lex_init();
}


int Mysql_sql_parser_fe::parse_sql_script(const char *sql_script, fe_process_sql_statement_callback cb, void *user_data)
{
  Context context= {cb, user_data, 0, ignore_dml, processing_create_statements, processing_alter_statements, processing_drop_statements};

  CHARSET_INFO *cs= get_charset_by_name(MYSQL_DEFAULT_CHARSET, MYF(0));
  myx_process_sql_statements(sql_script, cs, &process_sql_statement_cb, &context, MYX_SPM_NORMAL_MODE);

  return context.err_count;
}


int Mysql_sql_parser_fe::parse_sql_script_file(const char *sql_script_filename, fe_process_sql_statement_callback cb, void *user_data)
{
  Context context= {cb, user_data, 0, ignore_dml, processing_create_statements, processing_alter_statements, processing_drop_statements};

  CHARSET_INFO *cs= get_charset_by_name(MYSQL_DEFAULT_CHARSET, MYF(0));
  myx_process_sql_statements_from_file(sql_script_filename, cs, &process_sql_statement_cb, &context, MYX_SPM_DELIMS_REQUIRED);

  return context.err_count;
}

int Mysql_sql_parser_fe::process_sql_statement_cb(const MyxStatementParser *splitter, const char *statement, void *context_ptr)
{
  // possible values for result:
  // -1 - statement was ignored
  // 0 - statement was successfully processed
  // 1 - error occured during statement processing
  int result= 0;

  Context *context= reinterpret_cast <Context *> (context_ptr);

  if (!context || !context->cb)
    return -1;

  // check if statement is in utf8 encoding
  if (!g_utf8_validate(statement, -1, NULL))
  {
    int stmt_lc= 1;
    {
      const char *c= statement - 1;
      while (c)
      {
        switch (*++c)
        {
        case '\n':
          ++stmt_lc;
        case '\r':
          break;
        default:
          c= NULL;
          break;
        }
      }
    }
    std::string err_msg= "SQL statement starting from pointed line contains non UTF8 characters";
    context->cb(context->data, splitter, statement, NULL, stmt_lc, err_msg);
    context->err_count++;
    return 1;
  }

  std::string sql(statement);

  // stripe comments before further statement processing because
  // mysqldump puts the whole DDL in comments e.g. for triggers
  bool ignore_statement= false;
  remove_versioning_comments(sql, get_charset_by_name(MYSQL_DEFAULT_CHARSET, MYF(0)), &ignore_statement);

  // filter inappropriate statements
  if (ignore_statement || (context->ignore_dml && !is_statement_relevant(sql.c_str(), context)))
    return -1; // ignored

  LEX_HELPER(sql.c_str())
/*
  //! debugging
  yytokentype res;
  do
  {
    void *yyitem;
    res= (yytokentype) yylex(&yyitem);
    MyxSQLTreeItem *item= (MyxSQLTreeItem *) yyitem;
    std::cout << (item->get_name() ? item->get_name() : "") << ": " << item->get_value() << std::endl;
  }
  while (res);
*/
  myx_parse();
  //this is a workaround for memory corruption 
  LEX lex_copy;
  memcpy(&lex_copy, _lex_helper.lex(), sizeof(lex_copy));

  const MyxSQLTreeItem *tree= MyxSQLTreeItem::tree();
/*
  //! debugging
  if (tree)
  {
    std::ofstream fs("treedump.xml");
    fs << *tree;
    fs.flush();
  }
*/
  // in case of syntax error extend err message with context
  std::string err_msg= myx_get_err_msg();
  if (!tree && (0 == strcmp(err_msg.c_str(), "syntax error")))
  {
    //const MyxSQLTreeItem *item= *(_lex_helper.lex()->yylval);
    const MyxSQLTreeItem *item = *lex_copy.yylval;
    if (item)
    {
      static const size_t MAX_SQL_CONTEXT_SIZE= 80;
      std::string statement_= statement;
      size_t boffset= item->stmt_boffset();
      std::string err_context= statement_.substr(boffset,
        std::min<size_t>(statement_.size() - boffset, MAX_SQL_CONTEXT_SIZE));
      err_msg.clear();
      err_msg
        .append("SQL syntax error near '")
        .append(err_context)
        .append("'");
    }
  }

  result= context->cb(context->data, splitter, sql.c_str(), tree, _lex_helper.lex()->yylineno, err_msg);
  if (0 != result)
    context->err_count++;

  return result;
}
