#include "stdafx.h"

#include "wb_module_validation_mysql.h"
#include <algorithm>
#include "grtpp_module_cpp.h"
#include "grtdb/db_object_helpers.h"
#include "grts/structs.db.mysql.h"
#include "grts/structs.workbench.physical.h"
#include "db.mysql/src/db_mysql_params.h"
#include "grtpp.h"
#include "myx_util_functions.h"


//#define DEBUG

static bool is_int_correct(const char *str, std::string *error);
static bool is_double_correct(const char *str, std::string *error);
static bool is_string_correct(const char *str, std::string *error);
static bool is_datetime_correct(const char *str, std::string *error);
static bool is_bool_correct(const char *str, std::string *error);

//-----------------------------------------------------------------------------
void MySQLValidator::create_integrity_checks(val::ChainsSet *set)
{
  val::Chain<db_UserRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_user));
  val::Chain<db_mysql_ViewRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_view));
  val::Chain<db_mysql_IndexRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_index));
  val::Chain<db_mysql_TriggerRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_trigger));
  
  val::Chain<db_mysql_ColumnRef>::Ptr ccol = val::Chain<db_mysql_ColumnRef>::chain(set);
  ccol->add_check(val::atom(this, &MySQLValidator::integrity_check_column));
  ccol->add_check(val::atom(this, &MySQLValidator::integrity_check_column_default_value));
  ccol->add_check(val::atom(this, &MySQLValidator::integrity_check_partition_scheme));

  val::Chain<db_mysql_TableRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_table));
  val::Chain<db_mysql_ForeignKeyRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_foreign_key));
  val::Chain<db_mysql_ForeignKeyRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::integrity_check_foreign_key_with_engine_type));
}

//-----------------------------------------------------------------------------
void MySQLValidator::create_syntax_checks(val::ChainsSet *set)
{
  val::Chain<db_mysql_ViewRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_view));
  val::Chain<db_mysql_TableRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_table));
  val::Chain<db_mysql_ColumnRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_column));
  val::Chain<db_mysql_IndexRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_index));
  val::Chain<db_mysql_TriggerRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_trigger));
  val::Chain<db_mysql_RoutineRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_routine));
  val::Chain<db_mysql_ForeignKeyRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_foreign_key));

  create_dup_ids_validation(set);
}

//-----------------------------------------------------------------------------
void MySQLValidator::create_dup_ids_validation(val::ChainsSet *set)
{
  val::Chain<db_mysql_TableRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_dup_names_in_tables));
  val::Chain<db_mysql_ViewRef>::chain(set)->add_check(val::atom(this, &MySQLValidator::syntax_check_dup_names_in_views));
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_view(const db_mysql_ViewRef& view) const
{
  check_for_reserwed_words("View", view);
  check_for_invalid_chars("View", view);


  std::string sqlDefinition= view->sqlDefinition();
  int code = -1;
  const int len = sqlDefinition.size();
  if ( len == 0 )
  {
    resultsList->add_error("Syntax error in view '%s'. View code is empty", view->name().c_str());
  }
  else if ( 1 != ( code = _sql_syntax_check->checkViewSyntax(sqlDefinition.c_str())) )
  {
    if ( len > 32 )
    {
      sqlDefinition.erase(32, -1);
      sqlDefinition += "...";
    }
    resultsList->add_error("Syntax error in view '%s'. View code is '%s'", view->name().c_str(), sqlDefinition.c_str());
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_trigger(const db_mysql_TriggerRef& trigger) const
{
  check_for_reserwed_words("Trigger", trigger);
  check_for_invalid_chars("Trigger", trigger);

  std::string sqlDefinition= trigger->sqlDefinition();
  sqlDefinition= "DELIMITER //\n"+ sqlDefinition +"//";

  if ( 1 != _sql_syntax_check->checkTriggerSyntax(trigger->sqlDefinition().c_str()) )
  {
    resultsList->add_error("Syntax error in trigger %s", trigger->name().c_str());
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_routine(const db_mysql_RoutineRef& routine) const
{
  std::string sqlDefinition = routine->sqlDefinition();
  check_name_length("Routine", routine, 64);

  sqlDefinition= "DELIMITER //\n"+ sqlDefinition +"//";
  if ( 1 != _sql_syntax_check->checkRoutineSyntax(sqlDefinition.c_str()) )
  {
    resultsList->add_error("Syntax error in routine %s", routine->name().c_str());
  }

  check_for_reserwed_words("Routine", routine);
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_table(const db_mysql_TableRef& object) const
{
  check_for_reserwed_words("Table", object);
  check_for_invalid_chars("Table", object);
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_column(const db_mysql_ColumnRef& object) const
{
  check_for_reserwed_words("Column", object);
  check_for_invalid_chars("Column", object);
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_index(const db_mysql_IndexRef& object) const
{
  check_for_reserwed_words("Index", object);
  check_for_invalid_chars("Index", object);
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_foreign_key(const db_mysql_ForeignKeyRef& object) const
{
  check_for_reserwed_words("Foreign key", object);
  check_for_invalid_chars("Foreign key", object);
}

//-----------------------------------------------------------------------------
void MySQLValidator::check_for_reserwed_words(const char* type, const GrtNamedObjectRef& object) const
{
  if (db_DatabaseObjectRef::can_wrap(object) && !db_RoutineGroupRef::can_wrap(object))
  {
    if (dbmysql::is_word_reserved(object->name().c_str(), object.get_grt()))
      resultsList->add_error("%s name '%s' is a reserved word", type, object->name().c_str());
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::check_for_invalid_chars(const char* type, const GrtNamedObjectRef& object) const
{
  if (db_DatabaseObjectRef::can_wrap(object) && !db_RoutineGroupRef::can_wrap(object))
  {
    if (!dbmysql::check_valid_characters(object->name().c_str()))
      resultsList->add_error("%s name '%s' contains invalid characters", type, object->name().c_str());
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_dup_names_in_tables(const db_mysql_TableRef& table) const
{
  check_dup_names("Table", table);
}

//-----------------------------------------------------------------------------
void MySQLValidator::syntax_check_dup_names_in_views(const db_mysql_ViewRef& view) const
{
  check_dup_names("View", view);
}

//-----------------------------------------------------------------------------
void MySQLValidator::check_dup_names(const char* type, const GrtNamedObjectRef& object) const
{
  const std::string obj_name = object->name().c_str();
  
  Names::const_iterator it = names.find(obj_name);
  if ( names.end() == it )
    names.insert(std::pair<std::string, GrtNamedObjectRef>(obj_name, object));
  else
    resultsList->add_error("Duplicated name for %s. %s with name '%s' already exists.", 
                   type, 
                   it->second->get_metaclass()->name().c_str(),
                   obj_name.c_str()
                );
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_schema(const db_mysql_SchemaRef& schema) const
{
  _schema = schema;
  if ( !schema.is_valid() )
  {
    resultsList->add_error("Invalid schema");
    return;
  }
  
  add_total_items_to_check(schema->tables().count());
  add_total_items_to_check(schema->views().count());
  add_total_items_to_check(schema->routines().count());

  val::Chain<db_mysql_SchemaRef>::validate(schema, cs);
  
  reporter->report_info("Validating tables"); 
  forEach(schema->tables()  , this, &MySQLValidator::walk_table  );
  reporter->report_info("Validating views"); 
  forEach(schema->views()   , this, &MySQLValidator::walk_view   );
  reporter->report_info("Validating routines"); 
  forEach(schema->routines(), this, &MySQLValidator::walk_routine);
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_user(const db_UserRef& user) const
{
  if ( !user.is_valid() )
  {
    resultsList->add_error("Invalid user");
    return;
  }
  val::Chain<db_UserRef>::validate(user, cs);
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_table(const db_mysql_TableRef& table) const
{
  if ( !table.is_valid() )
  {
    resultsList->add_error("Invalid table in schema '%s'", _schema->name().c_str());
    return;
  }
  _table = table;
  
  const char* nm = table->name().is_valid() ? table->name().c_str() : "<null>";
  val::Chain<db_mysql_TableRef>::validate(table, cs);

  //Order is important, check of columns should go before check of indices
  forEach(table->columns()    , this, &MySQLValidator::walk_column);
  forEach(table->indices()    , this, &MySQLValidator::walk_index);
  forEach(table->foreignKeys(), this, &MySQLValidator::walk_foreign_key);
  forEach(table->triggers()   , this, &MySQLValidator::walk_trigger);

  if ( autoIncColumns.size() != 0 )
  {
    StringList::const_iterator       it   = autoIncColumns.begin();
    const StringList::const_iterator last = autoIncColumns.end();
    for (; last != it; ++it)
    {
      resultsList->add_error("Column '%s'.'%s' has the auto_increment attribute set without being part of an index.",
                            nm,
                            it->c_str()
                           );
    }
  }
  item_check_passed(table.get_grt());
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_view(const db_mysql_ViewRef& view) const
{
  if ( !view.is_valid() )
  {
    resultsList->add_error("Invalid view in schema '%s'", _schema->name().c_str());
    return;
  }

  val::Chain<db_mysql_ViewRef>::validate(view, cs);
  item_check_passed(view.get_grt());
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_routine(const db_mysql_RoutineRef& routine) const
{
  if ( !routine.is_valid() )
  {
    resultsList->add_error("Invalid routine in schema '%s'", _schema->name().c_str());
    return;
  }
  val::Chain<db_mysql_RoutineRef>::validate(routine, cs);
  item_check_passed(routine.get_grt());
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_column(const db_mysql_ColumnRef& column) const
{
  if ( !column.is_valid() )
  {
    resultsList->add_error("Invalid column in table '%s'", _table->name().c_str());
    return;
  }

  val::Chain<db_mysql_ColumnRef>::validate(column, cs);
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_index(const db_mysql_IndexRef& index) const
{
  if ( !index.is_valid() )
  {
    resultsList->add_error("Invalid index in table '%s'", _table->name().c_str());
    return;
  }
  
  _index = index;
  val::Chain<db_mysql_IndexRef>::validate(index, cs);
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_foreign_key(const db_mysql_ForeignKeyRef& fk) const
{
  if ( !fk.is_valid() )
  {
    resultsList->add_error("Invalid foreign key in table '%s'", _table->name().c_str());
    return;
  }

  val::Chain<db_mysql_ForeignKeyRef>::validate(fk, cs);
}

//-----------------------------------------------------------------------------
void MySQLValidator::walk_trigger(const db_mysql_TriggerRef& trigger) const
{
  if ( !trigger.is_valid() )
  {
    resultsList->add_error("Invalid trigger in table '%s'", _table->name().c_str());
    return;
  }

  val::Chain<db_mysql_TriggerRef>::validate(trigger, cs);
}

//-----------------------------------------------------------------------------
void MySQLValidator::validate(const db_mysql_CatalogRef& cat, val::ChainsSet *set)
{
  totalItemsToCheck = 0.0;
  passedChecks      = 0;
  
  names.clear();
  foreignKeys.clear();
  autoIncColumnExists = 0;
  autoIncColumns.clear();
  autoIncPos = 0;

  cs = set;

  forEach(cat->schemata() , this, &MySQLValidator::walk_schema);
  forEach(cat->users()    , this, &MySQLValidator::walk_user);
  
  names.clear();
  foreignKeys.clear();
  autoIncColumnExists = 0;
  autoIncColumns.clear();
  autoIncPos = 0;
  
  //oresultsList->print();
}

//-----------------------------------------------------------------------------
void MySQLValidator::check_name_length(const char           *type, 
                                     const GrtNamedObjectRef &obj, 
                                     const int             maxNameLength
                                    ) const
{
  if ( obj->name().is_valid() )
  {
    if ( maxNameLength < (int)strlen(obj->name().c_str()) )
    {
      std::string nm = obj->name().c_str();
      nm.erase(64, -1);
      nm += "...";
      resultsList->add_error("%s name is too long. Maximum allowed length is %i characters. Table '%s'", 
                      type,
                      maxNameLength,
                      nm.c_str()
                  );
    }
  }
  else
  {
    resultsList->add_error("%s name is invalid.", type);
  }  
}

//-----------------------------------------------------------------------------
void MySQLValidator::check_table_comment(const db_mysql_TableRef& table) const
{
  unsigned int maxLength = 60;

  if ( table->comment().is_valid() )
  {
    if ( maxLength < strlen(table->comment().c_str()) )
    {
      resultsList->add_error("Table comment is too long. Maximum allowed length is %i characters. Table '%s'", 
                      maxLength,
                      table->name().is_valid() ? table->name().c_str() : "<null>"
                  );
    }
  }
  else
  {
    resultsList->add_error("Table comment is invalid. Table %s", 
                            table->name().is_valid() ? table->name().c_str() : "<null>"
                          );
  }  
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_view(const db_mysql_ViewRef& view) const
{
  check_name_length("View", view, 64);
}

//=============================================================================
//
//=============================================================================
struct DupRCCheck
{
  DupRCCheck(const MySQLValidator *v, const db_mysql_TableRef& table)
      : val(v)
      , _table(table)
      {}
  // This is called when we walk FKs in a table
  void walk_fks(const db_mysql_ForeignKeyRef& key)
  {
    if ( !key.is_valid() )
    {
      val->resultsList->add_error("Invalid foreign key in table '%s'", _table->name().c_str());
      return;
    }
    
    // Store ptr to fk for nested iteration
    fk = const_cast<db_mysql_ForeignKeyRef*>(&key);
    
    dups.clear();
    // We have all gathered, walk referenced columns
    //ct::for_each<ct::ReferedColumns>(key, *this);
    forEach(key->referencedColumns(),  this, &DupRCCheck::walk_columns);
  }

  // This is a nested iteration to actually check for duplicates
  void walk_columns(const db_ColumnRef& col)
  {
    if ( col.is_valid() )
    {
      // Build complete name to identify referenced column, using prebuilt name from operator() above
      grt::StringRef name = col->owner().is_valid() ? col->owner()->name() : "<null>";
      if ( !name.is_valid() )
        name = "'null'";
      std::string nm = name.c_str();
      nm += '.';
      nm += col.is_valid() ? col->name().c_str() : "'null'";
      // Make a lookup in Names which is std::map
      Names::const_iterator it = dups.find(nm);
      if ( it == dups.end() )
        dups.insert(std::pair<std::string, GrtNamedObjectRef>(nm, *fk));
      else
      {
        val->resultsList->add_error("Column '%s'.'%s' is referenced more than once in table '%s' foreign key '%s'.", 
                                    name.c_str(),
                                    col.is_valid() ? col->name().c_str() : "'null'",
                                    (*fk)->owner()->name().c_str(), 
                                    (*fk)->name().c_str()
                                  );
      }
    }
    else
      val->resultsList->add_error("Foreign key '%s' refers invalid column", (*fk)->name().c_str());
  }
  private:
    db_mysql_ForeignKeyRef       *fk;
    Names                      dups;
    const MySQLValidator      *val;
    const db_mysql_TableRef      &_table;
};

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_user(const db_UserRef& user) const
{
  check_name_length("User", user, 64);
}

//-----------------------------------------------------------------------------
static db_mysql_StorageEngineRef get_engine_by_name(grt::GRT *grt, const std::string &name);
void MySQLValidator::integrity_check_table(const db_mysql_TableRef& table) const
{
  const grt::StringRef engineName = table->tableEngine();  
  if (engineName.is_valid() )
  {
	  std::string ename = engineName;
  db_mysql_StorageEngineRef engine= get_engine_by_name(table.get_grt(), *engineName);  
  if (!engine.is_valid() && !ename.empty())//is emtpy engine name ok?
    resultsList->add_error("Table '%s' has a invalid Engine.",table->name().c_str());
  }
  autoIncColumnExists = 0;
  autoIncColumns.clear();
  
  check_name_length("Table", table, 64);
  check_table_comment(table);
  
  // Check for duplication of refenced columns
  // e.g. when table has two foreign keys each referrs to the same column in another table
  DupRCCheck checkDupRC(const_cast<MySQLValidator*>(this), table);
  //const grt::ListRef<db_mysql_ForeignKey> &tableFKs = table->foreignKeys();
  forEach(table->foreignKeys(), checkDupRC, &DupRCCheck::walk_fks);
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_column(const db_mysql_ColumnRef& column) const
{
  check_name_length("Column", column, 64);
  if ( (int)(column->autoIncrement()) )
  {
    const std::string colName = column->name().c_str();
    autoIncColumns.push_back(colName);
    
    if ( !autoIncColumnExists )
      autoIncColumnExists = 1;
    else
    {
      resultsList->add_error("Column '%s'.'%s' cannot get the AUTO INCREMENT attribute. "
                            "There is already a column with the AUTO INCREMENT attribute set, in table '%s'",
                            column->owner()->name().c_str(),
                            colName.c_str(),
                            column->owner()->name().c_str()
                           );
    }
      
    if ( strlen(column->defaultValue().c_str()) )
    {
      resultsList->add_error("Column '%s'.'%s'"
                            " cannot have a default value and at the same time the AUTO INCREMENT attribute set.",
                            column->owner()->name().c_str(),
                            colName.c_str()              
                           );
    }
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_auto_inc(const db_mysql_IndexColumnRef& col) const
{
  if ( !col.is_valid() )
  {
    resultsList->add_error("Invalid key in index '%s'.'%s'", 
                            _table->name().c_str(), 
                            _index->name().c_str()
                          );
    return;
  }
  
  ++autoIncPos;
  const char *idx = col->owner().is_valid() ? col->owner()->name().c_str() : "<null>";
  const char *tbl = col->owner()->owner()->name().is_valid() ? col->owner()->owner()->name().c_str() : "<null>";
  const db_ColumnRef rcol = col->referencedColumn();
  if ( !rcol.is_valid() )
  {
    resultsList->add_error("Empty column in index '%s'.'%s'", 
                            _table->name().c_str(),
                            _index->name().c_str()
                          );
    return;
  }
  
  const std::string name = rcol->name().c_str();

  StringList::iterator found = std::find(autoIncColumns.begin(),
                                         autoIncColumns.end(),
                                         name
                                        );

  if ( found != autoIncColumns.end() )
  {
    autoIncColumns.erase(found);
    if ( autoIncPos != 0 )
      resultsList->add_error("Index '%s'.'%s'->'%s' with auto increment should listed first. "
        "To correct this go to the table editor and re-check auto increment check box for the "
        "specified column. This will then automatically re-order all the columns.", tbl,idx, name.c_str());

    static const std::string numeric = "numeric";
    if ( rcol->simpleType()->group()->name() != numeric )
      resultsList->add_error("Column '%s'.'%s' is set to AUTO INCREMENT which is only valid for numeric types.", 
      tbl, name.c_str());
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_index(const db_mysql_IndexRef& index) const
{
  autoIncPos = -1;
  check_name_length("Index", index, 64);
  forEach(index->columns(), this, &MySQLValidator::integrity_check_auto_inc);
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_trigger(const db_mysql_TriggerRef& trigger) const
{
  check_name_length("Trigger", trigger, 64);
}


static db_mysql_StorageEngineRef get_engine_by_name(grt::GRT *grt, const std::string &name)
{
  grt::ListRef<db_mysql_StorageEngine> engines;
  Module *module= grt->get_module("DbMySQL");
  
  if (!module)
    throw std::logic_error("module DbMySQL not found");
  
  grt::BaseListRef args(grt);
  engines= grt::ListRef<db_mysql_StorageEngine>::cast_from(module->call_function("getKnownEngines", args));

  if (engines.is_valid())
  {
    for (grt::ListRef<db_mysql_StorageEngine>::const_iterator iter= engines.begin();
         iter != engines.end(); ++iter)
    {
      if ((*iter)->name() == name)
        return *iter;
    }
  }
  return db_mysql_StorageEngineRef();
}


//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_foreign_key_with_engine_type(const db_mysql_ForeignKeyRef& fk) const
{
  const db_mysql_TableRef table= db_mysql_TableRef::cast_from(fk->owner());
  if ( !table.is_valid() )
  {
    resultsList->add_error("Foreign key '%s' is orphaned. Foreign key is in list for table '%s'."
                          , fk->name().c_str()
                          , _table->name().c_str()
                          );
    return;
  }
  
  const grt::StringRef engineName = table->tableEngine();

  if (!engineName.is_valid())
    return;

  db_mysql_StorageEngineRef engine= get_engine_by_name(fk.get_grt(), *engineName);

  if (engine.is_valid() && !engine->supportsForeignKeys())
    resultsList->add_error("Table '%s' has a foreign key defined though the storage engine type does not support foreign keys.", 
                  table->name().c_str()
                );
}

template <typename T>
struct Print
{
  Print(ResultsList* rl, char* prefix) : rl_(rl), prefix_(prefix) {};
  
  void print(const T& t) const
  {
    rl_->add_warning("%s'%s'", prefix_, t->name().c_str());
  }
  
  ResultsList *rl_;
  char  *prefix_;
};


static inline void concatFlags(std::string& to, const std::string& sep, const grt::StringRef& ap)
{
  to.append(ap);
  to.append(sep);
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_ref_column(const db_mysql_ForeignKeyRef& fk,
                                                const db_mysql_TableRef& table,
                                                const db_ColumnRef& column,
                                                const db_ColumnRef &refColumn
                                               ) const
{
  if ( !column.is_valid() )
  {
    resultsList->add_error("Foreign key '%s'.'%s' referencing column is invalid.", 
                  table->name().c_str(),
                  fk->name().c_str()
                );

    if ( ! column->name().is_valid() )
      resultsList->add_error("The referencing column name is invalid for foreign key %s",
                    fk->name().c_str()
                  );

    return;
  }

  if ( !refColumn.is_valid() )
  {
    resultsList->add_error("Foreign key '%s'.'%s' does not reference a column.", 
                  table->name().c_str(), 
                  fk->name().c_str()
                );

    if ( ! column->name().is_valid() )
      resultsList->add_error("The referenced column name is invalid for foreign key %s",
                    fk->name().c_str()
                  );

    return;
  }


  const std::string colType = bec::ColumnHelper::format_column_type(column, 0, true);
  const std::string refColType = bec::ColumnHelper::format_column_type(refColumn, 0, true);
  
  int res = 0;
  if (( res = bec::ColumnHelper::compare_column_types(column, refColumn) ))
  {
    if ( res == COLUMNS_FLAGS_DIFFER )
    {
      std::string attr_col;
      std::string attr_refCol;
      static const std::string sep = "";      
      const grt::StringListRef &colFlags    = column->flags();
      const grt::StringListRef &refColFlags = refColumn->flags();
      std::for_each(colFlags.begin(), colFlags.end(),
                     boost::bind(&concatFlags, boost::ref(attr_col), sep, _1)
                   );
      std::for_each(refColFlags.begin(), refColFlags.end(),
                     boost::bind(&concatFlags, boost::ref(attr_refCol), sep, _1)
                   );
      const char* refColumnOwnerName = refColumn->owner().is_valid() ? refColumn->owner()->name().c_str() : "<null>";
      resultsList->add_error("The type of '%s'.'%s' <%s %s> does not match the referenced column: '%s'.'%s' <%s %s>. code: %i.",
                            table.is_valid() ? table->name().c_str() : "<null>", column->name().c_str(),
                            colType.c_str(), attr_col.c_str(),
                            refColumnOwnerName, refColumn->name().c_str(),
                            refColType.c_str(), attr_refCol.c_str(),
                            res
                           );
    }
    else
    {
      const char* refColumnOwnerName = refColumn->owner().is_valid() ? refColumn->owner()->name().c_str() : "<null>";
      resultsList->add_error("The type of '%s'.'%s' <%s> does not match the referenced column:'%s'.'%s' <%s>. code: %i.",
                            table.is_valid() ? table->name().c_str() : "<null>", column->name().c_str(),
                            colType.c_str(),
                            refColumnOwnerName, refColumn->name().c_str(),
                            refColType.c_str(),
                            res
                           );
    }
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_foreign_key(const db_mysql_ForeignKeyRef& fk) const
{
  if ( !fk->name().is_valid() )
    resultsList->add_error("Foreign key name is invalid in the table '%s'.", fk->owner()->name().c_str());

  const db_mysql_TableRef table = db_mysql_TableRef::cast_from(fk->owner());
  if ( !table.is_valid() )
  {
    resultsList->add_error("Foreign key '%s' is owned by nothing", fk->name().c_str());
  }
  const grt::ListRef<db_Column> refCols = fk->referencedColumns();
  const grt::ListRef<db_Column> cols = fk->columns();

  grt::ListRef<db_Column>::const_iterator       rc_it  = refCols.begin();
  const grt::ListRef<db_Column>::const_iterator rlast  = refCols.end();

  grt::ListRef<db_Column>::const_iterator       c_it  = cols.begin();
  const grt::ListRef<db_Column>::const_iterator last  = cols.end();
  
  for (; rlast != rc_it && last != c_it; ++rc_it, ++c_it )
  {
    const bool c_it_valid = c_it != last;
    const bool rc_it_valid = rc_it != rlast;

    if ( c_it_valid && rc_it_valid )
      integrity_check_ref_column(fk, table, *c_it, *rc_it);
    else
    {
      if ( !c_it_valid && rc_it_valid )
      {}
      else if ( c_it_valid && !rc_it_valid )
      {
        resultsList->add_error("Foreign key '%s'.'%s' references nothing.",
                              table->name().c_str(),
                              fk->name().c_str());
      }
      else
      {}
    }
  }

  if ( fk->referencedTable().is_valid() )
  {
    if ( table->rowFormat() != fk->referencedTable()->rowFormat() )
    {
      const char* rowfmt1 = table->rowFormat().c_str();
      const char* rowfmt2 = fk->referencedTable()->rowFormat().c_str();
      resultsList->add_error("The row format of table '%s' ('%s') does not correspond to that of the referenced table '%s' ('%s')."
                            ,table->name().c_str()
                            ,rowfmt1
                            ,fk->referencedTable()->name().c_str()
                            ,rowfmt2
                           );
    }
  }
  else
  {
    resultsList->add_error("Foreign key '%s' refers no table", fk->name().c_str());
  }
  
  const std::string obj_name = fk->name().c_str();
  Names::const_iterator firstInstance = foreignKeys.find(obj_name);
  if ( foreignKeys.end() == firstInstance )
    foreignKeys.insert(std::pair<std::string, GrtNamedObjectRef>(obj_name, GrtNamedObjectRef::cast_from(fk->owner())));
  else 
    resultsList->add_error("Duplicate foreign key name '%s'.'%s'. It has been already defined in %s %s.", 
                   fk->owner()->name().c_str(), 
                   obj_name.c_str(), 
                   "table",
                   firstInstance->second->name().c_str()
                );
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_column_default_value(const db_mysql_ColumnRef& column) const
{
  const char* column_name = column->name().is_valid() ? column->name().c_str() : "<null>";
  std::string column_owner_name;
  if ( column->owner().is_valid() )
    column_owner_name = column->owner()->name().is_valid() ? column->owner()->name().c_str() : "<null>";

  if ( column->isNotNull() )
  {
    if ( column->defaultValueIsNull() || !strcasecmp(column->defaultValue().c_str(), "NULL"))
    {
      resultsList->add_warning("NULL default value for not-null column '%s' in table '%s'."
                          ,column_name
                          ,column_owner_name.c_str());
    }
    else if ( !column->defaultValue().is_valid() )
    {
      resultsList->add_warning("Invalid or unset default value for not-null column '%s' in table '%s'."
                          ,column_name
                          ,column_owner_name.c_str());
    }
  }
  else
  {
    if ( !column->defaultValueIsNull() && !column->defaultValue().is_valid() )
    {
      resultsList->add_warning("Invalid or unset default value for column '%s' in table '%s'."
                          ,column_name
                          ,column_owner_name.c_str());
    }
  }

  const db_SimpleDatatypeRef simpleType = column->simpleType();
  if (!simpleType.is_valid() || !simpleType->group().is_valid())
  {
    resultsList->add_warning("Invalid type for column '%s' in table '%s'."
                          ,column_name
                          ,column_owner_name.c_str());
  }

  const std::string defaultValue= column->defaultValue().is_valid() ? column->defaultValue() : "";

  bool ok = false;
  if ( simpleType.is_valid() )
  {
    const std::string group= simpleType->group().id();
    std::string message;

    if(group == "com.mysql.rdbms.common.typegroup.numeric")
    {
      if (simpleType->numericScale()>0 || column->scale()>0)
        ok= is_double_correct(defaultValue.c_str(), &message);
      else
        ok= is_int_correct(defaultValue.c_str(), &message);
    }
    else if (group == "com.mysql.rdbms.common.typegroup.string")
    {
      if ( column->defaultValueIsNull() || !strcasecmp(column->defaultValue().c_str(), "NULL") )
        ok = true;
      else
      {
        ok= is_string_correct(defaultValue.c_str(), &message);
        if ( !ok && defaultValue.empty() )
          ok = true;
      }
    }
    else if (group == "com.mysql.rdbms.common.typegroup.datetime")
      ok= is_datetime_correct(defaultValue.c_str(), &message);
    else if (group == "com.mysql.rdbms.common.typegroup.various")
    {
      if(strcmp(simpleType->name().c_str(), "BOOLEAN") == 0)
        ok= is_bool_correct(defaultValue.c_str(), &message);
    }

    if ( defaultValue.empty() )
    {
      ok = true; // workaround to turn off warnings about normal things
      
      //TODO: Uncomment later, when we have options to turn on/off warnings
      //resultsList->add_warning("Empty default value for column '%s' in table '%s'"
      //                      ,column->name().c_str()
      //                      ,column->owner()->name().c_str());
    }
    
    if (!ok)
    {
      resultsList->add_warning("Invalid default value '%s' for column `%s`.`%s`: %s." 
                            ,defaultValue.c_str()
                            ,column_owner_name.c_str()
                            ,column_name
                            ,message.c_str()
                           );

    }
  }
  else if ( column->userType().is_valid() )
  {
    //TODO: add handling of type
  }
  else
  {
    resultsList->add_error("Type for column '%s'.'%s' is empty",
                            column->owner()->name().c_str(),
                            column->name().c_str()
                           );
  }
}

//-----------------------------------------------------------------------------
void MySQLValidator::integrity_check_partition_scheme(const db_mysql_ColumnRef&) const
{}

std::string WbModuleValidationMySQLImpl::getValidationDescription(GrtObjectRef root)
{
  if (workbench_physical_ModelRef::can_wrap(root))
    root= workbench_physical_ModelRef::cast_from(root)->catalog();

  if (db_mysql_CatalogRef::can_wrap(root))
    return "MySQL specific validations";

  return "";
}


int WbModuleValidationMySQLImpl::validateAll(GrtObjectRef root)
{
  get_grt()->make_output_visible();

  ResultsList result(this);
  //Validation::Ptr val(new Validation);
  
  MySQLValidator mysqlValidator(&result, this);
  //mysqlValidator.clearValidations();
  
  val::ChainsSet cs;
  mysqlValidator.create_integrity_checks(&cs); 
  mysqlValidator.create_syntax_checks(&cs); 

  if ( workbench_physical_ModelRef::can_wrap(root) )
  {
    db_mysql_CatalogRef cat(db_mysql_CatalogRef::cast_from(((workbench_physical_ModelRef::cast_from(root))->catalog())));
    mysqlValidator.validate(cat, &cs);
  }
  else if ( db_mysql_CatalogRef::can_wrap(root) )
  {
    db_mysql_CatalogRef cat(db_mysql_CatalogRef::cast_from(root));
    mysqlValidator.validate(cat, &cs);
  }
  
  report_summary("MySQL specific validate all");
    
  return result.errors_count() + result.warnings_count();
}

int WbModuleValidationMySQLImpl::validateIntegrity(GrtObjectRef root)
{
  get_grt()->make_output_visible();
 
  ResultsList result(this);
  
  MySQLValidator mysqlValidator(&result, this);

  val::ChainsSet cs;
  mysqlValidator.create_integrity_checks(&cs);

  if ( workbench_physical_ModelRef::can_wrap(root) )
  {
    db_mysql_CatalogRef cat(db_mysql_CatalogRef::cast_from(((workbench_physical_ModelRef::cast_from(root))->catalog())));
    mysqlValidator.validate(cat, &cs);
  }

  return result.errors_count() + result.warnings_count();
}

int WbModuleValidationMySQLImpl::validateSyntax(GrtObjectRef root)
{
  get_grt()->make_output_visible();

  ResultsList result(this);
  
  MySQLValidator mysqlValidator(&result, this);
  val::ChainsSet cs;
  mysqlValidator.create_syntax_checks(&cs); 

  mysqlValidator.validate(db_mysql_CatalogRef::cast_from(((workbench_physical_ModelRef::cast_from(root))->catalog())), &cs);

  return result.errors_count() + result.warnings_count();
}

int WbModuleValidationMySQLImpl::validateDuplicatesAdditions(GrtObjectRef object)
{
  const db_SchemaRef schema = db_SchemaRef::cast_from(object);
  
  if (!schema->views().count() || !schema->tables().count())
    return 1;

  int res= 1;

  std::set<std::string> table_names;
  for (size_t i= 0, count= schema->tables().count(); i < count; i++)
    table_names.insert(schema->tables().get(i)->name().c_str());

  for (size_t i= 0, count= schema->views().count(); i < count; i++)
    if (table_names.find(schema->views().get(i)->name().c_str()) != table_names.end()) 
    {
      report_error(
        "There is a table and a view with name '%s' in schema '%s'. Tables and views share the same"
        " namespace, so they must be unique.", 
        schema->views().get(i)->name().c_str(),
        schema->name().c_str());
      res= 0;
    };

  return res;
}


//-----------------------------------------------------------------------------
static bool is_bool_correct(const char *str, std::string *error)
{
  if((strcasecmp(str, "TRUE") == 0) || (strcasecmp(str, "FALSE") == 0))
    return true;
  return is_int_correct(str, error);
}

//-----------------------------------------------------------------------------
static bool is_int_correct(const char *str, std::string *error)
{
  // warning: todo: could be wrong result for realy long values
  if (*str == '\'') 
    str++; // this should solve DEFAULT '1' case but in general this will be handled by syntax check soon

  int i;
  return !!sscanf(str, "%d", &i);
}

//-----------------------------------------------------------------------------
static bool is_double_correct(const char *str, std::string *error)
{
  // warning: todo: not sure which formats are supported
  if (*str == '\'') 
    str++; // this should solve DEFAULT '1.0' case but in general this will be handled by syntax check soon

  float f;
  return !!sscanf(str, "%f", &f);
}

//-----------------------------------------------------------------------------
static bool is_string_correct(const char *str, std::string *error)
{
  if ( str )
  {
    if (*str != '\'' || *(str + strlen(str)-1) != '\'')
    {
      if (error)
        *error= "String defaults should be quoted with (') symbol.";

      return false;
    }
  }
  return true;
}

//-----------------------------------------------------------------------------
static bool is_datetime_correct(const char *str, std::string *error)
{
  if (*str=='\'')
  {
    glong len= g_utf8_strlen(str, -1);
    if (*(str + len - 1) != '\'')
    {
      if (error)
        *error= "Expected closing quote (').";
      return false;
    }
    if (len < 3)
    {
      if (error)
        *error= "Expression is too short.";
      return false;
    }
    // let's omit first and last character
    const char *beg= str + 1;
    const char *end= str + len -2;

    for (; beg < end; beg= max(beg+1, g_utf8_next_char(beg)))
      if (g_unichar_isalpha(*beg))
      {
        if (error)
          *error= "Alpha character found.";
        return false;
      }

    // TODO ensure that syntax is correct
    return true;
  }
  
  if (!str_is_numeric(str))
  {
    const char* ct= "CURRENT_TIMESTAMP";
    char *upper_str= g_strdup(str);
    str_toupper(upper_str);
    const char* res= strfindword(upper_str, ct);
    g_free(upper_str);

    if (res)
      return true; // TODO more advanced analysis should be here
    
    if (error)
      *error= "Quoted string should be used if non-digit characters are present.";
    return false;
  }
  size_t len= strlen(str);
  switch (len)
  {
    case 14: //YYYYMMDDHHMMSS
    case 12: //YYMMDDHHMMSS
    case 11: //YMMDDHHMMSS
    case 8: // YYYYMMDD 
    case 6: // YYMMDD
    case 5: // YMMDD
    case 4: // MMDD
    case 3: // MDD
    case 2: // YY
      return true;
  }
  if (error)
    *error= "Numeric value for date should have either 2, 3, 4, 5, 6, 8, 11, 12 or 14 characters.";
  return false;
}

GRT_MODULE_ENTRY_POINT(WbModuleValidationMySQLImpl);
