#ifndef _WB_MODULE_VALIDATION_H_
#define _WB_MODULE_VALIDATION_H_

//#define DEBUG

//! \addtogroup MySQLValidations MySQL specific validations
//!     MySQL specific validations module
//!
//! \code
//!  ResultsList result;
//!  Validation::Ptr val(new Validation);
//!  
//!  MySQLValidator mysqlValidator(&result, this);
//!  MySQLValidator::create_integrity_checks(&mysqlValidator, val); 
//!  MySQLValidator::create_syntax_checks(&mysqlValidator, val); 
//!
//!  mysqlValidator.validate(db_mysql_CatalogRef::cast_from(((workbench_physical_Model::cast_from(root)).catalog())), val);
//!
//!  result.print(*this);
//! \endcode
//!     @{

#include "grtsqlparser/sql_facade.h"
#include "grt/grt_reporter.h"
#include "grts/structs.db.h"
#include "grts/structs.db.mysql.h"


#include "wb_validation.h"

//=============================================================================
//
//=============================================================================
//! \class MySQLValidator
//! \brief Implements MySQL specific checks
//!
//! MySQLValidator class is passed as t parameter in \ref check.
//!
class MySQLValidator : public Validators
{
  public:
    MySQLValidator(ResultsList *rl, bec::Reporter *rep) 
      : Validators(rl, rep)
    {
      _sql_syntax_check= SqlFacade::instance_for_rdbms_name(rep->grt(), "Mysql");
    }

    //! Starts validation of specified catalog \a cat with Validation rules in \a val 
    void validate(const db_mysql_CatalogRef&, val::ChainsSet *cs);

    void walk_schema(const db_mysql_SchemaRef&) const;
    void walk_user(const db_UserRef&) const;
    void walk_table(const db_mysql_TableRef&) const;
    void walk_view(const db_mysql_ViewRef&) const;
    void walk_routine(const db_mysql_RoutineRef&) const;
    void walk_column(const db_mysql_ColumnRef&) const;
    void walk_index(const db_mysql_IndexRef&) const;
    void walk_foreign_key(const db_mysql_ForeignKeyRef&) const;
    void walk_trigger(const db_mysql_TriggerRef&) const;

    void check_name_length(const char* type, const GrtNamedObjectRef& obj, const int maxNameLength) const;
    void check_table_comment(const db_mysql_TableRef&) const;

    void integrity_check_user(const db_UserRef&) const;
    void integrity_check_view(const db_mysql_ViewRef&) const;
    void integrity_check_table(const db_mysql_TableRef&) const;
    void integrity_check_column(const db_mysql_ColumnRef&) const;
    void integrity_check_auto_inc(const db_mysql_IndexColumnRef&) const;
    void integrity_check_index(const db_mysql_IndexRef&) const;
    void integrity_check_trigger(const db_mysql_TriggerRef&) const;
    void integrity_check_foreign_key_with_engine_type(const db_mysql_ForeignKeyRef&) const;
    void integrity_check_foreign_key(const db_mysql_ForeignKeyRef&) const;
    void integrity_check_ref_column(const db_mysql_ForeignKeyRef&, const db_mysql_TableRef& table, const db_ColumnRef&, const db_ColumnRef &referencedColumns) const;
    void integrity_check_column_default_value(const db_mysql_ColumnRef&) const;
    void integrity_check_partition_scheme(const db_mysql_ColumnRef&) const;

    void syntax_check_view(const db_mysql_ViewRef&) const;
    void syntax_check_table(const db_mysql_TableRef&) const;
    void syntax_check_column(const db_mysql_ColumnRef&) const;
    void syntax_check_index(const db_mysql_IndexRef&) const;
    void syntax_check_routine(const db_mysql_RoutineRef&) const;
    void syntax_check_trigger(const db_mysql_TriggerRef&) const;
    void syntax_check_foreign_key(const db_mysql_ForeignKeyRef&) const;

    void check_for_reserwed_words(const char* type, const GrtNamedObjectRef& object) const;
    void check_for_invalid_chars(const char* type, const GrtNamedObjectRef& object) const;
    
    void syntax_check_dup_names_in_tables(const db_mysql_TableRef&) const;  
    void syntax_check_dup_names_in_views(const db_mysql_ViewRef&) const;  

    void check_dup_names(const char* type, const GrtNamedObjectRef&) const;  

    //! Adds to passed \a val with integrity checks
    void create_integrity_checks(val::ChainsSet *set);
    //! Adds to passed \a val with syntax checks
    void create_syntax_checks(val::ChainsSet *set);
    void create_dup_ids_validation(val::ChainsSet *set);

    void item_check_passed(grt::GRT *g) const
    {
      ++passedChecks;
      g->send_progress(passedChecks/totalItemsToCheck, "MySQL Validation");
    }
    void add_total_items_to_check(const int count) const
    {
      totalItemsToCheck += count;
    }
  private:
    //Validation              val;
    val::ChainsSet          *cs;

    mutable Names           names;
    mutable Names           foreignKeys;
    mutable float           totalItemsToCheck;
    mutable int             passedChecks;
    
    mutable int             autoIncColumnExists;
    mutable StringList      autoIncColumns;
    mutable int             autoIncPos;
    
    mutable db_mysql_TableRef  _table;       
    mutable db_mysql_SchemaRef _schema;
    mutable db_mysql_IndexRef  _index;

    SqlFacade *_sql_syntax_check;
};

//=============================================================================
//=============================================================================
//=============================================================================

#include "wb_module_validation_mysql_public_interface.h"
#include "grtpp_module_cpp.h"
#include "grtdb/db_object_helpers.h"
#include "interfaces/plugin.h"
#include "interfaces/wbvalidation.h"

#include "grts/structs.db.mgmt.h"
#include "grt/grt_walker.h"

using namespace bec;
using namespace grt;

#define WbModuleValidationMySQL_VERSION "1.0"


class WB_MODULE_VALIDATION_MYSQLPUBLIC_FUNC WbModuleValidationMySQLImpl
  : public ModuleImplBase,
    public PluginInterfaceImpl,
    public WbValidationInterfaceImpl,
    protected Reporter
{
public:

  WbModuleValidationMySQLImpl(grt::CPPModuleLoader *ldr) : grt::ModuleImplBase(ldr), Reporter(ldr->get_grt()) {}

  DEFINE_INIT_MODULE(WbModuleValidationMySQL_VERSION, "MySQL AB", ModuleImplBase,
                DECLARE_MODULE_FUNCTION(WbModuleValidationMySQLImpl::validateAll),                     
                DECLARE_MODULE_FUNCTION(WbModuleValidationMySQLImpl::validateIntegrity),               
                DECLARE_MODULE_FUNCTION(WbModuleValidationMySQLImpl::validateSyntax),                  
                DECLARE_MODULE_FUNCTION(WbModuleValidationMySQLImpl::validateDuplicatesAdditions), 

                DECLARE_MODULE_FUNCTION(WbModuleValidationMySQLImpl::getValidationDescription),
                     
                DECLARE_MODULE_FUNCTION(WbModuleValidationMySQLImpl::getPluginInfo));

   virtual int validateAll(GrtObjectRef root);
   virtual std::string getValidationDescription(GrtObjectRef root);
  
   int validateIntegrity(GrtObjectRef root);
   int validateSyntax(GrtObjectRef root);
   int validateDuplicatesAdditions(GrtObjectRef root);

   grt::ListRef<app_Plugin> getPluginInfo();
private:
};


#endif // _WB_MODULE_VALIDATION_H_

//!  @}
