//!
//!    \addtogroup GVal General Validation
//!             General validation module
//!
//!    @{
//!
#ifndef _WB_MODULE_VALIDATION_H__
#define _WB_MODULE_VALIDATION_H__

#include "wb_module_validation_public_interface.h"
#include "grtdb/db_object_helpers.h"
#include "grts/structs.db.h"
#include "interfaces/plugin.h"
#include "interfaces/wbvalidation.h"

#include "grts/structs.db.mgmt.h"
#include "grt/grt_walker.h"
#include "grt/grt_reporter.h"

#include "wb_validation.h"

using namespace bec;

#define WbModuleValidation_VERSION "1.0"

//=============================================================================
//
//=============================================================================
/**
  \brief General validator
*/
class GeneralValidator : public Validators
{
  public:
    GeneralValidator(ResultsList *rl, bec::Reporter *rep) 
                     : Validators(rl, rep)
                     {}                 

    //! Starts validation of specified catalog \a cat with Validation rules in \a val 
    //void validate(const db_CatalogRef& cat, Validation::Ptr val);
    void validate(const db_CatalogRef& cat, val::ChainsSet *cs);
    
    void add_efficiency_validations(val::ChainsSet *cs);
    void add_consistency_validations(val::ChainsSet *cs);
    void add_logic_validations(val::ChainsSet *cs);
    void add_empty_validations(val::ChainsSet *cs);
    void add_dup_validations(val::ChainsSet *cs);

    void walk_schema(const db_SchemaRef&) const;
    void walk_table(const db_TableRef&) const;
    void walk_view(const db_ViewRef&) const;
    void walk_routine(const db_RoutineRef&) const;
    void walk_routine_group(const db_RoutineGroupRef&) const;
    void walk_user(const db_UserRef&) const;
    void walk_column(const db_ColumnRef&) const;
    void walk_index(const db_IndexRef&) const;
    void walk_foreign_key(const db_ForeignKeyRef&) const;
    void walk_trigger(const db_TriggerRef&) const;
    void walk_role(const db_RoleRef&) const;

    void dup_check_table_name(const db_TableRef&) const;
    void dup_check_view_name(const db_ViewRef&) const;
    void dup_check_column_name(const db_ColumnRef&) const;
    void dup_check_index_name(const db_IndexRef&) const;
    void dup_check_trigger_name(const db_TriggerRef&) const;
    void dup_check_routine_name(const db_RoutineRef&) const;
    void dup_check_routine_group_name(const db_RoutineGroupRef&) const;
    void dup_check_role_name(const db_RoleRef&) const;
    void dup_check_user_name(const db_UserRef&) const;

    void efficiency_check_table(const db_TableRef& table) const;
    void efficiency_check_foreign_key(const db_ForeignKeyRef& table) const;

    void consistency_check_column(const db_ColumnRef& column) const;

    void logic_check_fk_refers_primary_key(const db_ForeignKeyRef& fk) const;
    void logic_check_table(const db_TableRef& table) const;

    void empty_check_table(const db_TableRef&) const;
    void empty_check_view(const db_ViewRef&) const;
    void empty_check_user(const db_UserRef&) const;
    void empty_check_routine(const db_RoutineRef&) const;
    void empty_check_role(const db_RoleRef&) const;
    void empty_check_routine_group(const db_RoutineGroupRef&) const;
    void empty_check_object_is_referenced_by_role(const char* type, const GrtNamedObjectRef&, const db_CatalogRef&) const;
    template <typename T>
    void empty_check_object_is_used_by_view(const T&, const workbench_physical_ModelRef& phys_model_) const;
    
    void item_check_passed(grt::GRT *g)
    {
      ++passedChecks;
      g->send_progress(passedChecks/totalItemsToCheck, "MySQL Validation");
    }
    void add_total_items_to_check(const int count)
    {
      totalItemsToCheck += count;
    }
        
  private:
    mutable float            totalItemsToCheck;
    mutable int              passedChecks;
    mutable Names            consistencyColumns;

    mutable Names            tableNames_;
    mutable Names            viewNames_;
    mutable Names            columnNames_;
    mutable Names            indexNames_;
    mutable Names            triggerNames_;
    mutable Names            routineNames_;
    mutable Names            routineGroupNames_;
    mutable Names            roleNames_;
    mutable Names            userNames_;

    val::ChainsSet           *val;
    db_CatalogRef               cat_;
    workbench_physical_ModelRef phys_model_;
};

//=============================================================================
//
//=============================================================================
/**
    \brief General validator module
*/
class WB_MODULE_VALIDATION_PUBLIC_FUNC WbModuleValidationImpl
  :  public grt::ModuleImplBase,
    public PluginInterfaceImpl,
    public WbValidationInterfaceImpl,
    protected Reporter,
    public CatalogIterator<WbModuleValidationImpl>
{
public:

  WbModuleValidationImpl(grt::CPPModuleLoader *ldr) : grt::ModuleImplBase(ldr), Reporter(ldr->get_grt()) {}

  DEFINE_INIT_MODULE(WbModuleValidation_VERSION, "MySQL AB", grt::ModuleImplBase,
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::validateAll),
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::validateEmptyContent),
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::validateTableEfficiency),
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::validateDuplicateIdentifiers),
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::validateConsistency),
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::validateLogic),

                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::getValidationDescription),
                DECLARE_MODULE_FUNCTION(WbModuleValidationImpl::getPluginInfo));

  //// All Validations
  virtual int validateAll(GrtObjectRef root);

  virtual std::string getValidationDescription(GrtObjectRef root);

  //// Empty Content Validation
  int validateEmptyContent(GrtObjectRef root);
  // Table Efficiency Validation
  int validateTableEfficiency(GrtObjectRef root);
  // Duplicated Identifiers Validation
  int validateDuplicateIdentifiers(GrtObjectRef root);
  // Consistency Validation
  int validateConsistency(GrtObjectRef root);
  // Logic Validation
  int validateLogic(GrtObjectRef root);

  virtual grt::ListRef<app_Plugin> getPluginInfo();
};


#endif // _WB_MODULE_VALIDATION_H_
//!
//!     @}
//!
