#include "wb_config.h"
#include "linux_utilities/plugin_editor_base.h"
#include "../backend/mysql_routinegroup_editor.h"
#include "linux_utilities/MGGladeXML.h"
#include "grtdb/db_object_helpers.h"
#include "treemodel_wrapper.h"
#include "sql_editor_fe.h"
#include "errors_list.h"
#include <sstream>
#include "menu_manager.h"

class MGGladeXML;
//static const char* dbgdomain = "MySQLRoutineGroupEditorFE";

//==============================================================================
//
//==============================================================================
class DbMySQLRoutineGroupEditor : public PluginEditorBase
{
  MySQLRoutineGroupEditorBE                 _be;
  MGGladeXML                               *_xml;
  ErrorsList                                _errors;
  SqlEditorFE                               _code;
  TrackedPane                              *_pane;
  Glib::RefPtr<Gtk::ListStore>              _routines_model;
  Gtk::TreeView                            *_rg_list;
  TextListColumnsModel                     *_routines_columns;
  
  virtual bec::BaseEditor *get_be();
  
  void parse_sql(const std::string& sql);
  void error_selected(const int lineno, const std::string& msg);

  bool process_event(GdkEvent* event);
  void menu_action_on_node(Gtk::MenuItem* item, const Gtk::TreePath path);

  void set_group_name(const std::string &);
  
 public:
  DbMySQLRoutineGroupEditor(bec::GRTManager *grtm, const grt::BaseListRef &args);
  
  virtual ~DbMySQLRoutineGroupEditor();
  virtual void do_refresh_form_data();
  virtual std::string get_title();

  void on_routine_drop(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y,
                       const Gtk::SelectionData& selection_data, guint info, guint time);
};

//------------------------------------------------------------------------------
DbMySQLRoutineGroupEditor::DbMySQLRoutineGroupEditor(bec::GRTManager *grtm, const grt::BaseListRef &args)
    : PluginEditorBase(grtm, args)
    , _be(grtm, db_mysql_RoutineGroupRef::cast_from(args[0]), get_rdbms_for_db_object(args[0]))
    , _errors(&_be)
    , _routines_model(model_from_string_list(std::vector<std::string>(), &_routines_columns))
{
  _xml= new MGGladeXML(grtm->get_data_file_path("modules/data/editor_rg.glade"));

  Gtk::Notebook *editor_window(0);
  _xml->get("mysql_rg_editor_notebook", &editor_window);

  Gtk::Image *image;
  _xml->get("rg_image", &image);
  image->set(ImageCache::get_instance()->image_from_filename("db.RoutineGroup.editor.48x48.png", false));
  
  _be.set_refresh_ui_slot(sigc::mem_fun(this, &DbMySQLRoutineGroupEditor::refresh_form_data));
  
  editor_window->reparent(*this);
  editor_window->show();

  Gtk::Entry* entry(0);
  _xml->get("rg_name", &entry);
  add_entry_change_timer(entry, sigc::mem_fun(this, &DbMySQLRoutineGroupEditor::set_group_name));
  
  Gtk::TextView* tv;
  _xml->get("rg_comment", &tv);
  add_text_change_timer(tv, sigc::mem_fun(&_be, &MySQLRoutineGroupEditorBE::set_comment));
  
  Gtk::VBox* code_win;
  _xml->get("rg_code_holder", &code_win);
  _code.widget().set_size_request(-1, 100);
  add_sqleditor_text_change_timer(&_code, sigc::mem_fun(this, &DbMySQLRoutineGroupEditor::parse_sql));
  code_win->add(_code.widget());
  code_win->resize_children();
  
  Gtk::ScrolledWindow* win(0);
  _xml->get("rg_errors", &win);
  win->add(_errors.widget());
  _errors.signal_error_selected().connect(sigc::mem_fun(this, &DbMySQLRoutineGroupEditor::error_selected));

  Gtk::Paned* pane(0);
  _xml->get("rg_pane", &pane);
  _pane = new TrackedPane(pane, &_be);

  refresh_form_data();
  
  _xml->get("rg_list", &_rg_list);
  
  _rg_list->set_model(_routines_model);
  _rg_list->append_column("Routine", _routines_columns->item);
  _rg_list->set_headers_visible(false);

  // Setup DnD
  std::vector<Gtk::TargetEntry> targets;  
  targets.push_back(Gtk::TargetEntry("db.mysql.Routine", Gtk::TARGET_SAME_APP));
  _rg_list->drag_dest_set(targets, Gtk::DEST_DEFAULT_ALL,Gdk::ACTION_COPY);
  _rg_list->signal_drag_data_received().connect(sigc::mem_fun(this, &DbMySQLRoutineGroupEditor::on_routine_drop));
  _rg_list->signal_event().connect(sigc::mem_fun(*this, &DbMySQLRoutineGroupEditor::process_event));

  _code.set_text(_be.get_routines_sql());

  show_all();
}

//------------------------------------------------------------------------------
DbMySQLRoutineGroupEditor::~DbMySQLRoutineGroupEditor()
{
  delete _pane;
  delete _xml;
}

//------------------------------------------------------------------------------
std::string DbMySQLRoutineGroupEditor::get_title()
{
  return strfmt(_("Group: %s"), _be.get_name().c_str());
}

//------------------------------------------------------------------------------

void DbMySQLRoutineGroupEditor::set_group_name(const std::string &name)
{
  _be.set_name(name);
  _signal_title_changed.emit(get_title());
}

//------------------------------------------------------------------------------
bec::BaseEditor *DbMySQLRoutineGroupEditor::get_be()
{
  return &_be;
}

//------------------------------------------------------------------------------
void DbMySQLRoutineGroupEditor::do_refresh_form_data()
{
  Gtk::Entry* entry(0);
  _xml->get("rg_name", &entry);
  if (entry->get_text() != _be.get_name())
  {
    entry->set_text(_be.get_name());
    _signal_title_changed.emit(_be.get_name());
  }
  Gtk::TextView* tv;
  _xml->get("rg_comment", &tv);
  tv->get_buffer()->set_text(_be.get_comment());
  
  recreate_model_from_string_list(_routines_model, _be.get_routines_names());
}

//------------------------------------------------------------------------------
void DbMySQLRoutineGroupEditor::parse_sql(const std::string& sql)
{
  _pane->max1st();
  _errors.clear_list();
  _be.set_routines_sql(sql);
  recreate_model_from_string_list(_routines_model, _be.get_routines_names());
}

//------------------------------------------------------------------------------
void DbMySQLRoutineGroupEditor::error_selected(const int lineno, const std::string& msg)
{}

//------------------------------------------------------------------------------
bool DbMySQLRoutineGroupEditor::process_event(GdkEvent* event)
{
  if ( event->type == GDK_BUTTON_PRESS && event->button.button == 3 )
  {
    Gtk::TreeModel::Path   path;
    Gtk::TreeView::Column *column(0);
    int                    cell_x(-1);
    int                    cell_y(-1);
    
    if ( _rg_list->get_path_at_pos((int)event->button.x, (int)event->button.y, path, column, cell_x, cell_y) )
    {
      bec::MenuItemList  menuitems;
      bec::MenuItem      item;                                                                                                            
      item.caption = "Remove routine from the group";                                                                                            
      item.name    = "remove_routine_from_the_group";                                                                             
      menuitems.push_back(item); 
      
      MenuManager::run_popup(menuitems
                             ,event->button.x
                             ,event->button.y
                             ,event->button.time
                             ,sigc::bind(sigc::mem_fun(this, &DbMySQLRoutineGroupEditor::menu_action_on_node), path)
                             );
    }
  }
  return false;
}

//------------------------------------------------------------------------------
void DbMySQLRoutineGroupEditor::menu_action_on_node(Gtk::MenuItem* item, const Gtk::TreePath path)
{
  const std::string item_name = item->get_name();
  
  if ( item_name == "remove_routine_from_the_group" )
  {
    const std::string name = (*(_routines_model->get_iter(path)))[_routines_columns->item];
    _be.delete_routine_with_name(name);
    do_refresh_form_data();
    _code.set_text(_be.get_routines_sql());
  }
}

//------------------------------------------------------------------------------
void DbMySQLRoutineGroupEditor::on_routine_drop(const Glib::RefPtr<Gdk::DragContext>& context
                                               ,int x, int y
                                               , const Gtk::SelectionData& selection_data
                                               , guint info, guint time)
{
  bool dnd_status = false;
  
  if ( selection_data.get_target() == "db.mysql.Routine" )
  {
    std::string type;
    std::string oid;

    const std::string selection = selection_data.get_data_as_string();
    // Selection is compiled by <typename>\n<oid>\n<typename>\n<oid>\n....
    std::istringstream is(selection);
    while ( !is.eof() )
    {
      if ( std::getline(is, type) )
      {
        if ( std::getline(is, oid) )
        {
          grt::ObjectRef obj = _be.get_grt()->find_object_by_id(oid,"/");
          if ( obj.is_valid() )
          {
            db_mysql_RoutineRef routine = db_mysql_RoutineRef::cast_from(obj);
            if ( routine.is_valid() )
            {
              _be.append_routine_with_id(routine.id());
              std::string routine_name = _be.get_routine_name(routine.id());
            }
          }
        }
      }
    }

    recreate_model_from_string_list(_routines_model, _be.get_routines_names());
    _code.set_text(_be.get_routines_sql());

    dnd_status = true;
  }
  context->drag_finish(dnd_status, false, time);
}

//------------------------------------------------------------------------------
extern "C" 
{
  GUIPluginBase *createDbMysqlRoutineGroupEditor(bec::GRTManager *grtm, const grt::BaseListRef &args)
  {
    return Gtk::manage(new DbMySQLRoutineGroupEditor(grtm, args));
  }
};
