
#include "grtui/grt_wizard_plugin.h"
#include "grtui/wizard_progress_page.h"
#include "grtui/wizard_finished_page.h"

#include "grtui/grtdb_object_filter.h"

#include "db_mysql_sql_export.h"


using namespace grtui;
using namespace mforms;


class ExportInputPage : public WizardPage
{
public:
  ExportInputPage(WizardForm *form)
    : WizardPage(form, "options"), _frame(true), _contents(true), _options(true), _options_box(false)
  {
    set_title(_("SQL Export Options"));
    set_subtitle(_("Set various options for SQL script to be exported."));

    _options.set_title(_("SQL Options"));

    _options.add(&_options_box);
    _options_box.set_padding(12);
    _options_box.set_spacing(8);

    _generate_drop_check.set_text(_("Generate DROP Statements Before Each CREATE Statement"));
    _options_box.add(&_generate_drop_check, false, false);
    _omit_schema_qualifier_check.set_text(_("Omit Schema Qualifier in Object Names"));
    _options_box.add(&_omit_schema_qualifier_check, false, false);
    _generate_create_index_check.set_text(_("Generate Separate CREATE INDEX Statements"));
    _options_box.add(&_generate_create_index_check, false, false);
    _generate_show_warnings_check.set_text(_("Add SHOW WARNINGS After Every DDL Statement"));
    _options_box.add(&_generate_show_warnings_check, false, false);
    _skip_users_check.set_text(_("Do Not Create Users. Only Export Privileges"));
    _options_box.add(&_skip_users_check, false, false);
    _generate_insert_check.set_text(_("Generate INSERT Statements for Tables"));
    _options_box.add(&_generate_insert_check, false, false);

    add(&_options, false, false);

    add(&_frame, false, true);

    _frame.add(&_contents);
    _contents.add(&_caption, false, false);
    _contents.add(&_filename, true, true);
    _contents.add(&_browse_button, false, false);
    
    enable_file_browsing(&_filename, &_browse_button, mforms::SaveFile);
    
    _caption.set_text(_("Output SQL Script File:"));
    _browse_button.set_text(_("Browse..."));

    _filename.signal_changed().connect(sigc::mem_fun(this, &ExportInputPage::file_changed));

    _contents.set_padding(12);
    _contents.set_spacing(8);

    _frame.set_title(_("Output File"));
    
    signal_leave().connect(sigc::mem_fun(this, &ExportInputPage::gather_options));
  }
  

  void file_changed()
  {
    validate();
  }

  virtual bool allow_next()
  {
    return !_filename.get_string_value().empty();
  }

  void gather_options(grt::DictRef values)
  {
    values.gset("OutputFileName", _filename.get_string_value());
    
    values.gset("GenerateDrops", _generate_drop_check.get_active());
    values.gset("GenerateWarnings", _generate_show_warnings_check.get_active());
    values.gset("GenerateCreateIndex", _generate_create_index_check.get_active());
    values.gset("NoUsersJustPrivileges", _skip_users_check.get_active());
    values.gset("GenerateInserts", _generate_insert_check.get_active());
    values.gset("OmitSchemata", _omit_schema_qualifier_check.get_active());
  }

protected:
  Panel _frame;
  Box _contents;
  Label _caption;
  TextEntry _filename;
  Button _browse_button;

  Panel _options;
  Box _options_box;

  CheckBox _generate_drop_check;
  CheckBox _generate_create_index_check;
  CheckBox _generate_show_warnings_check;
  CheckBox _skip_users_check;
  CheckBox _generate_insert_check;
  CheckBox _omit_schema_qualifier_check;
};

//--------------------------------------------------------------------------------


class ExportFilterPage : public WizardPage
{
public:
  ExportFilterPage(WizardForm *form)
    : WizardPage(form, "filter"), _export_be(form->grtm()), _scroll_panel(false), _box(false)
  {
    set_title(_("SQL Export Filter"));
    set_subtitle(_("Select objects to be exported."));

    add(&_scroll_panel, true, true);
    _scroll_panel.add(&_box);

    setup_filters();
  }

  virtual ~ExportFilterPage()
  {
   
  }

  std::string get_summary()
  {    
    return _summary;
  }
  

protected:
  void setup_filters()
  {
    bec::GrtStringListModel *users_model;
    bec::GrtStringListModel *users_imodel;
    bec::GrtStringListModel *tables_model;
    bec::GrtStringListModel *tables_imodel;
    bec::GrtStringListModel *views_model;
    bec::GrtStringListModel *views_imodel;
    bec::GrtStringListModel *routines_model;
    bec::GrtStringListModel *routines_imodel;
    bec::GrtStringListModel *triggers_model;
    bec::GrtStringListModel *triggers_imodel;

    _export_be.setup_grt_string_list_models_from_catalog(&users_model, &users_imodel, 
                                                       &tables_model, &tables_imodel,
                                                       &views_model, &views_imodel, 
                                                       &routines_model, &routines_imodel, 
                                                       &triggers_model, &triggers_imodel);
    

    _table_filter=
      add_filter_type(db_mysql_Table::static_class_name(),
                    tables_model, tables_imodel);

    _view_filter=
      add_filter_type(db_mysql_View::static_class_name(),
                    views_model, views_imodel);

    _routine_filter=
      add_filter_type(db_mysql_Routine::static_class_name(),
                    routines_model, routines_imodel);

    _trigger_filter=
      add_filter_type(db_mysql_Trigger::static_class_name(),
                    triggers_model, triggers_imodel);

    _user_filter=
      add_filter_type(db_User::static_class_name(),
                    users_model, users_imodel);
  }
  
  
  DBObjectFilterFrame *add_filter_type(const std::string &class_name, 
                       bec::GrtStringListModel *model,
                       bec::GrtStringListModel *excl_model)
  {
    DBObjectFilterFrame *filter;
    
    filter= new DBObjectFilterFrame(_form->grtm());
    filter->set_object_class(class_name, _("Export %s Objects"));
    filter->set_models(model, excl_model);
    _box.add(filter, false, true);

    return filter;
  }
  
  
  virtual void execute()
  {
    _export_be.set_option("OutputFileName", values().get_string("OutputFileName"));
    _export_be.set_option("GenerateDrops", values().get_int("GenerateDrops"));
    _export_be.set_option("GenerateWarnings", values().get_int("GenerateWarnings"));
    _export_be.set_option("GenerateCreateIndex", values().get_int("GenerateCreateIndex"));
    _export_be.set_option("NoUsersJustPrivileges", values().get_int("NoUsersJustPrivileges"));
    _export_be.set_option("GenerateInserts", values().get_int("GenerateInserts"));
    _export_be.set_option("OmitSchemata", values().get_int("OmitSchemata"));

    _summary= strfmt(_("The script was exported to the file '%s'"),
                     values().get_string("OutputFileName").c_str());

    _done= false;
  }
  
  virtual bool override_next_action() { return true; }

  virtual bool execute_overriden_next_action()
  {
    if (!_done)
    {
      _export_be.set_option("TablesAreSelected", _table_filter->get_active());
      _export_be.set_option("TriggersAreSelected", _trigger_filter->get_active());
      _export_be.set_option("RoutinesAreSelected", _routine_filter->get_active());
      _export_be.set_option("ViewsAreSelected", _view_filter->get_active());
      _export_be.set_option("UsersAreSelected", _user_filter->get_active());

      _export_be.start_export();
      _done= true;

      _form->go_to_next();
      return true;
    }
    return false;
  }

  virtual std::string next_button_caption()
  {
    if (_done)
      return "";
    return _("_Execute >");
  }
  
protected:
  DbMySQLSQLExport _export_be;
  mforms::ScrollPanel _scroll_panel;
  mforms::Box _box;
  DBObjectFilterFrame *_table_filter;
  DBObjectFilterFrame *_view_filter;
  DBObjectFilterFrame *_routine_filter;
  DBObjectFilterFrame *_trigger_filter;
  DBObjectFilterFrame *_user_filter;

  std::string _summary;

  bool _done;
};


//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------

class WbPluginSQLExport : public WizardPlugin
{
  ExportInputPage *_input_page;
  ExportFilterPage *_filter_page;
  WizardFinishedPage *_finish_page;
  
public:
  WbPluginSQLExport(bec::GRTManager *grtm, const grt::BaseListRef &args)
    : WizardPlugin(grtm, args)
  {
    _input_page= new ExportInputPage(this);
    _filter_page= new ExportFilterPage(this);
    _finish_page= new WizardFinishedPage(this, _("SQL Export Finished"));
    add_page(_input_page);
    add_page(_filter_page);
    add_page(_finish_page);
    
    _filter_page->signal_leave().connect(sigc::hide(sigc::mem_fun(this, &WbPluginSQLExport::update_summary)));

    set_title(_("Forward Engineer SQL Script"));
  }

  void update_summary()
  {
    _finish_page->set_title(_("SQL Export Finished Successfully"));
    _finish_page->set_summary(_filter_page->get_summary());
  }
};






extern "C" 
{
  GUIPluginBase *createWBPluginSQLExport(bec::GRTManager *grtm, const grt::BaseListRef &args)
  {
    return new WbPluginSQLExport(grtm, args);
  }
};
