
#include "grtui/grt_wizard_plugin.h"
#include "grtui/wizard_progress_page.h"
#include "grtui/wizard_finished_page.h"

#include "db_rev_eng_be.h"


using namespace grtui;
using namespace mforms;


namespace ScriptImport {

class ImportInputPage : public WizardPage
{
public:
  ImportInputPage(WizardForm *form)
    : WizardPage(form, "options"), _frame(true), _contents(true), _options(true), _options_box(false)
  {
    set_title(_("SQL Import Options"));
    set_subtitle(_("Set options for SQL script to be imported."));

    add(&_frame, false, true);

    _frame.add(&_contents);
    _contents.add(&_caption, false, false);
    _contents.add(&_filename, true, true);
    _contents.add(&_browse_button, false, false);
    
    enable_file_browsing(&_filename, &_browse_button, mforms::OpenFile);
    
    _caption.set_text(_("Input SQL Script File:"));
    _browse_button.set_text(_("Browse..."));

    _filename.signal_changed().connect(sigc::mem_fun(this, &ImportInputPage::file_changed));

    _contents.set_padding(12);
    _contents.set_spacing(8);

    _frame.set_title(_("Input File"));

    _options.set_title(_("Options"));

    _options.add(&_options_box);
    _options_box.set_padding(12);
    _options_box.set_spacing(8);
    _options_box.add(&_autoplace_check, false, false);
    _autoplace_check.set_text(_("Place Imported Objects on a New Diagram"));
    _autoplace_check.set_active(true);

    add(&_options, false, false);

    signal_leave().connect(sigc::mem_fun(this, &ImportInputPage::gather_options));
  }
  

  void file_changed()
  {
    validate();
  }


  virtual void do_validate()
  {
    if (!_filename.get_string_value().empty())
    {
      if (!g_file_test(_filename.get_string_value().c_str(), (GFileTest)(G_FILE_TEST_EXISTS|G_FILE_TEST_IS_REGULAR)))
      {
        _form->set_problem(_("Invalid path"));
        return;
      }
    }
    _form->clear_problem();
  }


  virtual bool allow_next()
  {
    return !_filename.get_string_value().empty();
  }


  void gather_options(grt::DictRef values)
  {
    values.gset("import.filename", _filename.get_string_value());
    values.gset("import.place_figures", _autoplace_check.get_active());
  }

protected:
  Panel _frame;
  Box _contents;
  Label _caption;
  TextEntry _filename;
  Button _browse_button;

  Panel _options;
  Box _options_box;
  CheckBox _autoplace_check;
};


//--------------------------------------------------------------------------------

class ImportProgressPage : public WizardProgressPage
{
  Sql_import _import_be;
  TaskRow *_auto_place_task;
  bool _auto_place;
  bool _done;
  
public:
  ImportProgressPage(WizardForm *form) : WizardProgressPage(form, "progress")
  {
    set_title(_("SQL Import Progress"));
    set_subtitle(_("Progress of SQL script reverse engineering."));

    _auto_place= false;
    _done= false;
    
    _import_be.grtm(form->grtm());

    //showAdvancedImportLogPage = false;
    //caseSensitiveNamesCB.Checked = sqlImport.get_bool_option("SqlIdentifiersCS");


    TaskRow *task=
      add_async_task(_("Reverse Engineer Objects"), 
                     sigc::mem_fun(this, &ImportProgressPage::import_objects),
                     _("Reverse engineering and importing objects from script..."));
    task->process_finish= sigc::mem_fun(this, &ImportProgressPage::import_objects_finished);

    add_task(_("Verify Results"), 
             sigc::mem_fun(this, &ImportProgressPage::verify_results),
             _("Verifying imported objects..."));

    _auto_place_task= 
      add_async_task(_("Place Objects on Diagram"), 
                     sigc::mem_fun(this, &ImportProgressPage::place_objects),
                     _("Placing imported objects on a new diagram..."));

    end_adding_tasks(true,
                     _("Import finished successfully."));

    set_status_text("");
  }

  
  void import_objects_finished(grt::ValueRef value)
  {
    _form->grtm()->get_grt()->send_info(grt::StringRef::cast_from(value));
  }


  bool import_objects()
  {
    execute_grt_task(_import_be.get_task_slot(), false);
    return true;
  }


  bool verify_results()
  {
    return true;
  }


  bool place_objects()
  {
    if (_auto_place)
    {
      execute_grt_task(_import_be.get_autoplace_task_slot(), false);
      return true;
    }
    return false;
  }


  virtual void execute()
  {
    _import_be.sql_script(values().get_string("import.filename"));

    _auto_place= values().get_int("import.place_figures");

    _auto_place_task->set_enabled(_auto_place);

    // sqlImport.set_option("case_sensitive_names", caseSensitiveNamesCB.Checked);
     // sqlImport.set_option("processing_create_statements", procCreateStmtCB.Checked);
     // sqlImport.set_option("processing_alter_statements", procAlterStmtCB.Checked);
    //sqlImport.set_option("processing_drop_statements", procDropStmtCB.Checked);

    WizardProgressPage::execute();
  }


  std::string get_summary()
  {
    std::string summary;
    int schemas= 0, tables= 0, views= 0, procedures= 0;
    
    grt::ListRef<GrtObject> created_objects= _import_be.get_created_objects();
    
    for (grt::ListRef<GrtObject>::const_iterator iter= created_objects.begin();
         iter != created_objects.end(); ++iter)
    {
      if ((*iter).is_instance<db_Schema>())
        schemas++;
      else if ((*iter).is_instance<db_Table>())
        tables++;
      else if ((*iter).is_instance<db_View>())
        views++;
      else if ((*iter).is_instance<db_Routine>())
        procedures++;
    }

    summary= strfmt(_("Import of SQL script file '%s' has finished successfully.\n\n"
                      "%i tables, %i views and %i stored procedures were imported in %i schemas.\n\n"),
                    _import_be.sql_script().c_str(),
                    tables, views, procedures, schemas);

    if (_got_error_messages)
      summary.append(_("There were errors during the import.\n"));
    else if (_got_warning_messages)
      summary.append(_("There were warnings during the import.\n"));

    summary.append(_("Press [< Back] to review the logs."));
  

    return summary;
  }
};


//--------------------------------------------------------------------------------

class WbPluginSQLImport : public WizardPlugin
{
  ImportInputPage *_input_page;
  ImportProgressPage *_progress_page;
  WizardFinishedPage *_finish_page;
  
public:
  WbPluginSQLImport(bec::GRTManager *grtm, const grt::BaseListRef &args)
    : WizardPlugin(grtm, args)
  {
    _input_page= new ImportInputPage(this);
    _progress_page= new ImportProgressPage(this);
    
    _progress_page->signal_tasks_finished().
      connect(sigc::mem_fun(this, &WbPluginSQLImport::update_summary));
    
    _finish_page= new WizardFinishedPage(this, _("SQL Import Finished"));

    add_page(_input_page);
    add_page(_progress_page);
    add_page(_finish_page);
    
    set_title(_("Reverse Engineer SQL Script"));
  }

  void update_summary(bool success)
  {
    _finish_page->set_title(success ? _("SQL Import Finished Successfully")
                            : _("SQL Import Failed"));
    _finish_page->set_summary(_progress_page->get_summary());
  }
};


}; // namespace ScriptImport



extern "C" 
{
  GUIPluginBase *createWBPluginSQLImport(bec::GRTManager *grtm, const grt::BaseListRef &args)
  {
    return new ScriptImport::WbPluginSQLImport(grtm, args);
  }
};
